package felix.compiler;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;

import tuffy.mln.Literal;
import tuffy.mln.Term;
import tuffy.ra.Expression;
import tuffy.ra.Function;
import tuffy.util.ExceptionMan;
import tuffy.util.StringMan;
import tuffy.util.Timer;


import felix.dstruct.FelixClause;
import felix.dstruct.FelixPredicate;
import felix.dstruct.FelixQuery;
import felix.dstruct.FelixPredicate.FPProperty;
import felix.dstruct.StatOperator.OPType;
import felix.parser.FelixCommandOptions;

/**
 * The class of static analyzer that parses a given 
 * MLN program. It takes as input
 * the whole MLN program, and assign properties to each 
 * predicate. Properties include
 * whether a predicate is SYMM, REFLEX, TRANS etc., or 
 * whether a clause specifies a CRF
 * chain etc.
 * 
 * @author Ce Zhang
 *
 */
public class StaticAnalyzer {

	/**
	 * Felix Query.
	 */
	FelixQuery fq;
	
	/**
	 * Command line options.
	 */
	FelixCommandOptions options;
	
	/**
	 * The constructor.
	 * @param _fq
	 * @param _opt
	 */
	public StaticAnalyzer(FelixQuery _fq, FelixCommandOptions _opt){
		fq = _fq;
		options = _opt;
	}
	
		
	/**
	 * Analyze the input MLN program and assign properties to each predicate.
	 */
	public void parse(){
		
		//classify clauses and register them to predicates
		//Note: the invocation order matters!
		this.parseKeyConstraintRelation();
		this.parseReflexiveRelation();
		this.parseSymmetricRelation();
		this.parseTransitiveRelation();
		this.parseChainRecursiveRelation();
		this.parseNonRecursiveRelation();
		this.parseOtherRecursiveRelation();
		this.parseSpecialPredicate();
		this.parseEmbededWeightRule();
		//this.parseSymmetricRelationPi2P();
	}
	
	/**
	 * Parse the predicate serves as the linear-representation of COREF.
	 */
	public void parseSpecialPredicate(){
		
		for(FelixPredicate p : fq.getAllPred()){
			String name = p.getName();
			if(name.endsWith("_map")){
			
				String oriName = name.replaceAll("_map$", "");
				
				if(fq.getPredByName(oriName) == null){
					ExceptionMan.die("Cannot use predicate P_map without predicate P, please change to another name.");
				}
				
				FelixPredicate orip = fq.getPredByName(oriName);
				
				if(orip.isClosedWorld()){
					ExceptionMan.die("Cannot use predicate P with predicate P as closed. Please change P to co-ref predicate, or" +
							" change the name of P_map");
				}
				
				if(!p.isClosedWorld()){
					ExceptionMan.die("Cannot use predicate P_map with open-world setting. Plase change P_map to closed-world, or" +
							" change the name of P_map");
				}
				
				orip.mustbe = OPType.COREF;
				p.oriCorefPredicate = orip;
				p.isCorefMapPredicate = true;
				
				orip.corefMAPPredicate = p;
			
			}
		}
		
	}
	
	/**
	 * Parse predicate with key constraints.
	 */
	public void parseKeyConstraintRelation(){
		
		// get key constraint from predicate declaration
		for(FelixPredicate fp : fq.getAllPred()){
			if(fp.hasDependentAttributes()){
				
				Object[] _labelpos = fp.getLabelAttrPositions().toArray();
				int[] labelpos = new int[_labelpos.length];
				
				for(int i=0;i<_labelpos.length;i++){
					labelpos[i] = (Integer) _labelpos[i];
				}
				
				fp.registerProperty(
						FelixPredicate.FPProperty.KEY_CONSTRAINT, 
						null, 
						labelpos);
			}
		}
		
		// discover rules like R(a,b1), [b1!=b2], => !R(a,b2)
		for(FelixClause fc : fq.getAllClause()){
			
			if(fc.getRegLiterals().size() != 2 || 
					fc.isHardClause() == false){
				continue;
			}
			
			HashSet<FelixPredicate> processedPred = new HashSet<FelixPredicate>();
			
			for(Literal l : fc.getRegLiterals()){
			
				FelixPredicate p = fq.getPredByName(l.getPred().getName());
				
				if(processedPred.contains(p)){
					continue;
				}
				
				processedPred.add(p);
				
				if(p.isClosedWorld()){
					continue;
				}
				
				Literal l1 = fc.getRegLiterals().get(0);
				Literal l2 = fc.getRegLiterals().get(1);
				
				if(fc.getConstraints().size() != 1 || 
						l1.getPred().equals(l2.getPred()) == false){
					continue;
				}
				
				Expression e = fc.getConstraints().get(0);
				String var1 = "", var2 = "";
				
				boolean eqNeqFlag = false;
				if(e.getFunction() == Function.NOT){
					e = e.getArgs().get(0);
					eqNeqFlag = true;
				}
					
				if (  ( eqNeqFlag == false && e.getFunction() == Function.Neq ) ||
						( eqNeqFlag == true && e.getFunction() == Function.Eq ) ){
					
					if(l1.getSense() == true || l2.getSense() == true){
						continue;
					}
					
					if(e.getArgs().size() != 2){
						continue;
					}
					
					Expression e1 = e.getArgs().get(0);
					Expression e2 = e.getArgs().get(1);
					
					if(e1.getFunction() != Function.VariableBinding 
							|| e2.getFunction() != Function.VariableBinding){
						continue;
					}
					
					var1 = e1.toString();
					var2 = e2.toString();
				}else{
					continue;
				}
				
				//check
				ArrayList<Term> ta = l1.getTerms();
				ArrayList<Term> tb = l2.getTerms();
				
				if(ta.size() != tb.size()) continue;
				
				boolean isLR = true;
				boolean isFirst = false;
				int firstI = -1;
				ArrayList<Integer> __labelpos = new ArrayList<Integer>();
				for(int i=0;i<ta.size();i++){
					if(ta.get(i).toString().equals(tb.get(i).toString()) == false){
					
						if( (ta.get(i).toString().equals(var1) && tb.get(i).toString().equals(var2) && isFirst ==false) || 
								(tb.get(i).toString().equals(var1) && ta.get(i).toString().equals(var2) && isFirst ==false) ){
							firstI = i;
							isFirst = true;
						}else{
							isLR = false;
						}
						
						__labelpos.add(i);
						
					}else{
						//__keypos.add(i);
					}
				}
				
				if(firstI == -1 || isLR == false){
					continue;
				}
				
				Object[] _labelpos = __labelpos.toArray();
				int[] labelpos = new int[_labelpos.length];
				
				for(int i=0;i<_labelpos.length;i++){
					labelpos[i] = (Integer) _labelpos[i];
				}
				
				p.registerProperty(
						FelixPredicate.FPProperty.KEY_CONSTRAINT, 
						fc, 
						labelpos);	
			}
		}
		
	}
	
	/**
	 * Generates all permutations of the given terms.
	 * @param terms
	 * @return
	 */
	public HashSet<ArrayList<String>> generateAllPermutations(HashSet<String> terms){
		HashSet<ArrayList<String>> ret = new HashSet<ArrayList<String>>();
		
		if(terms.size() == 1){
			ArrayList<String> tmp = new ArrayList<String>();
			tmp.add(terms.iterator().next());
			ret.add(tmp);
			return ret;
		}
		
		for(String first : terms){
			
			HashSet<String> newTerms = (HashSet<String>) terms.clone();
			newTerms.remove(first);
			
			HashSet<ArrayList<String>> others = this.generateAllPermutations(newTerms);
			for(ArrayList<String> a : others){
				a.add(first);
				ret.add(a);
			}
		}
		
		return ret;
	}
	

	/**
	 * Parse predicate which is symmetric.
	 */
	public void parseSymmetricRelation(){
		
		for(FelixClause fc : fq.getAllClause()){
			
			if(fc.getRegLiterals().size() != 2 || 
					fc.isHardClause() == false){
				continue;
			}
			
			HashSet<FelixPredicate> processedPred = new HashSet<FelixPredicate>();
			
			for(Literal l : fc.getRegLiterals()){
			
				FelixPredicate p = fq.getPredByName(l.getPred().getName());
				
				if(processedPred.contains(p)){
					continue;
				}
				
				processedPred.add(p);
				
				if(p.isClosedWorld()){
					continue;
				}
		
				Literal l1 = fc.getRegLiterals().get(0);
				Literal l2 = fc.getRegLiterals().get(1);
				
				if(!l1.getPred().equals(l2.getPred())){
					continue;
				}
				
				if(p.arity() != 2){
					continue;
				}
				
				if(l1.getSense() != l2.getSense()){
					if(l1.getTerms().get(0).toString().equals(l2.getTerms().get(1).toString()) && 
							l1.getTerms().get(1).toString().equals(l2.getTerms().get(0).toString()) && 
							!l1.getTerms().get(1).toString().equals(l1.getTerms().get(0).toString())){
						p.registerProperty(
								FPProperty.SYMM, 
								fc);
					}
				}
				
			}
		}
		
	}
	
	/**
	 * Parse predicate which is reflexive.
	 */
	public void parseReflexiveRelation(){
		
		for(FelixClause fc : fq.getAllClause()){
			
			if(fc.getRegLiterals().size() != 1 || 
					fc.isHardClause() == false){
				continue;
			}
			
			HashSet<FelixPredicate> processedPred = new HashSet<FelixPredicate>();
			
			for(Literal l : fc.getRegLiterals()){
			
				FelixPredicate p = fq.getPredByName(l.getPred().getName());
				
				if(processedPred.contains(p)){
					continue;
				}
				
				processedPred.add(p);
				
				if(p.isClosedWorld()){
					continue;
				}
				
				if(l.getTerms().size() == 2 && 
						(l.getSense() == true)){
					if(l.getTerms().get(0).var() != null && 
							l.getTerms().get(0).var().equals(l.getTerms().get(1).var())){
						p.registerProperty(
								FPProperty.REFLEX, 
								fc);
					}
				}

			}
		}
	}
	
	/**
	 * Parse predicate which is transitive.
	 */
	public void parseTransitiveRelation(){
		
		for(FelixClause fc : fq.getAllClause()){
			
			if(fc.getRegLiterals().size() != 3 || 
					fc.isHardClause() == false){
				continue;
			}
			
			HashSet<FelixPredicate> processedPred = new HashSet<FelixPredicate>();
			
			for(Literal l : fc.getRegLiterals()){
			
				FelixPredicate p = fq.getPredByName(l.getPred().getName());
				
				if(processedPred.contains(p)){
					continue;
				}
				
				processedPred.add(p);
				
				if(p.arity() != 2){
					continue;
				}
				
				if(p.isClosedWorld()){
					continue;
				}		
				
				if(fc.getRegLiterals().size() == 3){
					
					ArrayList<Literal> ls = new ArrayList<Literal>();
					ls.add(fc.getRegLiterals().get(0));
					ls.add(fc.getRegLiterals().get(1));
					ls.add(fc.getRegLiterals().get(2));
					
					if(!ls.get(0).getPred().equals(ls.get(1).getPred()) ||
							!ls.get(2).getPred().equals(ls.get(1).getPred()) ||
							!ls.get(0).getPred().equals(ls.get(2).getPred())){
						continue;
					}
				
					int head = -1;
					int[] aa = new int[2];
					int ct = 0;
					for(int i=0;i<3;i++){
						if(ls.get(i).getSense() == true){
							if(head != -1){
								head = -1;
								break;
							}
							head = i;
						}else{
							if(ct < 2){
								aa[ct++] = i; 
							}
						}
					}
					
					if(head == -1){
						continue;
					}
					
					Literal l1 = ls.get(aa[0]);
					Literal l2 = ls.get(aa[1]);
					Literal h = ls.get(head);
					
					String common = "";
					String o1 = "";
					String o2 = "";
					for(int i=0;i<2;i++){
						for(int j=0;j<2;j++){
							if(l1.getTerms().get(i).toString().equals(l2.getTerms().get(j).toString())){
								common = l1.getTerms().get(i).toString();
								o1 = l1.getTerms().get((i+1)%2).toString();
								o2 = l2.getTerms().get((j+1)%2).toString();
								break;
							}
						}
						if(common.equals("") == false){
							break;
						}
					}
					
					if(common.equals("")){
						continue;
					}
					
					if( h.getTerms().get(0).toString().equals(o1) && h.getTerms().get(1).toString().equals(o2) ){
						p.registerProperty(
								FPProperty.TRANS,
								fc);
					}
					
					if( h.getTerms().get(1).toString().equals(o1) && h.getTerms().get(0).toString().equals(o2) ){
						p.registerProperty(
								FPProperty.TRANS,
								fc);
					}
				}
			}
		}
		
	}
	
	/**
	 * Parse clause which specifies a CRF chain rule.
	 */
	public void parseChainRecursiveRelation(){
		
		for(FelixClause fc : fq.getAllClause()){
			
			HashSet<FelixPredicate> processedPred = new HashSet<FelixPredicate>();
			
			for(Literal l : fc.getRegLiterals()){
			
				FelixPredicate p = fq.getPredByName(l.getPred().getName());
				
				if(processedPred.contains(p)){
					continue;
				}
				
				processedPred.add(p);
				
				if(p.isClosedWorld()){
					continue;
				}

				if(fc.getLiteralsOfPredicate(l.getPred()).size() != 2){
					continue;
				}
				
				Literal l1 = fc.getLiteralsOfPredicate(l.getPred()).get(0);
				Literal l2 = fc.getLiteralsOfPredicate(l.getPred()).get(1);
			
				boolean isCRFRule = true;
							
				int numberOfDifference = 0;
				for(int i=0;i<l1.getTerms().size();i++){
					Term t1 = l1.getTerms().get(i);
					Term t2 = l2.getTerms().get(i);
	
					if(t1.toString().equals(t2.toString())){
						// if these two terms are the same, good!
						continue;
					}else{
						// if not, we need to see whether in this rule,
						// t1 and t2 is representing a chain. Here we know 
						// there can be only one label-field, so, only one 
						// such t1-t2 pair can exist.
						
						if(!p.getKeyPositions().contains(i)){
							continue;
						}
						
						numberOfDifference ++;
										
						boolean isChain = false;
						
						// find in condition [t1 = t2 + 1]
						//TODO: We may want to change it smarter in the future
						for(Expression e : fc.getConstraints()){
							if(e.toString().equals(t1+" = (" + t2 + " + 1)") ||
								e.toString().equals(t2+" = (" + t1 + " + 1)") ||
								e.toString().equals("(" + t1 + " + 1) = " + t2) ||
								e.toString().equals("(" + t2 + " + 1) = " + t1) ||
								e.toString().equals(t1+" = (" + t2 + " - 1)") ||
								e.toString().equals(t2+" = (" + t1 + " - 1)") ||
								//TODO
								e.toString().equals("(" + t1 + " - 1) = (" + t2 + " + 0)") ||
								e.toString().equals("(" + t2 + " - 1) = (" + t1 + " + 0)") ||
								e.toString().equals("(" + t1 + " - " + t2 + ") = " + 1) ||
								e.toString().equals(1 + " = (" + t1 + " - " + t2 + ")") || 
								e.toString().equals("(" + t2 + " - " + t1 + ") = " + 1) ||
								e.toString().equals(1 + " = (" + t2 + " - " + t1 + ")")){
									isChain = true;
							}
						}
						
						if(!isChain){
							isCRFRule = false;
						}
										
						if(numberOfDifference != 1){
							isCRFRule = false;
						}
					}
				}
				
				if(isCRFRule == true && numberOfDifference > 0){
					if(p.getKeyPositions().size() != 0){
						p.registerProperty(
								FPProperty.CHAIN_RECUR, 
								fc);
					}
				}
				
			}
		}
		
	}
	
	// rules other than special rules
	/**
	 * Parse clause which does NOT specify 1) NON-RECURSIVE rule and 
	 * 2) Chain rule.
	 * 
	 */
	public void parseOtherRecursiveRelation(){
		for(FelixClause fc : fq.getAllClause()){
			
			HashSet<FelixPredicate> processedPred = new HashSet<FelixPredicate>();
			
			HashSet<FelixPredicate> appeared = new HashSet<FelixPredicate>();
			
			for(Literal l : fc.getRegLiterals()){
				FelixPredicate p = fq.getPredByName(l.getPred().getName());
				
				if(p.isClosedWorld()){
					continue;
				}
				
				appeared.add(p);
			}
			
			
			
			
			for(Literal l : fc.getRegLiterals()){
			
				FelixPredicate p = fq.getPredByName(l.getPred().getName());
			
				if(p.isClosedWorld()){
					continue;
				}
				
				if(processedPred.contains(p)){
					continue;
				}
				
				processedPred.add(p);
				
				if(appeared.size() == 1){
					if(fc.getLiteralsOfPredicate(l.getPred()).size() > 1){
						p.registerProperty(
								FPProperty.OTHER_RECUR, 
								fc);
					}
				}else{
					if(fc.getLiteralsOfPredicate(l.getPred()).size() > 1){
						p.registerProperty(
								FPProperty.OTHER_RECUR_WITHOTHER_OPENPRED, 
								fc);
					}
				}

			}
		}
	}
	
	
	/**
	 * Parse clause which specifies non-recursive rules.
	 */
	public void parseNonRecursiveRelation(){

		for(FelixClause fc : fq.getAllClause()){
			
			HashSet<FelixPredicate> processedPred = new HashSet<FelixPredicate>();
			
			for(Literal l : fc.getRegLiterals()){
			
				FelixPredicate p = fq.getPredByName(l.getPred().getName());
			
				if(p.isClosedWorld()){
					continue;
				}
				
				if(processedPred.contains(p)){
					continue;
				}
				
				processedPred.add(p);
				
				if(fc.getLiteralsOfPredicate(l.getPred()).size() == 1){
					p.registerProperty(
							FPProperty.NON_RECUR, 
							fc);
				}

			}
		}
	}
	
	/**
	 * Parse clause whose weights are specified by embedded weight.
	 */
	public void parseEmbededWeightRule(){

		for(FelixClause fc : fq.getAllClause()){
			
			for(Literal l : fc.getRegLiterals()){
			
				FelixPredicate p = fq.getPredByName(l.getPred().getName());
			
				if(p.isClosedWorld()){
					continue;
				}
				
				if(fc.hasEmbeddedWeight()){
					p.registerProperty(FPProperty.EMBED_WEIGHT_RULE, fc);
				}

			}
			
		}
	}
	
}
