package felix.dstruct;


import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import felix.dstruct.StatOperator.OPType;
import felix.main.Felix;
import felix.parser.FelixCommandOptions;
import felix.society.TaskList;
import felix.society.TaskSet;
import felix.task.ExecuteOperatorTask;
import felix.task.OptimizeDMOTask;
import felix.util.FelixConfig;


import tuffy.db.RDB;
import tuffy.infer.DataMover;
import tuffy.mln.Clause;
import tuffy.mln.Literal;
import tuffy.mln.MarkovLogicNetwork;
import tuffy.mln.Predicate;
import tuffy.mln.Term;
import tuffy.mln.Type;
import tuffy.ra.ConjunctiveQuery;
import tuffy.ra.Expression;
import tuffy.util.Config;
import tuffy.util.ExceptionMan;
import tuffy.util.FileMan;
import tuffy.util.StringMan;
import tuffy.util.UIMan;


/**
 * An object of ConcurrentOperatorsBucket contains 
 * multiple statistical operators that can be executed in
 * parallel. ConcurrentOperatorsBucket is the basic unit
 * of the {@link ExecutionPlan}.
 */
public class ConcurrentOperatorsBucket extends Thread{

	/**
	 * Set of statistical operators in this bucket.
	 */
	HashSet<StatOperator> concurrentOperators = new HashSet<StatOperator>();
	
	/**
	 * When executing this bucket, this variable records the remaining operators.
	 */
	ArrayList<StatOperator> operatorStacks = new ArrayList<StatOperator>();
	
	/**
	 * Predicates used as inputs.
	 */
	public HashSet<FelixPredicate> inputPredicates = new HashSet<FelixPredicate>();
	
	/**
	 * Auxiliary structure for clearing the old data table and
	 * merge the output of statistical operators in this bucket.
	 * This is used for not clearing a table twice.
	 */
	public HashSet<String> cleared = new HashSet<String>();
	
	/**
	 * Set of predicates that can potentially share with others (through
	 * dual decomposition).
	 */
	public HashSet<FelixPredicate> commonCandidates = new HashSet<FelixPredicate>();
	
	/**
	 * Predicates output by this bucket.
	 */
	public HashSet<FelixPredicate> outputPredicates = new HashSet<FelixPredicate>();
	
	/**
	 * If this bucket is a tuffy bucket, this set contains the result table 
	 * name of their corresponding MLN.
	 */
	HashSet<String> subTuffyMLNRelTable = new HashSet<String>();
	
	/**
	 * Register a table in {@link ConcurrentOperatorsBucket#subTuffyMLNRelTable}.
	 * @param tableName
	 */
	public void addMLNRelTable(String tableName){
		this.subTuffyMLNRelTable.add(tableName);
	}
	
	/**
	 * Predicates as input.
	 */
	public HashSet<String> inputPredicateScope = new HashSet<String>();
	
	/**
	 * Adds input predicate scopes.
	 * @param _predicates
	 * @deprecated
	 */
	public void pushPredicateScopes(HashSet<FelixPredicate> _predicates){
		for(FelixPredicate fp : _predicates){
			this.inputPredicateScope.add(fp.getName());
		}
	}
	
	/**
	 * Degree of concurrent running of operators.
	 * @deprecated
	 */
	int nCore = 1;
		
	/**
	 * MAP/marginal.
	 */
	boolean isMarginal = false;
	
	/**
	 * Whether this bucket has ran its first operator. This variable is used to
	 * control some consistency issues.
	 * @deprecated
	 */
	boolean started = false;
	
	/**
	 * ID of this bucket.
	 */
	public int id = -1;
	
	/**
	 * Type of operators in this bucket. The construction of buckets
	 * ensures all operators in it are with the same type.
	 */
	public OPType type;
	
	/**
	 * Precedence of this bucket. The construction of buckets
	 * ensures all operators in it are with the same precedence.
	 */
	public int precedence = -1;
	
	/**
	 * Command line options.
	 */
	FelixCommandOptions options;
	
	
	/**
	 * Whether this bucket runs in marginal mode.
	 * @return
	 */
	public boolean isMarginal(){
		return this.isMarginal;
	}
	
	/**
	 * The constructor.
	 * @param isMarginal
	 */
	public ConcurrentOperatorsBucket(boolean isMarginal){
		this.isMarginal = isMarginal;
	}
	
	/**
	 * Get all operators in this bucket.
	 * @return
	 */
	public HashSet<StatOperator> getOperators(){
		return concurrentOperators;
	}
	
	/**
	 * Get the precedence of this bucket.
	 * @return
	 */
	public int getPrecedence(){
		return precedence;
	}
	
	/**
	 * Number of rules that only contains evidence relations.
	 * Intuitively, the larger this number, the earlier
	 * we should solve it while scheduling.
	 */
	public int nStartingRule;
	
	/**
	 * Add an operator into this bucket.
	 * @param sop
	 */
	public void addOperator(StatOperator sop){
		this.concurrentOperators.add(sop);
		outputPredicates.addAll(sop.outputPredicates);
		inputPredicates.addAll(sop.inputPredicates);
		
		for(FelixPredicate fp : sop.outputPredicates){
			fp.belongsTo = this;
		}
		
		this.commonCandidates.addAll(sop.commonCandidate);
		
		this.nStartingRule = sop.nStartingRules;
		options = sop.options;
		this.type = sop.type;
		this.precedence = sop.getPrecedence();
		sop.belongsToBucket = this;
	}
	
	/**
	 * Set the degree of concurrency.
	 * @param _nCore
	 */
	public void setNCore(int _nCore){
		nCore = _nCore;
	}
	
	/**
	 * Relations that are shared with other buckets through 
	 * dual decomposition.
	 */
	public HashSet<FelixPredicate> dd_CommonOutput = new HashSet<FelixPredicate>();
	
	/**
	 * Map from relations in {@link #dd_CommonOutput}
	 * to tables containing their temporary output result in
	 * each iteration.
	 */
	public HashMap<FelixPredicate, String> dd_commonOutputPredicate_2_tableName = 
											new HashMap<FelixPredicate, String>();
	
	/**
	 * Map from relations in {@link #dd_CommonOutput}
	 * to tables containing their corresponding Langragian
	 * Multipliers.
	 */
	public HashMap<FelixPredicate, String> dd_commonOutputPredicate_2_priorTableName = 
		new HashMap<FelixPredicate, String>();

	/**
	 * Adds predicate to {@link #dd_CommonOutput}.
	 * @param _fp
	 */
	public void addCommonOutput(FelixPredicate _fp){
		for(StatOperator sop : this.concurrentOperators){
			sop.dd_CommonOutput.add(_fp);
		}
		this.dd_CommonOutput.add(_fp);
	}
	
	/**
	 * Prepares database for dual decomposition -- 1) add
	 * new rules for Lagaragian Multipliers; 2) create tables
	 * for intermediate output and priors etc..
	 */
	public void prepareDB44DD(){
		if(options.useDualDecomposition){
			
			RDB newDB = RDB.getRDBbyConfig(Config.db_schema);
			
			for(FelixPredicate fp : this.dd_CommonOutput){
				
				// generate output table and build the map from predicate name to table name
				if(fp.isCorefPredicate){
					String tableName = "_dd_" + fp.getName() + "_map_of_op" +this.id + "_" + this.type;
					String viewName = "_dd_" + fp.getName() + "_of_op" +this.id + "_" + this.type;
					
					newDB.dropTable(tableName);
					newDB.dropView(viewName);
					newDB.execute("CREATE TABLE " + tableName + " AS (SELECT * FROM " + fp.corefMAPPredicate.getRelName() + " WHERE 1=2)");
					
					this.dd_commonOutputPredicate_2_tableName.put(fp, viewName);
					this.dd_commonOutputPredicate_2_tableName.put(fp.corefMAPPredicate, tableName);
										
				}else{
					String tableName = "_dd_" + fp.getName() + "_of_op" +this.id + "_" + this.type;
					this.dd_commonOutputPredicate_2_tableName.put(fp, tableName);
					
					newDB.dropTable(tableName);
					newDB.execute("CREATE TABLE " + tableName + " AS (SELECT * FROM " + fp.getRelName() + " WHERE 1=2)");
				}
	
				// generate prior predicate and new rules
				String priorTableName = "_dd_prior_" + fp.getName() + "_of_op" + this.id + "_" + this.type;
				FelixPredicate tmpPredicate = new FelixPredicate(priorTableName, true);
				for(int i=0;i<fp.getArgs().size();i++){
					tmpPredicate.appendArgument(fp.getTypeAt(i));
				}
				tmpPredicate.appendArgument(new Type("float_"));
				tmpPredicate.prepareDB(Felix.db);
				dd_commonOutputPredicate_2_priorTableName.put(fp, tmpPredicate.getRelName());
				
				Literal tmpLiteral = new Literal(tmpPredicate, false);
				for(int i=0;i<fp.getArgs().size();i++){
					tmpLiteral.appendTerm(new Term(String.valueOf((char)('a' + i))));
				}
				tmpLiteral.appendTerm(new Term("wgt"));
				
				Literal tmpLiteral2 = new Literal(fp, true);
				for(int i=0;i<fp.getArgs().size();i++){
					tmpLiteral2.appendTerm(new Term(String.valueOf((char)('a' + i))));
				}
				
				FelixClause fc = new FelixClause();
				fc.addLiteral(tmpLiteral);
				fc.addLiteral(tmpLiteral2);
				fc.setVarWeight("wgt");
				
				
				
				for(StatOperator sop : this.concurrentOperators){
					
					FelixClause fff = (FelixClause) fc.clone();
					
					if(sop.dc == null){
						sop.dd_PriorClauses.add(fff);
					}else{
						HashSet<Expression> e = 
								sop.dc.getExpressions(fff, null, sop.dPart, sop.nPart, false);
						sop.clauseConstraints.put(fff, e);
						sop.dd_PriorClauses.add(fff);
					}
					
					
				}
				
				
				if(fp.isCorefPredicate && this.type == OPType.COREF){
					
					FelixPredicate pmap = fp.corefMAPPredicate;
					
					priorTableName = "_dd_prior_" + pmap.getName() + "_of_op" + this.id + "_" + this.type;
					tmpPredicate = new FelixPredicate(priorTableName, true);
					for(int i=0;i<fp.getArgs().size();i++){
						tmpPredicate.appendArgument(fp.getTypeAt(i));
					}
					tmpPredicate.appendArgument(new Type("float_"));
					tmpPredicate.prepareDB(Felix.db);
					dd_commonOutputPredicate_2_priorTableName.put(pmap, tmpPredicate.getRelName());
					
					tmpLiteral = new Literal(tmpPredicate, false);
					for(int i=0;i<fp.getArgs().size();i++){
						tmpLiteral.appendTerm(new Term(String.valueOf((char)('a' + i))));
					}
					tmpLiteral.appendTerm(new Term("wgt"));
					
					tmpLiteral2 = new Literal(fp, true);
					for(int i=0;i<fp.getArgs().size();i++){
						tmpLiteral2.appendTerm(new Term(String.valueOf((char)('a' + i))));
					}
					
					fc = new FelixClause();
					fc.addLiteral(tmpLiteral);
					fc.addLiteral(tmpLiteral2);
					fc.setVarWeight("wgt");
					
					for(StatOperator sop : this.concurrentOperators){
						sop.dd_PriorClauses.add(fc);
					}
				}
				
				
			}
			
			
			newDB.close();
			
			for(StatOperator sop : this.concurrentOperators){
				sop.dd_commonOutputPredicate_2_tableName = this.dd_commonOutputPredicate_2_tableName;
			}
			
		}
	}
	
	/**
	 * Run all operators in this bucket.
	 */
	public void run(){
				
		cleared.clear();
		subTuffyMLNRelTable.clear();
		
		for(Predicate p : outputPredicates){
			RDB newDB = RDB.getRDBbyConfig(Config.db_schema);
			
			if(p.isCurrentlyView == false && p.isClosedWorld() == false){
				if(FelixConfig.isFirstRunOfDD || !this.options.useDualDecomposition){
					String sql = "DELETE FROM " + p.getRelName();
					newDB.execute(sql);
				}
			}
			if(this.type.equals(OPType.COREF)){
				if(!options.isDLearningMode){
					if(FelixConfig.isFirstRunOfDD || !this.options.useDualDecomposition){
						String sql = "DELETE FROM " + p.getRelName() + "_map";
						newDB.execute(sql);
					}
				}
			}
			newDB.close();
			//p.nextTupleID = 0;
		}
		
		operatorStacks.clear();
		operatorStacks.addAll(concurrentOperators);
		
		if(this.type.equals(OPType.TUFFY)){
			nCore = 1;
		}
		

		
		if(options.useDualDecomposition){
			
			RDB newDB = RDB.getRDBbyConfig(Config.db_schema);
			for(FelixPredicate fp : this.dd_CommonOutput){
				String tableName = this.dd_commonOutputPredicate_2_tableName.get(fp);
				String sql = "DELETE from " + tableName;
				newDB.execute(sql);
			}
			newDB.close();
			
			if(this.type == OPType.TUFFY){
				
				if(FelixConfig.isFirstRunOfDD){
					for(StatOperator sop : concurrentOperators){
						sop.currentState = false;
						sop.run();
					}
					this.cleanUp();
				}else{
									
					// second, run tuffy on the output table,
					// take prior as evidences
					for(StatOperator sop : concurrentOperators){
						sop.currentState = false;
						sop.run();
					}
					this.cleanUp();
					
				}
				
			}else{
			
				ExecutorService pool = Executors.newFixedThreadPool(Config.getNumThreads());
				
				TaskList tasks = new TaskList();
				
				TaskSet taskset1 = new TaskSet();
					
				for(StatOperator sop : concurrentOperators){
					taskset1.addSubTask(new ExecuteOperatorTask(sop));
				}
				
				tasks.addSubTask(taskset1);
				try {
					tasks.execute(pool);
				} catch (Exception e) {
					e.printStackTrace();
				}
				
				pool.shutdown();
				
				this.cleanUp();
			}
			
			newDB.close();
			
		}else{
			for(int i=0;i<nCore; i++){
				if(operatorStacks.size() == 0){
					continue;
				}
				StatOperator sop = operatorStacks.remove(operatorStacks.size() - 1);
				sop.start();
				started = true;
			}
		}
	}
	
	/**
	 * Merge the output of operators into a consistent DB table 
	 * and dump them into files.
	 */
	public void cleanUp(){
		try{
			RDB newDB = RDB.getRDBbyConfig();
			newDB.execute("SET search_path = " + Config.db_schema);
			
			if(this.type.equals(OPType.TUFFY)){
				MarkovLogicNetwork tmpMLN = new MarkovLogicNetwork();
				tmpMLN.setDB(newDB);
				DataMover datamover = new DataMover(tmpMLN);
				
				for(FelixPredicate p : outputPredicates){
					
					if(p.isClosedWorld() == true){
						continue;
					}
					
					datamover.updateOriTable(subTuffyMLNRelTable, p);
					datamover.updateOriTable(subTuffyMLNRelTable, p, "_copy_of_" + p.getRelName());
				}
				
				for(FelixPredicate p : this.dd_CommonOutput){
					
					
					String tableName = this.dd_commonOutputPredicate_2_tableName.get(p);
						
					
					newDB.execute("DELETE FROM " + this.dd_commonOutputPredicate_2_tableName.get(p));
						
					datamover.updateOriTable(subTuffyMLNRelTable, p, 
							this.dd_commonOutputPredicate_2_tableName.get(p));	

					
				}
				
			}
			
			newDB.close();
		}catch(Exception e){
			e.printStackTrace();
		}
	}
	
	/**
	 * Run the next operator in this bucket.
	 * @deprecated
	 */
	public synchronized void runNextOperatorInBucket(){
		
		if(operatorStacks.size() == 0){
			return;
		}
		
		StatOperator sop = operatorStacks.remove(operatorStacks.size() - 1);
		sop.start();
	}
	
	/**
	 * Return only if all the operators have been executed.
	 * @deprecated
	 */
	public void myJoin(){

		try{
			int goal = this.concurrentOperators.size();
			
			while(true){
				int ct = 0;
				for(StatOperator sop : this.concurrentOperators){
					if(operatorStacks.contains(sop) || !started){
						continue;
					}
					sop.join();
					ct ++;
				}
				if(ct == goal){
					break;
				}
			}
			
			this.cleanUp();
		}catch(Exception e){
			e.printStackTrace();
		}
	
	}
	
	/**
	 * Returns string representation of bucket.
	 */
	public String toString(){

		String ret = "";
		
		for(StatOperator sop : this.concurrentOperators){
			ret += sop;
			ret += "\n";
		}
		
		return ret;
	}
	
	/**
	 * Returns string representation of bucket without partitioning info..
	 * @return
	 */
	public String toNoParString(){

		String ret = "";
		
		for(StatOperator sop : this.concurrentOperators){
			ret += sop.toNoParString();
			ret += "\n";
			break;
		}
		
		return ret;
	}
	
	/**
	 * Output the results of this bucket.
	 * @param db
	 * @param fout
	 * @param p
	 * @deprecated
	 */
	public void dumpMapAnswerForPredicate(RDB db, String fout, FelixPredicate p) {
	//	spreadTruth();
		
		int digits = 4;
		
		HashMap<Long,String> cmap = db.loadIdSymbolMapFromTable();
		try {
			BufferedWriter bufferedWriter = FileMan.getBufferedWriterMaybeGZ(fout);
			String sql;
			String tableName = p.getRelName();
			String predName = p.getName();
			
			if(this.type.equals(OPType.COREF)){
				tableName += "_map";
				predName += "_map";
			}
			
			sql = "SELECT * FROM " + tableName + " WHERE truth=TRUE";
			
			
			ResultSet rs = db.query(sql);
			while(rs == null){
				rs = db.query(sql);
			}
			while(rs.next()) {
				String line = "";
				ArrayList<String> cs = new ArrayList<String>();
				for(String a : p.getArgs()) {
					int c = rs.getInt(a);
					cs.add("\"" + StringMan.escapeJavaString(cmap.get(c)) + "\"");
				}
				line += StringMan.commaList(cs) + ")";
				
				if(isMarginal){
					
					if(Config.output_prolog_format){
						line = "tuffyPrediction(" + UIMan.decimalRound(digits, rs.getDouble("prior")) +
							", " + predName + "(" + line + ".";
					}else{
						if(!this.type.equals(OPType.COREF)){
							line = rs.getString("prior") + "\t" + predName + "(" + line;
						}
					}
					
				}else{
					line =  predName + "(" + line;
				}
				
				bufferedWriter.append(line + "\n");
			}
			rs.close();
			bufferedWriter.close();
		} catch (Exception e) {
			ExceptionMan.handle(e);
		}
	}
		
	
}
