package felix.dstruct;


import java.sql.Array;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;


import felix.util.FelixStringMan;
import felix.util.FelixUIMan;

import tuffy.db.RDB;
import tuffy.mln.Literal;
import tuffy.ra.ConjunctiveQuery;
import tuffy.util.ExceptionMan;
import tuffy.util.UIMan;


/**
 * A Data Movement Operator (DMO) is a bridge from physical database
 * objects (e.g., view, relational table etc.) to logical
 * requirements of the data. A statistical operator (e.g., CRF, Coref etc.)
 * only knows the logic of DMO instead of the physical execution
 * plan. The physical execution plan of the DMO is instantiated by 
 * the Optimizer. The logic definition of the DMO is instantiated by
 * corresponding statistical operators.
 * 
 * @author Ce Zhang
 *
 */
public class DataMovementOperator {

	/**
	 * RDB object on which this DMO works on.
	 */
	public RDB db;
	
	/**
	 * Estimated number of Bound-Bound invocations.
	 */
	public int predictedBB = 0;
	
	/**
	 * Estimated number of Bound-Free invocations.
	 */
	public int PredictedBF = 0;
	
	/**
	 * Estimated number of Free-Free invocations.
	 */
	public int PredictedFF = 0;
	
	/**
	 * Whether this DMO can be used as view/table by other DMOs.
	 */
	public boolean isIntermediaDMO = false;
	
	/**
	 * Name of table/view generated by this DMO. This name
	 * can be used by other DMOs as a view/table directly.
	 */
	public String name = "";
	
	/**
	 * Counter for DMO ID.
	 */
	public static int globalID = 1;
	
	/**
	 * Whether this DMO allowing optimizations from the optimizer.
	 */
	public boolean allowOptimization = true;
	
	/**
	 * Whether this DMO allows pre-fetching.
	 */
	public boolean hasKnownFetchingOrder = false;
	
	/**
	 * Parameter for pre-fetching optimization.
	 */
	public int preFetchFactor = -1;
	
	/**
	 * The statistical operator who instantiates this DMO.
	 */
	public StatOperator parentStatOperator = null;
	
	/**
	 * Conjunctive query or prepared statement acting as the *logic*
	 * description of this DMO. This plan cannot be executed by
	 * the executor.
	 */
	public QuerySet logicQueryPlan = null;
	
	/**
	 * How the result of this DMO is ordered?
	 */
	public ArrayList<String> orderBy = new ArrayList<String>();
	
	/**
	 * Do we need to build index for this DMO's output/intermediate
	 * results?
	 */
	public boolean needIndex = false;
	
	/**
	 * Datalog, Conjunctive query or prepared statement acting as the *physical*
	 * description of this DMO. When this DMO is executed, it first
	 * evaluates all Datalog rules (only evaluate once), and then evaluate
	 * the conjunctive query or prepared statement (which may have been optimized).
	 * This variable cannot be touched by any ad hoc statistical operator.
	 */
	public QuerySet physicalQueryPlan = null;
	
	/**
	 * The schema for the output of this DMO.
	 */
	public ArrayList<String> finalSelList = new ArrayList<String>();
	
	/**
	 * Part of the schema of the output of this DMO, which is from the
	 * conjunctive queries or prepared statement in {@link DataMovementOperator#logicQueryPlan}.
	 */
	public ArrayList<String> selListFromRule = new ArrayList<String>();
	
	/**
	 * Part of the schema of the output of this DMO, which is additionally
	 * assigned (i.e., may not be founded in the conjunctive query or
	 * prepared statement in {@link DataMovementOperator#logicQueryPlan}).
	 */
	public ArrayList<String> additionalSelList = new ArrayList<String>();
	
	/**
	 * Whether prepare() method has been invoked.
	 */
	public boolean hasPrepared = false;	
	
	/**
	 * Whether the output is a {@link ResultSet} object or others. This is maintained
	 * by DMO, and is not visible to others. 
	 * 
	 */
	boolean isResultSet = true;
	
	/**
	 * Whether the output (if any) of this DMO should be regarded as view or materialized
	 * table.
	 */
	public boolean asView = true;
	
	/**
	 * Set of variables which will be bounded while evaluating this DMO.
	 */
	public HashSet<String> whichToBound = new HashSet<String>();
	
	/**
	 * Object as the output of this DMO. This can be a {@link ResultSet} object or other.
	 * This variable is not visible to other class. Other class must use specified data-exchange
	 * interface of DMO instead of trying to fetching data from this variable directly.
	 */
	Object resultSet = null;
	
	/**
	 * Get the union of results from a set of DMOs. Here each DMO must share exactly the same output schema.
	 * @param db
	 * @param _parentStatOperator
	 * @param toBeUnioned set of DMOs to be unioned.
	 * @param bindingPattern bindingPattern of the resulting union. See {@link ConjunctiveQuery#psMap}.
	 * @param bindings 
	 * @return
	 */
	@SuppressWarnings("unchecked")
	public static DataMovementOperator UnionAll(RDB db, StatOperator _parentStatOperator, 
			List<DataMovementOperator> toBeUnioned, String bindingPattern, ArrayList<Integer> bindings, int... cid){
		
		//return a dummy dmo for unionAll
		DataMovementOperator ret = new DataMovementOperator(db, _parentStatOperator);
		ret.allowOptimization = false;
		ret.predictedBB = 0;
		ret.PredictedBF = 0;
		ret.PredictedFF = 0;
		
		String sql = "";
		ArrayList<String> subqueries = new ArrayList<String>();
		ArrayList<String> unionAllSelList = null;
		
		for(DataMovementOperator dmo : toBeUnioned){
			subqueries.add(" SELECT * FROM " + dmo.getAllFreeViewName());
			// we assume all sub-DMOs should have the same selList.
			if(unionAllSelList == null){
				ret.selListFromRule = (ArrayList<String>) dmo.selListFromRule.clone();
				ret.additionalSelList = (ArrayList<String>) dmo.additionalSelList.clone();
				ret.finalSelList = (ArrayList<String>) dmo.finalSelList.clone();
				unionAllSelList =  (ArrayList<String>) dmo.finalSelList.clone();
			}
		}
				
		sql = "SELECT " + FelixStringMan.commaList(unionAllSelList) + " FROM ( " + 
			FelixStringMan.join(" UNION ALL ", subqueries) + ") TMPTMPTMP WHERE ";
		
		ArrayList<String> whereList = new ArrayList<String>();
		for(int i=0; i < unionAllSelList.size(); i++){
			if(bindingPattern.charAt(i) == '1'){
				whereList.add("(" + unionAllSelList.get(i) + " = ?" + ")");
			}
		}
		whereList.add("(1=1)");
		sql += FelixStringMan.join(" AND ", whereList);
				
		ret.logicQueryPlan.addQuery(db.getPrepareStatement(sql), ret.selListFromRule, ret.additionalSelList);
		
		
		return ret;
	}
	
	/**
	 * Generate a DMO which is a SELECT on the given DMO.
	 * @param db
	 * @param _parentStatOperator
	 * @param fromDMO the given DMO.
	 * @param bindings variables in the WHERE list.
	 * @return
	 */
	@SuppressWarnings("unchecked")
	public static DataMovementOperator Select(RDB db, StatOperator _parentStatOperator, 
			DataMovementOperator fromDMO, ArrayList<String> bindings, int... cid){
			
		//return a dummy dmo for unionAll
		DataMovementOperator ret = new DataMovementOperator(db, _parentStatOperator);
		ret.allowOptimization = false;
		ret.predictedBB = 0;
		ret.PredictedBF = 0;
		ret.PredictedFF = 0;
		
		ret.selListFromRule = (ArrayList<String>) fromDMO.selListFromRule.clone();
		ret.additionalSelList = (ArrayList<String>) fromDMO.additionalSelList.clone();
		ret.finalSelList = (ArrayList<String>) fromDMO.finalSelList.clone();
		
		String sql = "SELECT " + FelixStringMan.commaList(ret.finalSelList) + " FROM " + 
			fromDMO.getAllFreeViewName() + " WHERE ";
		
		ArrayList<String> whereList = new ArrayList<String>();
		for(int i=0; i < bindings.size(); i++){
			whereList.add("(" + bindings.get(i) + " = ?" + ")");
		}
		whereList.add("(1=1)");
		sql += FelixStringMan.join(" AND ", whereList);
				
		ret.logicQueryPlan.addQuery(db.getPrepareStatement(sql), ret.selListFromRule, ret.additionalSelList);

		return ret;
	}
	
	/**
	 * Generate a DMO which is a SELECT ... ORDER BY on the given DMO.
	 * @param db
	 * @param _parentStatOperator
	 * @param fromDMO the given DMO.
	 * @param bindings variables in the WHERE list.
	 * @param orderby
	 * @param cid
	 * @return
	 */
	@SuppressWarnings("unchecked")
	public static DataMovementOperator SelectOrderBy(RDB db, StatOperator _parentStatOperator, 
			DataMovementOperator fromDMO, ArrayList<String> bindings, String orderby, int... cid){
			
		//return a dummy dmo for unionAll
		DataMovementOperator ret = new DataMovementOperator(db, _parentStatOperator);
		ret.allowOptimization = false;
		ret.predictedBB = 0;
		ret.PredictedBF = 0;
		ret.PredictedFF = 0;
		
		ret.selListFromRule = (ArrayList<String>) fromDMO.selListFromRule.clone();
		ret.additionalSelList = (ArrayList<String>) fromDMO.additionalSelList.clone();
		ret.finalSelList = (ArrayList<String>) fromDMO.finalSelList.clone();
		
		String sql = "SELECT " + FelixStringMan.commaList(ret.finalSelList) + " FROM " + 
			fromDMO.getAllFreeViewName() + " WHERE ";
		
		ArrayList<String> whereList = new ArrayList<String>();
		for(int i=0; i < bindings.size(); i++){
			whereList.add("(" + bindings.get(i) + " = ?" + ")");
		}
		whereList.add("(1=1)");
		sql += FelixStringMan.join(" AND ", whereList);
		sql += orderby;
				
		ret.logicQueryPlan.addQuery(db.getPrepareStatement(sql), ret.selListFromRule, ret.additionalSelList);

		return ret;
	}
	
	/**
	 * the constructor.
	 * @param _db
	 * @param _parentStatOperator
	 */
	public DataMovementOperator(RDB _db, StatOperator _parentStatOperator){
		
		db = _db;
		parentStatOperator = _parentStatOperator;
		logicQueryPlan = new QuerySet();
		physicalQueryPlan = new QuerySet();
		
		this.name = "pred_DMO_" + (globalID++);
	}
	
	/**
	 * Methods that prepares for the physical execution of this DMO. Invocation
	 * of this method will: 1) execute all Datalog rules in the physical
	 * plan; 2) build indices for evaluating conjunctive queries; and 3) create
	 * view/table for this DMO. This function can be only called once in
	 * each Dual decomposition iteration.
	 * 
	 */
	public void prepare(){
		
		
		if(hasPrepared){
			for(ConjunctiveQuery cq : physicalQueryPlan.datalogQueries){
				//db.dropTable(cq.head.getPred().getRelName());
				String ss = "DELETE FROM " + cq.head.getPred().getRelName();
				db.execute(ss);
			}
			
			db.commit();
		}
		
//		if(hasPrepared == false){
			for(ConjunctiveQuery cq : physicalQueryPlan.datalogQueries){
				FelixUIMan.println(2, "Materailizing " + cq);
				cq.buildIndexes(db, null, null, null, false, new ArrayList<String>());
				cq.materialize(db, true, orderBy);
			}
			
			// TODO: do we need to build indices for PreparedStatement?
			if(physicalQueryPlan.objectConjunctiveQuery != null){
				physicalQueryPlan.objectConjunctiveQuery.buildIndexes(db, null, null, null, false, 
						this.additionalSelList);
	
				db.analyze(physicalQueryPlan.objectConjunctiveQuery.head.getPred().getRelName());
				
				for(Literal l : physicalQueryPlan.objectConjunctiveQuery.body){
					db.analyze(l.getPred().getRelName());
				}
			}
			
			// create view/table for this DMO
			if(!isIntermediaDMO && asView){
				
				db.dropTable(this.getAllFreeViewName());
				db.dropView(this.getAllFreeViewName());
				db.commit();
				
				String sql = "CREATE VIEW " + this.getAllFreeViewName() + " AS " + 
					this.getPS(null, new ArrayList<Integer>()).toString();
				
				FelixUIMan.println(2, "CREATE VIEW " + this.getAllFreeViewName() + " AS " + 
						this.getPS(null, new ArrayList<Integer>()).toString());
				
				db.execute(sql);
			}else if(!isIntermediaDMO && !asView){
				
				db.dropTable(this.getAllFreeViewName());
				db.dropView(this.getAllFreeViewName());
				db.commit();
				
				FelixUIMan.println(2, "CREATE TABLE "+ this.getAllFreeViewName() + " AS " + this.getPS(null, new ArrayList<Integer>()).toString());
				String sql = "CREATE TABLE " + this.getAllFreeViewName() + " AS " + 
				this.getPS(null, new ArrayList<Integer>()).toString();
				db.execute(sql);
				
				// needIndex should go back to the optimizer?
				needIndex = true;
				if(needIndex){
					for(String s : this.selListFromRule){
						FelixUIMan.println(2, "CREATE INDEX " + this.getAllFreeViewName() + "_" + s 
								+ " ON " + this.getAllFreeViewName() + "(" + s + ")");
						db.execute("CREATE INDEX  " + this.getAllFreeViewName() + "_" + s 
								+ " ON " + this.getAllFreeViewName() + "(" + s + ")");
					}
				}
				
				db.analyze(this.getAllFreeViewName());
			}
			
			hasPrepared = true;
//		}
	}
	
	/**
	 * Execute this DMO given bindings of variables. This method need to be invoked before
	 * {@link DataMovementOperator#next()}, {@link DataMovementOperator#getNext(int)}, 
	 * {@link DataMovementOperator#getNext(String)} etc. can be invoked. This interface
	 * is similar to ResultSet in JDBC.
	 * 
	 * @param bindingPattern binding patterns. If this DMO is a conjunctive query (in logic plan),
	 * use format as {@link ConjunctiveQuery#psMap} (a null value will correspond to all-free case). 
	 * If this DMO is a prepared statement (in logic plan), this value should be set as null.
	 * @param bindings ordered list of constants that will be used to bind variables.
	 * @return PreparedStatement that will be used to execute this DMO.
	 */
	public PreparedStatement execute(String bindingPattern, ArrayList<Integer> bindings){
		
		//TODO: add pre-fetch for CRF operator
		try{
			
			if(resultSet != null && resultSet.getClass().equals(ResultSet.class)){
				ResultSet rs = (ResultSet) resultSet;
				rs.close();
			}
			
			resultSet = null;
			
			//this.prepare();
			
			if(bindingPattern == null && physicalQueryPlan.objectPreparedStatement != null){
				
				for(int i=0;i<bindings.size();i++){
					physicalQueryPlan.objectPreparedStatement.setInt(i+1, bindings.get(i));
				}

				physicalQueryPlan.objectPreparedStatement.execute();
				
				resultSet = physicalQueryPlan.objectPreparedStatement.getResultSet();
				
				isResultSet = true;
				
				return physicalQueryPlan.objectPreparedStatement;
				
			}else if(bindingPattern != null && physicalQueryPlan.objectConjunctiveQuery != null){
				
				PreparedStatement ps = physicalQueryPlan.objectConjunctiveQuery.psMap.get(bindingPattern);
				
				for(int i=0;i<bindings.size();i++){
					physicalQueryPlan.objectPreparedStatement.setInt(i+1, bindings.get(i));
				}
				
				ps.execute();
				
				resultSet = ps.getResultSet();
				
				isResultSet = true;
				
				return ps;
				
			}else if(bindingPattern == null && physicalQueryPlan.objectConjunctiveQuery != null){
				
				PreparedStatement ps = physicalQueryPlan.objectConjunctiveQuery.psMap.get(
						physicalQueryPlan.objectConjunctiveQuery.getAllFreeBinding());
								
				ps.execute();
				
				resultSet = ps.getResultSet();
				
				isResultSet = true;
				
				return ps;
			}else{
				throw new Exception("Errors when trying to binding non-existing queries!");
			}
		}catch(Exception e){
			e.printStackTrace();
		}
		return null;
		
	}
	
	/**
	 * Get the prepared statement that will be used to execute this DMO. 
	 * See {@link DataMovementOperator#execute(String, ArrayList)} (the only difference
	 * is this method does not actually execute this DMO).
	 */
	public PreparedStatement getPS(String bindingPattern, ArrayList<Integer> bindings){
		
		//TODO: add pre-fetch for CRF operator
		try{
						
			if(bindingPattern == null && physicalQueryPlan.objectPreparedStatement != null){
				
				for(int i=0;i<bindings.size();i++){
					physicalQueryPlan.objectPreparedStatement.setInt(i+1, bindings.get(i));
				}
					
				return physicalQueryPlan.objectPreparedStatement;
				
			}else if(bindingPattern != null && physicalQueryPlan.objectConjunctiveQuery != null){
				
				PreparedStatement ps = physicalQueryPlan.objectConjunctiveQuery.psMap.get(bindingPattern);
				
				for(int i=0;i<bindings.size();i++){
					physicalQueryPlan.objectPreparedStatement.setInt(i+1, bindings.get(i));
				}
				
				return ps;
				
			}else if(bindingPattern == null && physicalQueryPlan.objectConjunctiveQuery != null){
				
				PreparedStatement ps = physicalQueryPlan.objectConjunctiveQuery.psMap.get(
						physicalQueryPlan.objectConjunctiveQuery.getAllFreeBinding());
								
				return ps;
			}else{
				throw new Exception("Errors when trying to binding non-existing queries!");
			}
		}catch(Exception e){
			
		}
		return null;
		
	}
	
	/**
	 * Gets the view/table name of this DMO in Free-free case. This name can be referred
	 * by other DMO directly w/o aware of the physical plan of this DMO.
	 */
	public String getAllFreeViewName(){
		try{
			if(this.isIntermediaDMO){
				throw new Exception("Requesting view name for intermedia relation!");
			}
			return this.name;
		}catch(Exception e){
			e.printStackTrace();
		}
		return null;
	}
	
	/**
	 * Whether there are more tuples in the result of this DMO. See {@link DataMovementOperator#execute(String, ArrayList)}.
	 * @return
	 */
	public Boolean next(){
		
		try{
			if(isResultSet){
				
				ResultSet rs = (ResultSet) resultSet;
								
				boolean aa = rs.next();
				
				if(aa){
					return aa;
				}else{
					rs.close();
					return aa;
				}	
			}
		
		}catch(Exception e){
			
		}
		
		return null;
	}
	
	/**
	 * Get int value in the result of this DMO. See {@link DataMovementOperator#execute(String, ArrayList)}.
	 * @return
	 */
	public Integer getNext(String columnName){
		
		try{
			if(isResultSet){
				
				ResultSet rs = (ResultSet) resultSet;
				return rs.getInt(columnName);
	
			}
			
			
		}catch(Exception e){
			
		}
		
		return null;
	}
	
	/**
	 * Get int value in the result of this DMO. See {@link DataMovementOperator#execute(String, ArrayList)}.
	 * @return
	 */
	public Array getNextArray(int column){
		
		try{
			if(isResultSet){
				
				ResultSet rs = (ResultSet) resultSet;
				return rs.getArray(column);
	
			}
			
			
		}catch(Exception e){
			e.printStackTrace();
		}
		
		return null;
	}
	
	/**
	 * Get int value in the result of this DMO. See {@link DataMovementOperator#execute(String, ArrayList)}.
	 * @return
	 */
	public Integer getNext(int column){
		
		try{
			if(isResultSet){
				
				ResultSet rs = (ResultSet) resultSet;
				return rs.getInt(column);
	
			}
			
			
		}catch(Exception e){
			e.printStackTrace();
		}
		
		return null;
	}
	
	/**
	 * Get double value in the result of this DMO. See {@link DataMovementOperator#execute(String, ArrayList)}.
	 * @return
	 */
	public Double getNextDouble(String columnName){
		
		try{
			if(isResultSet){
				
				ResultSet rs = (ResultSet) resultSet;
				return rs.getDouble(columnName);
	
			}
			
			
		}catch(Exception e){
			
		}
		
		return null;
	}
	
	/**
	 * Get double value in the result of this DMO. See {@link DataMovementOperator#execute(String, ArrayList)}.
	 * @return
	 */
	public Double getNextDouble(int column){
		
		try{
			if(isResultSet){
				
				ResultSet rs = (ResultSet) resultSet;
				return rs.getDouble(column);
	
			}
			
			
		}catch(Exception e){
			e.printStackTrace();
		}
		
		return null;
	}
	
	/**
	 * Get String value in the result of this DMO. See {@link DataMovementOperator#execute(String, ArrayList)}.
	 * @return
	 */
	public String getNextString(int column){
		
		try{
			if(isResultSet){
				
				ResultSet rs = (ResultSet) resultSet;
				return rs.getString(column);

	
			}
			
			
		}catch(Exception e){
			e.printStackTrace();
		}
		
		return null;
	}
	
	/**
	 * Cancels rule of this DMO.
	 */
	public void closePS(){
		if(this.physicalQueryPlan.objectPreparedStatement != null){
			try {
				UIMan.println("Cancel PS " + this.physicalQueryPlan.objectPreparedStatement);
				this.physicalQueryPlan.objectPreparedStatement.cancel();
			} catch (SQLException e) {
				e.printStackTrace();
				ExceptionMan.die("DMO - 708");
			}
		}
	}
	
	/**
	 * Copy logic plan to physical plan directly. If a DMO does not
	 * allow global optimization, its physical plan will be filled
	 * in this fashion. Notes: This function can not be called by ad hoc
	 * statistical operator, which should only be aware of the logic plan.
	 */
	public void copyLogicPlanToPhysicalPlanDirectly(){
		physicalQueryPlan.datalogQueries = new ArrayList<ConjunctiveQuery>();
		physicalQueryPlan.objectConjunctiveQuery = 
			logicQueryPlan.objectConjunctiveQuery;
		physicalQueryPlan.objectPreparedStatement = 
			logicQueryPlan.objectPreparedStatement;
	}
	
	
	/**
	 * A collection containing 1) a set of datalog rules; 2) a conjunctive query
	 * or prepared statement. 
	 * 
	 * @author Ce Zhang
	 *
	 */
	public class QuerySet{
		
		public ArrayList<ConjunctiveQuery> datalogQueries;
		
		public ConjunctiveQuery objectConjunctiveQuery = null;
		
		public PreparedStatement objectPreparedStatement = null;
		
		/**
		 * The constructor.
		 */
		public QuerySet(){
			datalogQueries = new ArrayList<ConjunctiveQuery>();
		}
		
		/**
		 * Add one query to this collection.
		 * @param query
		 * @param selList See {@link DataMovementOperator#selListFromRule}.
		 * @param _additionalSelList See {@link DataMovementOperator#additionalSelList}.
		 */
		@SuppressWarnings("unchecked")
		public void addQuery(Object query, ArrayList<String> selList, ArrayList<String> _additionalSelList){
			try{
				if(objectConjunctiveQuery != null || objectPreparedStatement != null){
					throw new Exception("You can have only one rule for each DMO.");
				}
				
				additionalSelList = (ArrayList<String>) _additionalSelList;
				
				if(query.getClass().equals(ConjunctiveQuery.class)){			
					objectConjunctiveQuery = (ConjunctiveQuery) query;
					if(selList != null){
						selListFromRule = (ArrayList<String>) selList.clone();
					}else{
						for(String arg : objectConjunctiveQuery.head.getPred().getArgs()){
							selListFromRule.add(arg);
						}
					}
				}else{
					if(selList != null){
						selListFromRule = (ArrayList<String>) selList.clone();
					}else{
						throw new Exception("You need to provide the sel-list (the field of resulting table" +
								"for this prepared statement)");
					}
					objectPreparedStatement = (PreparedStatement) query;
				}
				
				finalSelList.addAll(selListFromRule);
				finalSelList.addAll(additionalSelList);
			}catch(Exception e){
				e.printStackTrace();
			}
		}
	}
	
	
}










