package felix.dstruct;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;

import felix.optimizer.CostModel;



/**
 * The class of an execution plan, which is an 
 * ordered list of {@link ConcurrentOperatorsBucket} 
 * with parameters.
 * 
 * @author Ce Zhang
 *
 */
public class ExecutionPlan {

	/**
	 * Cost model used in this execution plan.
	 */
	CostModel cm;
	
	/**
	 * Set CostModel.
	 * @param _cm
	 */
	public void setCostModel(CostModel _cm){
		cm = _cm;
	}
	
	/**
	 * Get CostModel.
	 * @return
	 */
	public CostModel getCostModel(){
		return cm;
	}
	
	/**
	 * List of buckets which are organized in physical execution order.
	 */
	public ArrayList<ConcurrentOperatorsBucket> operators = new ArrayList<ConcurrentOperatorsBucket>();
	
	/**
	 * Set of predicates that are shared between different operators
	 * via dual decomposition.
	 */
	public HashSet<FelixPredicate> dd_CommonPredicates = new HashSet<FelixPredicate>();
	
	/**
	 * Map from predicates in {@link #dd_CommonPredicates} to
	 * their parents operators.
	 */
	public HashMap<FelixPredicate, HashSet<ConcurrentOperatorsBucket>> 
			dd_Predicate2OperatorBucket = new HashMap<FelixPredicate, HashSet<ConcurrentOperatorsBucket>>();
														
	
	/**
	 * Add a bucket in the end of the execution plan.
	 * @param sop
	 */
	public void addOperatorAfter(ConcurrentOperatorsBucket sop){
		operators.add(sop);
	}

	/**
	 * Add a bucket before the end of the execution plan. 
	 * @param sop
	 */
	public void addOperatorBefore(ConcurrentOperatorsBucket sop){
		operators.add(operators.size()-1, sop);
	}
	
	/**
	 * Get all statistical operators in this execution plan.
	 * @return
	 */
	public ArrayList<StatOperator> getAllStatOperator(){
		ArrayList<StatOperator> ret = new ArrayList<StatOperator>();
		
		for(ConcurrentOperatorsBucket cob : this.operators){
			for(StatOperator sop : cob.concurrentOperators){
				ret.add(sop);
			}
		}
		
		
		return ret;
	}
	
	/**
	 * Returns string representation of this execution plan.
	 */
	public String toString(){
		
		String ret = "";
		
		for(int i = operators.size()-1;i>=0; i--){
			ret += operators.get(i);
			ret += "\n";
		}
		
		return ret;
	}
	
	/**
	 * @deprecated
	 * Returns HTML representation of this execution plan 
	 * (for a very preliminary online demo).
	 * @return
	 */
	public String toHTMLString(){
		/*
		String ret = "";
		boolean flag = false;
		ret += "<UL CLASS=\"mktree\" ID=\"tree1\">\n";
		
		for(int i = operators.size()-1;i>=0; i--){
		
			ConcurrentOperatorsBucket cob = operators.get(i);
			
			ret += "<LI>" + cob.toNoParString() + " " + "with " + 
				cob.concurrentOperators.size() + " Concurrent Sub-Operators\n";
			ret += "<UL>\n";
				
			for(StatOperator sop : cob.getOperators()){
				ret += "<LI> Op" + sop.getId() + ": ";
				if(sop.dataCrackerSignature == null){
					ret += "Work on all data";
				}else{
					ret += "Partitioned by " + sop.dataCrackerSignature;
				}
				
				ret += "\n";
				
				ret += "<UL>";
				
				ret += "<LI> " + sop.inputPredicates.size() + " Input Open Predicates";
				ret += "<UL>";
				for(FelixPredicate fc : sop.inputPredicates){
					ret += "<LI>" + fc + "\n";
				}
				ret += "</UL>";
				
				ret += "<LI> " + sop.outputPredicates.size() + " Output Open Predicates";
				ret += "<UL>";
				for(FelixPredicate fc : sop.outputPredicates){
					ret += "<LI>" + fc + "\n";
				}
				ret += "</UL>";
				
				
				ret += "<LI> " + sop.allRelevantFelixClause.size() + " Relevant Clauses";
				ret += "<UL>";
				for(FelixClause fc : sop.allRelevantFelixClause){
					if(flag == false){
						flag = true;
						ret += "<LI ID=\"node234\">" + fc + "\n";
					}else{
						ret += "<LI ID=\"node234\">" + fc + "\n";
					}
				}
				ret += "</UL>";
				
				ret += "</UL>";
				
			}
				
			ret +="</UL>\n";
			
			ret += "<br/>";
			
		}
		
		ret += "</UL>\n";
		return ret;
		*/
		return null;
	}
	
}
