package felix.dstruct;

import java.io.FileInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;

import org.postgresql.PGConnection;

import felix.dstruct.StatOperator.OPType;
import felix.util.FelixConfig;
import felix.util.FelixUIMan;

import tuffy.db.RDB;
import tuffy.mln.Atom;
import tuffy.mln.Literal;
import tuffy.mln.Predicate;
import tuffy.util.ExceptionMan;
import tuffy.util.FileMan;
import tuffy.util.StringMan;
import tuffy.util.UIMan;


/**
 * The predicate object used in Felix, which extends
 * the Predicate class in Tuffy to contain Felix-related
 * fields and methods.
 * 
 * @author Ce Zhang
 *
 */
public class FelixPredicate extends Predicate{

	/**
	 * @deprecated
	 */
	String embeddedPythonCode = "";
	
	/**
	 * Properties that can be assigned to each predicate.
	 * Example properties include: SYMM, REPLEX, TRANS etc.
	 *
	 */
	public enum FPProperty {SYMM, REFLEX, TRANS, CHAIN_RECUR,
		OTHER_RECUR, OTHER_RECUR_WITHOTHER_OPENPRED, KEY_CONSTRAINT, NON_RECUR, EMBED_WEIGHT_RULE};
		
	/**
	 * Path the HDFS file that this predicate depends on.
	 */
	public String dependencyFile = null;
	
	/**
	 * Type of dependencies, e.g., "hdfs", "jdbc" etc.
	 */
	public String dependencyName = null;
	
	/**
	 * Python script for MAP.
	 */
	public String mapScript = null;
	
	/**
	 * Python script for running before all MAPs.
	 */
	public String mapinitScript = "";
	
	/**
	 * Python script for running before all REDUCEs.
	 */
	public String reduceinitScript = "";
	
	/**
	 * Python script for MAP.
	 */
	public String reduceScript = "\tfor v in _inputvalues:\n"+
								 "\t\tfelixio_push(_inputkey, v)";
	
	/**
	 * Do we need to extract features from HDFS for this relation?
	 */
	public boolean needExtractFeatures = false;
	
	/**
	 * If the input is XML tag, which &lt;xmltag&gt;&lt;/xmltag&gt;
	 * should we send to MAP as a unit?
	 */
	public String xmltag = null;
	
	/**
	 * Whether the extraction of this relation relies on other relations.
	 */
	public String jdbcdep = null;
	
	/**
	 * The name of MAP's input variable.
	 */
	public String mapinputvar = "_input";
	
	/**
	 * The name of REDUCE's input key variable.
	 */
	public String reduceinputkeyvar = "_inputkey";
	
	/**
	 * The name of REDUCE's input value variable.
	 */
	public String reduceinputvaluesvar = "_inputvalues";
	
	/**
	 * Whether the evidence file of this relation exists
	 * in some relational table instead of input evid. file.
	 */
	public boolean loadFromDatabase = false;
	
	/**
	 * see {@link #loadFromDatabase}
	 */
	public String loadingSchema = null;
	
	/**
	 * see {@link #loadFromDatabase}
	 */
	public String loadingTable = null;
	
	
	/**
	 * Adds atom as evidence to this predicate (Override Felix version) -
	 * the difference is that we sometimes flush it to file
	 * directly.
	 * @param a
	 */
	public void addEvidence(Atom a) {
		
		hasEvid = true;
		
		if (a.isSoftEvidence())
			setHasSoftEvidence(true);
		
		if(FelixConfig.mixturedLoading == true){
			
			ArrayList<String> towrite = new ArrayList<String>();
			
			try {
				
				this.loadingFileWriter.append((a.truth == true? "True" : "False"));
				this.loadingFileWriter.append("\t");
				this.loadingFileWriter.append(a.prior == null? "1" : a.prior.toString());;
				this.loadingFileWriter.append("\t");
			
				this.loadingFileWriter.append(StringMan.joinAndEscape("\t", a.sargs));
				this.loadingFileWriter.append("\n");
			} catch (IOException e) {
				e.printStackTrace();
			}
			
		}else{
			addEvidenceTuple(a);
		}
	}
		
	/**
	 * The type of operator this predicate must be assigned to.
	 */
	public OPType mustbe = null;
	
	/**
	 * The parent {@link ConcurrentOperatorsBucket}.
	 */
	public ConcurrentOperatorsBucket belongsTo = null;
		
	/**
	 * Whether this predicate is the view-based representation of
	 * a coref operator. This predicate must with a name suffix ``_map''.
	 */
	public boolean isCorefMapPredicate = false;
	
	/**
	 * Whether this predicate is a coref operator.
	 */
	public boolean isCorefPredicate = false;
	
	/**
	 * If {@link FelixPredicate#isCorefMapPredicate} is true, to which this
	 * predicate serves. 
	 */
	public FelixPredicate oriCorefPredicate = null;
	
	/**
	 * If {@link FelixPredicate#isCorefPredicate} is true, which
	 * relation serves as the linear-view-representation of it?
	 */
	public FelixPredicate corefMAPPredicate = null;
	
	/**
	 * Map from predicate properties ({@link FPProperty}) to
	 * clauses satisfying the corresponding property.
	 */
	HashMap<FPProperty, HashSet<FelixClause>> properities
		= new HashMap<FPProperty, HashSet<FelixClause>>();
	
	/**
	 * Clauses related to this predicate.
	 */
	HashSet<FelixClause> registeredClauses = new HashSet<FelixClause>();
	
	/**
	 * If this predicate has key constraints, this set records the position
	 * of keys.
	 */
	HashSet<Integer> labelPositions = new HashSet<Integer>();
	
	/**
	 * If this relation is defined as a view instead of
	 * a table, what is its view definition?
	 */
	public String viewDef = null; 
	
	/**
	 * Map from chain-rule clauses to possible partitions of sequence.
	 */
	HashMap<FelixClause, ArrayList<String>> chainRulePartitions =
		new HashMap<FelixClause, ArrayList<String>>();
	
	/**
	 * Get the label position if this predicate is LR or CRF.
	 * @return
	 */
	public ArrayList<Integer> getLabelPositions(){		
		ArrayList<Integer> ret = new ArrayList<Integer>();
		for(int i=0;i<this.arity();i++){
			if(!this.labelPositions.contains(i)){
				continue;
			}
			ret.add(i);
		}
		return ret;
	}
	
	/**
	 * Sets embedded python code for this predicate
	 * @param content
	 */
	public void setEmbeddedPythonCode(String content){
		
		this.embeddedPythonCode = content;
	
	}
	
	/**
	 * Get partitioning fields of sequence if this predicate is CRF.
	 * @return
	 */
	public ArrayList<String> getCRFPartitionFields(){

		if(chainRulePartitions.keySet().size() != 1){
			return null;
		}
		
		FelixClause key = chainRulePartitions.keySet().iterator().next();
		
		if(chainRulePartitions.get(key).size() == 0){
			return null;
		}
		
		return chainRulePartitions.get(key);
	
	}
	
	/**
	 * Get the key position if this predicate is LR or CRF.
	 * @return
	 */
	public ArrayList<String> getKeyFieldsArgs(){
		ArrayList<String> ret = new ArrayList<String>();
		for(int i=0;i<this.arity();i++){
			if(!labelPositions.contains(i)){
				ret.add(this.getArgs().get(i));
			}
		}
		return ret;
	}
	
	/**
	 * Get the label types if this predicate is LR or CRF.
	 * @return
	 */
	public ArrayList<String> getLabelFieldsTypeTable(){
		ArrayList<String> ret = new ArrayList<String>();
		for(int i=0;i<this.arity();i++){
			if(!labelPositions.contains(i)){
				continue;
			}
			ret.add(this.getTypeAt(i).getRelName());
		}
		return ret;
	}
	
	/**
	 * Get the label fields' name if this predicate is LR or CRF.
	 * @return
	 */
	public ArrayList<String> getLabelFieldsArgs(){
		ArrayList<String> ret = new ArrayList<String>();
		for(int i=0;i<this.arity();i++){
			if(!labelPositions.contains(i)){
				continue;
			}
			ret.add(this.getArgs().get(i));
		}
		return ret;
	}
	
	/**
	 * Global counter for temporary predicates.
	 */
	static int tmpPredCounter = 0;
	
	/**
	 * Get the name of the next temporary predicate.
	 * @return
	 */
	public static String getNextTmpPredicateName(){
		return "tmp_predicate_" + (tmpPredCounter++);
	}
	
	/**
	 * Whether this predicate is the view-based representation of
	 * a coref operator.
	 * @return
	 */
	public boolean isCorefMap(){
		return this.isCorefMapPredicate;
	}
	
	/**
	 * If {@link FelixPredicate#isCorefMapPredicate} is true, to which this
	 * predicate serves. 
	 * @return
	 */
	public FelixPredicate getOriCorefPredicate(){
		return this.oriCorefPredicate;
	}
	
	/**
	 * Get clauses associated to the given property.
	 * @param prop
	 * @return
	 */
	public HashSet<FelixClause> getPropertyClauses(FPProperty prop){
		if(this.hasProperty(prop)){
			return new HashSet<FelixClause>(properities.get(prop));
		}else{
			return new HashSet<FelixClause>();
		}
	}
	
	/**
	 * Whether this predicate has the given property.
	 * @param prop
	 * @return
	 */
	public boolean hasProperty(FPProperty prop){
		return properities.containsKey(prop);
	}
	
	/**
	 * The constructor.
	 * @param aname
	 * @param aClosedWorld
	 */
	public FelixPredicate(String aname, boolean aClosedWorld) {
		super(null, aname, aClosedWorld);
	}

	/**
	 * Get {@link FelixPredicate#keyPositions}.
	 * @return
	 */
	public HashSet<Integer> getKeyPositions(){
		HashSet<Integer> ret = new HashSet<Integer>();
		for(int i=0;i<this.arity();i++){
			if(this.labelPositions.contains(i)){
				continue;
			}
			ret.add(i);
		}
		return ret;
	}
	
	/**
	 * Flushes string-based evidence to database.
	 * @param rName
	 */
	public void flushStrEvidence(String rName) {
		try {
			
			//create table
			ArrayList<String> tableColumn = new ArrayList<String>();
			tableColumn.add("truth BOOL");
			tableColumn.add("prior FLOAT");
			for(String arg : this.getArgs()){
				tableColumn.add(arg + " TEXT");
			}
			String sql = "CREATE TABLE " + rName + " ( " + StringMan.commaList(tableColumn) + ");";
			RDB db = RDB.getRDBbyConfig(FelixConfig.db_schema);
			db.dropTable(rName);
			db.execute(sql);
			db.commit();
			
			// flush the file
			loadingFileWriter.close();
			loadingFileWriter = null;
			// copy into DB
			ArrayList<String> cols = new ArrayList<String>();
			cols.add("truth");
			cols.add("prior");
			cols.addAll(this.getArgs());
			FileInputStream in = new FileInputStream(loadingFile);
			PGConnection con = (PGConnection) db.getConnection();
			sql = "COPY " + rName + 
			StringMan.commaListParen(cols) + " FROM STDIN";
			con.getCopyAPI().copyIn(sql, in);
			in.close();
			db.commit();
			db.analyze(rName);
			FileMan.removeFile(loadingFile.getAbsolutePath());
			
			db.close();
			
		} catch (Exception e) {
			ExceptionMan.handle(e);
		}
	}
	
	/**
	 * Get all clauses related to this predicate.
	 * @return
	 */
	public HashSet<FelixClause> getRelevantClauses(){
		return registeredClauses;
	}
	
	/**
	 * Add a property to this predicate, along with the clause with this property.
	 * @param prop
	 * @param evid
	 * @param _pos if this property is key constraint, this parameter is the position
	 * of keys.
	 */
	public void registerProperty(FPProperty prop, FelixClause evid, int... _pos){
		
		FelixUIMan.println(2, 0, "\nRegister property {" + prop + "} to Predicate {"
				+ this.getName() + "(" + StringMan.join(",", this.getArgs()) + ")" 
				+ "} with evidence {" + evid + "} {" 
				+ FelixUIMan.joinArray(_pos) + "}");
		
		if(prop == FPProperty.EMBED_WEIGHT_RULE){
			if(!this.properities.containsKey(FPProperty.EMBED_WEIGHT_RULE)){
				this.properities.put(FPProperty.EMBED_WEIGHT_RULE, new HashSet<FelixClause>());
			}
			this.properities.get(FPProperty.EMBED_WEIGHT_RULE).add(evid);
			return;
		}
		
		if(prop == FPProperty.KEY_CONSTRAINT){
			
			if(_pos.length == 0){
				ExceptionMan.die("Cannot register key constriant without " +
						"any given positions");
			}
			
			if(this.labelPositions.size() != 0){
				UIMan.warn("Ignore second key constraints of predicate "
						+ this.getName());
			}else{
				for(int i=0;i<_pos.length;i++){
					this.labelPositions.add(_pos[i]);
				}
			}
		}
		
		if(this.hasDependentAttributes() == false && prop != FPProperty.SYMM 
				&& prop != FPProperty.REFLEX && evid == null){
			ExceptionMan.die("NULL clause assigned to non key_constraint predicate");
		}
		
		if(registeredClauses.contains(evid)){
			FelixUIMan.println(2, 0, "Ignore replicated clause registering for "
					+ "property {" + prop + "} to Predicate {"
					+ this.getName() + "(" + StringMan.join(",", this.getArgs()) + ")" 
					+ "} with evidence {" + evid + "} {" 
					+ FelixUIMan.joinArray(_pos) + "}");
		}else{
			if(evid != null){
				if(!properities.containsKey(prop)){
					properities.put(prop, new HashSet<FelixClause>());
				}
				properities.get(prop).add(evid);
				registeredClauses.add(evid);
				
				if(prop == FPProperty.CHAIN_RECUR){
					this.tryToExtractSeqFromChainRule(evid);
				}
			}else{
				if(!properities.containsKey(prop)){
					properities.put(prop, new HashSet<FelixClause>());
				}
			}
		}
	}
	
	/**
	 * If the given clause is a CRF chain rule, try to extract the
	 * partitions of sequence.
	 * @param fc
	 */
	public void tryToExtractSeqFromChainRule(FelixClause fc){
		
		ArrayList<String> ret = new ArrayList<String>();
		
		Literal l1 = null;
		Literal l2 = null;
		
		for(Literal l : fc.getRegLiterals()){
			if(l.getPred().getName().equals(this.getName())){
				if(l1 == null){
					l1 = l;
				}else{
					l2 = l;
				}
			}
		}
		
		for(int i=0;i<this.arity();i++){
			// seq partition must be on keys
			if(this.labelPositions.contains(i)){
				continue;
			}
			
			if(l1.getTerms().get(i).toString().equals(
					l2.getTerms().get(i).toString())){
				ret.add(this.getArgs().get(i));
			}
		}
		
		this.chainRulePartitions.put(fc, ret);
		
	}
	
	/**
	 * Returns string representation of this predicate.
	 */
	public String toString(){
		String ret = "";
		
		ret = this.getName();
		ret += "(";
		ret += StringMan.commaList(this.getArgs());
		ret += ")";
		
		return ret;
	}

	
}
