package felix.dstruct;

import java.io.BufferedReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.TreeMap;
import java.util.concurrent.ConcurrentHashMap;

import tuffy.db.RDB;
import tuffy.mln.Literal;
import tuffy.mln.Predicate;
import tuffy.mln.Type;
import tuffy.ra.ConjunctiveQuery;
import tuffy.ra.Function;
import tuffy.util.DebugMan;
import tuffy.util.ExceptionMan;
import tuffy.util.FileMan;
import tuffy.util.Timer;
import tuffy.util.UIMan;
import felix.parser.FelixInputParser;
import felix.task.ConcurrentEvidenceLoader;
import felix.util.*;


/**
 * An object of FelixQuery consists of program, query and evidence
 * from inputs.
 * @author Ce Zhang
 *
 */
public class FelixQuery {

	/**
	 * Map from predicate name to rules used for coref that is 
	 * difficult to be represented by first order logic.
	 */
	HashMap<String, ArrayList<ConjunctiveQuery>> specialClusteringRules = 
		new HashMap<String, ArrayList<ConjunctiveQuery>>();
	
	/**
	 * All input clauses.
	 */
	ArrayList<FelixClause> clauses = new ArrayList<FelixClause>();
	
	/**
	 * Map from integer IDs to clauses
	 */
	public HashMap<Integer, FelixClause> id2clause = new HashMap<Integer, FelixClause>();
	
	/**
	 * All input predicates.
	 */
	ArrayList<FelixPredicate> predicates = new ArrayList<FelixPredicate>();

	//TODO: scoping rule in statOperator?
	/**
	 * All input scoping rules
	 */
	ArrayList<ConjunctiveQuery> scopingRules = new ArrayList<ConjunctiveQuery>();
	
	/**
	 * All input datalog rules
	 */
	ArrayList<ConjunctiveQuery> datalogRules = new ArrayList<ConjunctiveQuery>();

	/**
	 * Map from predicate's name to predicate.
	 */
	HashMap<String, FelixPredicate> nameMapPred = new HashMap<String, FelixPredicate>();
	
	/**
	 * Map from type's name to type.
	 */
	private Hashtable<String, Type> nameMapType = new Hashtable<String, Type>();
	
	/**
	 * Map from function's name to function.
	 */
	private Hashtable<String, Function> nameMapFunc = new Hashtable<String, Function>();
	
	/**
	 * Map from constant to constant ID.
	 */
	public ConcurrentHashMap<String, Integer> mapConstantID = new ConcurrentHashMap<String, Integer>();
	
	/**
	 * Map from constant ID to constant.
	 */
	public ConcurrentHashMap<Integer, String> mapIDConstant = new ConcurrentHashMap<Integer, String>();
	
	/**
	 * @deprecated
	 */
	public ConcurrentHashMap<String, Integer> newMapConstantID = new ConcurrentHashMap<String, Integer>();
		
	
	/**
	 * The parser used to parse the input.
	 */
	FelixInputParser parser = new FelixInputParser(this);

	/**
	 * Add a clause to FelixQuery.
	 * @param fc
	 */
	public void addFelixClause(FelixClause fc) {
		fc.setId(this.clauses.size() + 1);
		this.clauses.add(fc);
		this.id2clause.put(fc.getId(), fc);
	}

	/**
	 * Add a predicate to FelixQuery.
	 * @param fp
	 */
	public void addFelixPredicate(FelixPredicate fp) {

		if (nameMapPred.containsKey(fp.getName())) {
			ExceptionMan.die("Duplicate predicate definitions - "
					+ fp.getName());
		}

		fp.setID(predicates.size());
		predicates.add(fp);
		nameMapPred.put(fp.getName(), fp);

	}

	/**
	 * Get all scoping rules.
	 * @return
	 */
	public ArrayList<ConjunctiveQuery> getScopingRules(){
		return this.scopingRules;
	}
	
	/**
	 * Add a scoping rule to FelixQuery.
	 * @param sr
	 */
	public void addScopingRule(ConjunctiveQuery sr) {
		this.scopingRules.add(sr);
	}

	/**
	 * Add a Datalog rule to FelixQuery.
	 * @param dr
	 */
	public void addDatalogRule(ConjunctiveQuery dr) {
		this.datalogRules.add(dr);
	}

	/**
	 * Get predicate by name.
	 * @param name
	 * @return
	 */
	public FelixPredicate getPredByName(String name) {
		return nameMapPred.get(name);
	}
	
	/**
	 * Get all predicates.
	 * @return
	 */
	public ArrayList<FelixPredicate> getPredicates(){
		return this.predicates;
	}

	/**
	 * Get or create type by name.
	 * @param name
	 * @return
	 */
	public Type getOrCreateTypeByName(String name) {
		Type t = nameMapType.get(name);
		if (t == null) {
			t = new Type(name);
			nameMapType.put(name, t);
		}
		return t;
	}

	/**
	 * Get function by name.
	 * @param name
	 * @return
	 */
	public Function getFunctionByName(String name) {
		Function f = Function.getBuiltInFunctionByName(name);
		if (f != null) {
			return f;
		}
		return nameMapFunc.get(name);
	}

	/**
	 * Add a constant to FelixQuery.
	 * @param symbol
	 * @return
	 */
	public int addConstant(String symbol) {
				
		Integer id = mapConstantID.get(symbol);
		boolean isnew = false;
		isnew = (id==null);
		if (id == null) {
			id = mapConstantID.size() + 1;
			mapConstantID.putIfAbsent(symbol, id);
			// Clause.mappingFromID2Const.put(id, symbol);
			
			if(FelixConfig.evidDBSchema != null && isnew && loadingEvid == true){
				newMapConstantID.putIfAbsent(symbol, id);
			}			
		}
		mapIDConstant.putIfAbsent(id, symbol);
		return id;

	}
	
	/**
	 * Whether we consider the evidence file while parsing operators.
	 */
	public boolean loadingEvid = false;

	/**
	 * Get or add (if not exists) a symbol to a type's constant table.
	 * @param symbol
	 * @param type
	 * @return
	 */
	public int getSymbolID(String symbol, Type type) {
	
		Integer id = mapConstantID.get(symbol);
		boolean isnew = false;
		isnew = (id==null);
		if (id == null) {
			id = mapConstantID.size() + 1;
			mapConstantID.putIfAbsent(symbol, id);
			
			if(FelixConfig.evidDBSchema != null && isnew && loadingEvid == true){
				newMapConstantID.putIfAbsent(symbol, id);
			}
			
			// Clause.mappingFromID2Const.put(id, symbol);
		}
		if (type != null && !type.isNonSymbolicType()) {
			if(FelixConfig.evidDBSchema != null && isnew && loadingEvid == false){
				return id;
			}
			type.addConstant(id);
		}
		return id;

	}

	/**
	 * Load MLN programs.
	 * @param progFiles
	 */
	public void loadPrograms(String[] progFiles) {
		for (String f : progFiles) {
			String g = FileMan.getGZIPVariant(f);
			if (g == null) {
				ExceptionMan.die("non-existent file: " + f);
			} else {
				f = g;
			}
			UIMan.println(">>> Parsing program file: " + f);
			parser.parseProgramFile(f);
		}
	}

	/**
	 * Load MLN queries.
	 * @param queryFiles
	 */
	public void loadQueries(String[] queryFiles) {
		for (String f : queryFiles) {
			String g = FileMan.getGZIPVariant(f);
			if (g == null) {
				ExceptionMan.die("non-existent file: " + f);
			} else {
				f = g;
			}
			UIMan.println(">>> Parsing query file: " + f);
			parser.parseQueryFile(f);
		}
	}

	/**
	 * Load MLN queries specified by command lines.
	 * @param queryAtoms
	 */
	public void parseQueryCommaList(String queryAtoms) {
		parser.parseQueryCommaList(queryAtoms);
	}
	
	/**
	 * Load MLN queries specified by command lines.
	 * @param queryAtoms
	 */
	public void parseProgFromString(String prog) {
		parser.parseProgramFileFromString(prog);
	}

	/**
	 * Close all predicate's loadingFile.
	 */
	public void closeFiles() {
		for (FelixPredicate p : predicates) {
			p.closeFiles();
		}
	}

	/**
	 * Get all predicates in this FelixQuery.
	 * @return
	 */
	public HashSet<FelixPredicate> getAllPred() {
		return new HashSet<FelixPredicate>(predicates);
	}

	/**
	 * Get all predicates which are open in this FelixQuery.
	 * @return
	 */
	public HashSet<FelixPredicate> getAllOpenPred(){
		HashSet<FelixPredicate> ret = new HashSet<FelixPredicate>();
		
		for(FelixPredicate fp : this.getAllPred()){
			if(fp.isClosedWorld() == false){
				ret.add(fp);
			}
		}
		
		return ret;
	}
	
	/**
	 * Get all clauses in this FelixQuery.
	 * @return
	 */
	public HashSet<FelixClause> getAllClause(){
		return new HashSet<FelixClause>(clauses);
	}
	
	/**
	 * Load MLN evidences.
	 * @param evidFiles
	 * @throws InterruptedException 
	 */
	public void loadEvidences(String[] evidFiles) throws InterruptedException {
		
		ArrayList<ConcurrentEvidenceLoader> loaders = 
				new ArrayList<ConcurrentEvidenceLoader>();
		
		for(String f : evidFiles){
			ConcurrentEvidenceLoader loader = 
					new ConcurrentEvidenceLoader(f, this);
			loaders.add(loader);
			loader.run();
		}
		
	}

	//TODO: think about whether there are repeated materializations
	//that can be reuse.
	/**
	 * Materialize database tables.
	 * @param db
	 */
	public void materializeTables(RDB db) {
		FelixUIMan.println(1, ">>> Storing symbol tables...");
		FelixUIMan.println(2, "# constants = " + mapConstantID.size());
		if(FelixConfig.evidDBSchema == null){
			db.createConstantTable(mapConstantID);
		}else{
			db.insertConstantTable(newMapConstantID);
		}
		mapConstantID = null;
		try {
			DebugMan.runGC();
			DebugMan.runGC();
			DebugMan.runGC();
		} catch (Exception e) {
			e.printStackTrace();
		}

		for (Type t : nameMapType.values()) {
			if(FelixConfig.evidDBSchema == null){
				t.storeConstantList(db);
			}else{
				t.storeConstantList(db,true);
			}
			//db.execute("CREATE INDEX index_type_" + t.name + "_id ON " + t.getRelName() + "(constantid)");
			//db.execute("CREATE INDEX index_type_" + t.name + "_value ON " + t.getRelName() + "(constantvalue)");
			//db.analyze(t.getRelName());
		}

		UIMan.println(">>> Storing evidence...");

		for (Predicate p : getAllPred()) {
			if(FelixConfig.evidDBSchema == null){
				p.flushEvidence();
			}else{
				p.flushEvidence(true);
			}
			
		}
	}

	/**
	 * Execute all Datalog rules.
	 * @param db
	 */
	public void executeAllDatalogRules(RDB db) {
		if (datalogRules.isEmpty())
			return;

		long total = 0;
		UIMan.println(">>> Executing Datalog rules...");
		for (ConjunctiveQuery cq : datalogRules) {
			UIMan.print(".");
			Predicate p = cq.head.getPred();
			FelixUIMan.println(1, cq.toString());
			Timer.start("datalogq");
			//cq.buildIndexes(db, null, null, null, false, new ArrayList<String>());
			cq.materialize(db, true, new ArrayList<String>());
			int ni = db.getLastUpdateRowCount();
			total += ni;
			FelixUIMan.println(1, "### inserted " + UIMan.comma(ni)
					+ (ni != 1 ? " new tuples" : " new tuple"));
			FelixUIMan.println(1, "### current cardinality of '" + p.getName()
					+ "' = " + UIMan.comma(db.countTuples(p.getRelName())));
			String tm = Timer.elapsed("datalogq");
			FelixUIMan.println(1, "### took time " + tm + "\n");
			db.analyze(p.getRelName());
		}
		UIMan.println();
	}
	
	/**
	 * Get Class and Tag rules for clustering predicates.
	 * @param _predName
	 * @return
	 */
	public ArrayList<ConjunctiveQuery> getSpecialClusteringRules(String _predName){
		if(this.specialClusteringRules.containsKey(_predName)){
			return this.specialClusteringRules.get(_predName);
		}else{
			return new ArrayList<ConjunctiveQuery>();
		}
	}
	
	/**
	 * Add Class or Tag rules for clustering predicates.
	 * @param cq
	 */
	public void registerClusteringRule(ConjunctiveQuery cq){
		
		Literal lhead = cq.head;
		String _pname = lhead.getPred().getName();
		if(!this.specialClusteringRules.containsKey(_pname)){
			this.specialClusteringRules.put(_pname, new ArrayList<ConjunctiveQuery>());
		}
		this.specialClusteringRules.get(_pname).add(cq);
	}

}





