package felix.dstruct;

import java.util.HashMap;
import java.util.HashSet;

import tuffy.mln.Predicate;


/**
 * The class of an OperatorBucketGraph. Each operator graph contains 1) a set of 
 * {@link ConcurrentOperatorsBucket}; and 2) Dependencies between different buckets.
 * Current assumption is DAG (although there are no data-structure-level constraints
 * for DAG, some algorithms may become trivial when dealing with cycles).
 * An OperatorBucketGraph is a complete description of the logic plan.
 * 
 * @author Ce Zhang
 *
 */
public class OperatorBucketGraph {
	
	/**
	 * Counter for buckets 
	 * in this OperatorBucketGraph.
	 */
	int counter = 1;

	/**
	 * Set of operators in this OperatorBucketGraph.
	 */
	HashSet<ConcurrentOperatorsBucket> operators = new HashSet<ConcurrentOperatorsBucket>();
	
	/**
	 * Map from the bucket to its upstream buckets.
	 */
	public HashMap<ConcurrentOperatorsBucket, HashSet<ConcurrentOperatorsBucket>> upStreams = 
		new HashMap<ConcurrentOperatorsBucket, HashSet<ConcurrentOperatorsBucket>>();
	
	/**
	 * Map from the bucket to its downstream buckets.
	 */
	public HashMap<ConcurrentOperatorsBucket, HashSet<ConcurrentOperatorsBucket>> downStreams = 
		new HashMap<ConcurrentOperatorsBucket, HashSet<ConcurrentOperatorsBucket>>();
	
	/**
	 * Add a new bucket to this OperatorBucketGraph. An ID will be
	 * automatically assigned to this bucket.
	 * @param sop statistical operator.
	 */
	public void addOperator(ConcurrentOperatorsBucket sop){
		assert(this.operators.contains(sop) == false);
		
		sop.id = this.counter ++;
		
		this.operators.add(sop);
		this.upStreams.put(sop, new HashSet<ConcurrentOperatorsBucket>());
		this.downStreams.put(sop, new HashSet<ConcurrentOperatorsBucket>());
	}
	
	/**
	 * Add a new bucket relation to this OperatorBucketGraph. Here by ``bucket relation''
	 * it means a directed edge between two buckets from upstream to downstream.
	 * @param upStream
	 * @param downStream
	 */
	private void addOperatorRelation(ConcurrentOperatorsBucket upStream, ConcurrentOperatorsBucket downStream){
		
		assert(this.operators.contains(upStream) == true);
		assert(this.operators.contains(downStream) == true);
		
		this.upStreams.get(downStream).add(upStream);
		this.downStreams.get(upStream).add(downStream);
		
	}
	
	/**
	 * Parse the dependency between buckets. This function will transform
	 * the input/output predicates information of each statistical bucket
	 * into the upstream/downstream relations between statistical buckets.
	 * Note that, this upstream/downstream relation is only a preference and is
	 * not necessarily as the same as the final execution plan.
	 * 
	 */
	public void parseDependency(){
		
		HashMap<Predicate, HashSet<ConcurrentOperatorsBucket>> regardPredicateAsInput = 
			new HashMap<Predicate, HashSet<ConcurrentOperatorsBucket>>();
		HashMap<Predicate, HashSet<ConcurrentOperatorsBucket>> regardPredicateAsOutput = 
			new HashMap<Predicate, HashSet<ConcurrentOperatorsBucket>>();
		
		for(ConcurrentOperatorsBucket op : this.operators){
			for(Predicate p : op.inputPredicates){
				if(!regardPredicateAsInput.containsKey(p)){
					regardPredicateAsInput.put(p, new HashSet<ConcurrentOperatorsBucket>());
				}
				regardPredicateAsInput.get(p).add(op);
			}
			for(Predicate p : op.outputPredicates){
				if(!regardPredicateAsOutput.containsKey(p)){
					regardPredicateAsOutput.put(p, new HashSet<ConcurrentOperatorsBucket>());
				}
				regardPredicateAsOutput.get(p).add(op);
			}
		}
		
		for(ConcurrentOperatorsBucket op : this.operators){
			for(Predicate p : op.inputPredicates){
				if(regardPredicateAsOutput.get(p) == null){
					continue;
				}
				for(ConcurrentOperatorsBucket op2 : regardPredicateAsOutput.get(p)){
					this.addOperatorRelation(op2, op);
				}
			}
		}
		
	}
	
	/**
	 * Return the set of buckets in this OperatorBucketGraph.
	 */
	public HashSet<ConcurrentOperatorsBucket> getOperators(){
		return this.operators;
	}
	
	/**
	 * Randomly pick one bucket in this OperatorBucketGraph.
	 */
	public ConcurrentOperatorsBucket getOneRandomOperator(){
		while(true){
			for(ConcurrentOperatorsBucket op : this.operators){
				if(Math.random() < 0.5){
					return op;
				}
			}
		}
	}
	
	/**
	 * Gets all upstream buckets of the given bucket.
	 */
	public HashSet<ConcurrentOperatorsBucket> getUpStreamOperator(ConcurrentOperatorsBucket sop){
		return this.upStreams.get(sop);
	}
	
	/**
	 * Gets all downstream buckets of the given bucket.
	 */
	public HashSet<ConcurrentOperatorsBucket> getDownStreamOperator(ConcurrentOperatorsBucket oup){
		return this.downStreams.get(oup);
	}
		
}
