package felix.dstruct;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;

import felix.dstruct.FelixPredicate.FPProperty;
import felix.optimizer.DataCracker1991;
import felix.parser.FelixCommandOptions;
import felix.util.FelixStringMan;
import felix.util.FelixUIMan;

import tuffy.db.RDB;
import tuffy.mln.Predicate;
import tuffy.ra.ConjunctiveQuery;
import tuffy.ra.Expression;
import tuffy.util.ExceptionMan;
import tuffy.util.StringMan;


/**
 * Abstract class of a statistical operator (e.g., Tuffy, Coref, CRF etc. ). 
 * A valid Felix operator should extend this class and implement methods 
 * like run(), prepare(), etc.
 * 
 * @author Ce Zhang
 *
 */
public abstract class StatOperator extends Thread implements Cloneable{
	
	/**
	 * Parameter assigning the degree of data partitioning.
	 */
	public int partitionedInto = 1;
	
	public DataCracker1991 dc = null;
	
	public int dPart = 1;
	
	public int nPart = 1;
	
	/**
	 * @deprecated
	 */
	public boolean currentState = false;
	
	/**
	 * The set of predicates that are shared with other
	 * operators via dual decomposition.
	 */
	protected HashSet<FelixPredicate> dd_CommonOutput = new HashSet<FelixPredicate>();
	
	/**
	 * Map from predicate to database table name
	 */
	protected HashMap<FelixPredicate, String> dd_commonOutputPredicate_2_tableName = 
		new HashMap<FelixPredicate, String>();
	
	/**
	 * Type of StatOperator.
	 */
	public enum OPType {CRF, LR, COREF, TUFFY};
	
	/**
	 * The human-readable representation of data partitions. This is only used in toString()
	 */
	public String dataCrackerSignature = null;
	
	/**
	 * Rules for Lagrangian Multipliers. 
	 */
	public HashSet<FelixClause> dd_PriorClauses = new HashSet<FelixClause>();
	
	/**
	 * Map from clauses to expressions partitioning the data. 
	 */
	public HashMap<FelixClause, HashSet<Expression>> clauseConstraints =
		new HashMap<FelixClause, HashSet<Expression>>();
	
	/**
	 * The ConcurrentOperatorsBucket this StatOperator belongs to.
	 * StatOperators in one ConcurrentOperatorsBucket can be executed
	 * in parallel.
	 */
	public ConcurrentOperatorsBucket belongsToBucket = null;
	
	/**
	 * Whether this operator works in marginal or MAP mode.
	 */
	public boolean isMarginal = true;
	
	/**
	 * Options parsed from command line or configuration file.
	 */
	protected FelixCommandOptions options = null;
	
	/**
	 * Predicates computed by previous operators.
	 */
	protected HashSet<String> inputPredicateScope = new HashSet<String>();
	
	/**
	 * Max time this operator can run. This parameter will be useful 
	 * when there are cycles in the operator graph.
	 * @deprecated
	 */
	int maxRunTime = 5;
	
	/**
	 * Set of predicates that can potentially be shared with
	 * other operators via dual decomposition.
	 */
	public HashSet<FelixPredicate> commonCandidate = new HashSet<FelixPredicate>();
	
	/**
	 * How many times this operator has run.
	 * @deprecated
	 */
	int currentRunTime = 0;
	
	/**
	 * Global counter of StatOperators ID.
	 */
	static int idCounter = 1;
	
	/**
	 * ID of this operator. Note that, the statistical operator returned by
	 * .clone() has a different ID.
	 */
	int id = -1;
	
	/**
	 * Type of this operator.
	 */
	public OPType type = null;
	
	/**
	 * Set of predicates whose values are used by this operator as inputs.
	 */
	public HashSet<FelixPredicate> inputPredicates = new HashSet<FelixPredicate>();
	
	/**
	 * Set of predicates whose values are output by this operator.
	 */
	public HashSet<FelixPredicate> outputPredicates = new HashSet<FelixPredicate>();
	
	/**
	 * Set of clauses assigned to this operator.
	 */
	public HashSet<FelixClause> allRelevantFelixClause = new HashSet<FelixClause>();
	 
	/**
	 * Database connection.
	 */
	public RDB db;
	
	/**
	 * Felix query.
	 */
	public FelixQuery fq;
	
	/**
	 * @deprecated
	 */
	public boolean isBinaryArbLR = false;
	
	/**
	 * The precedence of this operator. The larger this value, the earlier this operator 
	 * will be run in the final physical plan. 
	 */
	protected int precedence = -1;
	
	/**
	 * @deprecated
	 * @param _predicates
	 */
	public void pushPredicateScopes(HashSet<String> _predicates){
		this.inputPredicateScope.addAll(_predicates);
	}
	
	/**
	 * Returns a clone of this statistical operator.
	 */
	@SuppressWarnings("unchecked")
	public StatOperator clone(){
		try{
			StatOperator ret;
			
			if(this.type == OPType.LR){
				ret = this.getClass().getConstructor(
					this.fq.getClass(), 
					this.outputPredicates.getClass(), 
					this.options.getClass())
				.newInstance(this.fq, this.outputPredicates, this.options);
			}else{
				ret = this.getClass().getConstructor(
						this.fq.getClass(), 
						this.outputPredicates.getClass(), 
						this.options.getClass())
					.newInstance(this.fq, this.outputPredicates, this.options);
			}

			
			for(FelixClause fc : this.allRelevantFelixClause){
				ret.registerRelevantClause(fc);
			}
			ret.sealDefinition();
			
			
			ret.clauseConstraints = (HashMap<FelixClause, HashSet<Expression>>) 
										this.clauseConstraints.clone();
			ret.belongsToBucket = this.belongsToBucket;
			ret.isMarginal = this.isMarginal;
			ret.precedence = this.precedence;
			
			return ret;		
		}catch(Exception e){
			e.printStackTrace();
			return null;
		}

	}
	
	/**
	 * See {@link ConcurrentOperatorsBucket#nStartingRule}.
	 */
	public int nStartingRules = 0;
	
	/**
	 * Sets precedence of this statistical operator.
	 * @param _value
	 */
	public void setPrecedence(int _value){
		this.precedence = _value;
	}
	
	/**
	 * Returns set of all clauses associated with the given property.
	 * @param col
	 * @param prop
	 * @return
	 */
	HashSet<FelixClause> getPropertyClausesUnion(Collection<FelixPredicate> col, FPProperty prop){
		HashSet<FelixClause> ret = new HashSet<FelixClause>();
		for(FelixPredicate fp : col){
			ret.addAll(fp.getPropertyClauses(prop));
		}
		return ret;
	}
	
	/**
	 * the constructor.
	 * @param _oriMLN
	 * @param _opt
	 */
	public StatOperator(FelixQuery _fq, HashSet<FelixPredicate> _goalPredicates, FelixCommandOptions _opt){
		this.outputPredicates.clear();
		this.outputPredicates.addAll(_goalPredicates);
		options = _opt;
		id = idCounter++;
		fq = _fq;
		this.isMarginal = _opt.marginal;
			
	}
	
	/**
	 * Registers clause to this statistical operator.
	 * @param fc
	 */
	public void registerRelevantClause(FelixClause fc){
		this.allRelevantFelixClause.add(fc);
	}
	
	/**
	 * According to the clauses added via {@link #registerRelevantClause(FelixClause)},
	 * parse {@link #outputPredicates}, {@link #inputPredicates}, {@link #nStartingRules},
	 * and {@link #commonCandidate}.
	 * 
	 */
	public void sealDefinition(){
		
		for(FelixClause fc : this.allRelevantFelixClause){
			
			int nOpen = 0;
					
			for(Predicate p : fc.getReferencedPredicates()){
				
				if(!p.isClosedWorld() && !outputPredicates.contains(p)){
					nOpen ++;
				}
				
				FelixPredicate fp = fq.getPredByName(p.getName());
				
				if(this.outputPredicates.contains(fp)){
					continue;
				}
							
				if(fp.isCorefMapPredicate){
					
					this.inputPredicates.add(fp.getOriCorefPredicate());
					
				}else{
			
					if(fp.isClosedWorld() == true){
						continue;
					}
					
					if(this.isBinaryArbLR == false){
						this.inputPredicates.add(fp);
					}else{
						if(this.getPropertyClausesUnion(outputPredicates, FPProperty.NON_RECUR).contains(fc)){
							this.inputPredicates.add(fp);
						}
					}
				}
			
			}
							
			for(Predicate p : fc.getReferencedPredicates()){
				if(!p.isClosedWorld() || fq.getPredByName(p.getName()).isCorefMapPredicate){
					this.commonCandidate.add(fq.getPredByName(p.getName()));
				}
			}
				
			if(nOpen == 0){
				this.nStartingRules ++;
			}
			
		}
	}
	
	
	/**
	 * All {@link DataMovementOperator}s used in this statistical operator.
	 */
	protected List<DataMovementOperator> allDMOs = new ArrayList<DataMovementOperator>();
	
	
	/**
	 * Generate the operator-specified logic plan, i.e., all data movement operators 
	 * that will be used for inference.
	 * 
	 * This function should be invoked after a new instance of operator is created.
	 * Any valid Felix operator should implement this method.
	 */
	public abstract void prepare();
	
	/**
	 * Method that executes this operator. Any valid Felix operator should implement 
	 * this method.
	 */
	abstract public void run();
	
	/**
	 * @deprecated
	 */
	abstract public void learn();

	/**
	 * Human-readable representation of the logic plan. Any valid Felix operator should 
	 * implement this method.
	 * 
	 * TODO: need to think out a better to explain physical plan. (e.g., a graph?)
	 */
	abstract public String explain();
	
	/**
	 * @deprecated
	 */
	public HashSet<String> throwAwayPredicatesNames = new HashSet<String>();
	
	/**
	 * Get the target predicate of this StatOperator if this operator is
	 * CRF, LR or COREF.
	 * @return
	 */
	public FelixPredicate getTargetPredicateIfHasOnlyOne(){
		if(this.outputPredicates.size() != 1){
			ExceptionMan.die("Why this operator has >two or =zero output predicates?");
		}	
		return this.outputPredicates.iterator().next();
	}
	
	
	/**
	 * Given a first order logic clause and a target predicate, translate it into
	 * a conjunctive query with properly assigned weight.
	 * 
	 * @param target
	 * @param forceRecursive If there are multiple literals of the given predicate,
	 * this parameter specifies whether the generated conjunctive queries are recursive.
	 * CRF will set this parameter as FALSE, while COREF will set it as TRUE.
	 * @param props The property selected for the target predicate.
	 * @return
	 */
	public HashSet<ConjunctiveQuery> translateFelixClasesIntoFactorGraphEdgeQueries(
			FelixPredicate target, 
			boolean forceRecursive,
			HashSet<String> allowedOpenPredicates,
			FPProperty... props){
		
		HashSet<FelixClause> rules = new HashSet<FelixClause>();
		
		for(FPProperty prop : props){
			rules.addAll(target.getPropertyClauses(prop));
		}
		
		if(options.useDualDecomposition){
			for(FPProperty prop : props){
				if(prop == FPProperty.NON_RECUR){
					rules.addAll(this.dd_PriorClauses);
				}
			}
		}
			
		HashSet<ConjunctiveQuery> ret = new HashSet<ConjunctiveQuery> ();
		
		FelixUIMan.println(2, 0, "{" + StringMan.join(",", FelixStringMan.colToStringArray(props)) + 
				 "} ConjunctiveQueries for " + this.toString());
		for(FelixClause rule : rules){
			
			boolean flag = false;
			boolean notRelated = true;
			for(Predicate fp : rule.getReferencedPredicates()){
				if(!fp.isClosedWorld() && !fp.equals(target)){
					if(!allowedOpenPredicates.contains(fp.getName())){
						flag = true;
					}
					
					if(this.throwAwayPredicatesNames.contains(fp.getName())){
						flag = true;
					}
				}
				
				if(fp.equals(target)){
					notRelated = false;
				}
			}
			
			if(notRelated == true){
				continue;
			}
			
			if(flag == true){
				continue;
			}
			
			if(this.isBinaryArbLR == true){
				rule.isBinaryLRRules = true;
			}
			
			ConjunctiveQuery returnedCQ = rule.toSimplifiedFactorGraphQuery(this, target, forceRecursive);
			ret.add(returnedCQ);
			FelixUIMan.printobj(2, 0, returnedCQ);
		}
		
		return ret;
	}
	
	/**
	 * Get all DMOs used by this operator.
	 * @return
	 */
	public List<DataMovementOperator> getAllDMOs(){
		return this.allDMOs;
	}
	
	/**
	 * Get the precedence of this operator.
	 * @return
	 */
	public int getPrecedence(){
		return this.precedence;
	}
		
	public String toString(){
		String ret = FelixStringMan.indentHead();
		
		ret += "{" + this.type + "} ";
		ret += "Operator of {";
		ret += StringMan.join(",", FelixStringMan.colToStringArray(this.outputPredicates));
		ret += "} with {" + this.allRelevantFelixClause.size() + "} Relevant Clauses";
		
		if(this.clauseConstraints.size() == 0){
			ret += " Communicate { " + 
					StringMan.join(",", FelixStringMan.colToStringArray(this.dd_CommonOutput))
					+ "}";
			return ret;
		}
		
		ret += " Partitioned by: ";
		
		if(dataCrackerSignature != null){
			ret += dataCrackerSignature;
		}
		
		ret += " Communicating { " + 
				StringMan.join(",", FelixStringMan.colToStringArray(this.dd_CommonOutput))
				+ "}";
		
		return ret;
	}
	
	public String toNoParString(){
		String ret = FelixStringMan.indentHead();
		
		ret += "{" + this.type + "} ";
		ret += "Operator of {";
		ret += StringMan.join(",", FelixStringMan.colToStringArray(this.outputPredicates));
		ret += "} ";
		
		ret += " Communicating { " + 
				StringMan.join(",", FelixStringMan.colToStringArray(this.dd_CommonOutput))
				+ "}";
		
		return ret;
	}
	
	//TODO: Maybe add legibility-checking of rules?
}










