package felix.executor;

import java.io.BufferedWriter;
import java.io.File;
import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;

import com.google.common.io.Files;

import tuffy.db.RDB;
import tuffy.infer.DataMover;
import tuffy.mln.MarkovLogicNetwork;
import tuffy.mln.Predicate;
import tuffy.parse.CommandOptions;
import tuffy.util.Config;
import tuffy.util.ExceptionMan;
import tuffy.util.FileMan;
import tuffy.util.StringMan;
import tuffy.util.UIMan;

import felix.dstruct.ConcurrentOperatorsBucket;
import felix.dstruct.ExecutionPlan;
import felix.dstruct.FelixPredicate;
import felix.dstruct.FelixQuery;
import felix.dstruct.StatOperator;
import felix.dstruct.StatOperator.OPType;
import felix.main.Felix;
import felix.optimizer.DMOOptimizer;
import felix.parser.FelixCommandOptions;
import felix.util.FelixConfig;
import felix.util.FelixUIMan;



/**
 * Class for executing a given physical {@link ExecutionPlan} using 
 * dual decomposition.
 * 
 * @author Ce Zhang
 *
 */
public class DDExecutor {

	/**
	 * The execution plan to be executed.
	 */
	ExecutionPlan ep;
	
	/**
	 * The FelixQuery used by this Executor.
	 */
	FelixQuery fq;
	
	/**
	 * The FelixCommandOptions used by this Executor.
	 */
	FelixCommandOptions options;
	
	/**
	 * The DMOOptimizer used by this Executor.
	 */
	public DMOOptimizer dmoo;
		
	/**
	 * Execute this plan. If Felix runs in
	 * explain mode, just prints out the physical plan.
	 * 
	 * <p> TODO: Find a better way to the explain mode (e.g., a graph).
	 */
	@SuppressWarnings("unchecked")
	public void run(){
		
		try{
			
			if(FelixConfig.explainMode){
				FelixUIMan.println(0, 0, "\nExecution Plan:\n");
			}
			
			for(int i = ep.operators.size()-1; i >=0; i --){
				
					if(FelixConfig.explainMode){
						FelixUIMan.println("Concurrently:");
					}
					
					for(StatOperator o : ep.operators.get(i).getOperators()){
						if(FelixConfig.explainMode){
							FelixUIMan.println(0,0, o.toString());
						}
					}
					
					if(FelixConfig.explainMode){
						FelixUIMan.println("");
					}
					
				}
			
			if(FelixConfig.explainMode){
				return;
			}
				
			
			int sum = -1;
			int nIt = FelixConfig.nDDIT;
			
			for(int i = ep.operators.size()-1; i >=0; i --){
				ep.operators.get(i).setNCore(Config.getNumThreads());
				ep.operators.get(i).prepareDB44DD();				
			}

			// the first run of DD replicates the NON-DD executor.
			// this is used to decides the scope of the downstream MLN
			// operator.
			FelixConfig.isFirstRunOfDD = true;
			
			for(int i = ep.operators.size()-1; i >=0; i --){
				dmoo.optimizeDMO(ep.operators.get(i));
				ep.operators.get(i).run();
				
			}			
			
			BufferedWriter bw = FileMan.getBufferedWriterMaybeGZ(options.fout + "_snap_0" );
			RDB newDB = RDB.getRDBbyConfig(Config.db_schema);
			for(FelixPredicate fp : fq.getPredicates()){
				if(fp.hasQuery()){
					UIMan.println(">>> Dumping results for " + fp + "\n");
					this.dumpMapAnswerForPredicate(newDB, fp, bw);
				}
			}
			bw.close();
			
			File from = new File(options.fout + "_snap_" + (FelixConfig.nDDIT-nIt));
			File to = new File(options.fout);
			if(to.exists()){
				to.delete();
			}
			Files.copy(from, to);
			
			FelixConfig.isFirstRunOfDD = false;
			
			// copy state of each table
			HashSet<FelixPredicate> sharedPreds = new HashSet<FelixPredicate>();
			for(int i = ep.operators.size()-1; i >=0; i --){
				sharedPreds.addAll(ep.operators.get(i).dd_CommonOutput);
				if(ep.operators.get(i).type == OPType.TUFFY){
					sharedPreds.addAll(ep.operators.get(i).outputPredicates);
				}
			}
			for(FelixPredicate fp : sharedPreds){
				UIMan.println(">>> Backing up table " + fp.getName());
				newDB.dropTable("_copy_ori_" + fp.getRelName());
				String sql = "CREATE TABLE _copy_ori_" + fp.getRelName()
						+ " AS SELECT * FROM " + fp.getRelName() + ";";
				newDB.execute(sql);
			}
			newDB.close();
			
			
			while(sum != 0 && nIt-- > 0 ){
				
				newDB = RDB.getRDBbyConfig(Config.db_schema);
				
				for(FelixPredicate fp : sharedPreds){
					UIMan.println(">>> Restoring table " + fp.getName());
					newDB.dropTable(fp.getRelName());
					newDB.dropView(fp.getRelName());
					String sql = "CREATE TABLE " + fp.getRelName()
							+ " AS SELECT * FROM _copy_ori_" + fp.getRelName() + ";";
					newDB.execute(sql);
				}
				for(FelixPredicate fp : fq.getAllPred()){
					if(fp.isCorefPredicate && !sharedPreds.contains(fp)){
						newDB.dropTable(fp.getRelName());
						newDB.dropView(fp.getRelName());
						newDB.execute(fp.viewDef);
					}					
				}
				newDB.commit();
				newDB.close();
			
				for(int i = ep.operators.size()-1; i >=0; i --){
					dmoo.optimizeDMO(ep.operators.get(i));
					ep.operators.get(i).run();
				}
				
				bw = FileMan.getBufferedWriterMaybeGZ(options.fout + "_snap_" + (FelixConfig.nDDIT-nIt) );
				newDB = RDB.getRDBbyConfig(Config.db_schema);
				for(FelixPredicate fp : fq.getPredicates()){
					if(fp.hasQuery()){
						UIMan.println(">>> Dumping results for " + fp + "\n");
						this.dumpMapAnswerForPredicate(newDB, fp, bw);
					}
				}
				newDB.close();
				bw.close();
				
				from = new File(options.fout + "_snap_" + (FelixConfig.nDDIT-nIt));
				to = new File(options.fout);
				if(to.exists()){
					to.delete();
				}
				Files.copy(from, to);
				
				
				
				sum = 0;
				
				//CURRENTLY ASSUME ONE PREDICATE IS SHARED BY ONLY TWO OPERATORS 
				// this is ensured by Felix's compiler.
				for(FelixPredicate fp : fq.getAllPred()){
					
					ConcurrentOperatorsBucket cob1 = null;
					ConcurrentOperatorsBucket cob2 = null;
					
					for(int i = ep.operators.size()-1; i >=0; i --){
						if(ep.operators.get(i).dd_CommonOutput.contains(fp)){
							if(cob1 == null){
								cob1 = ep.operators.get(i);
							}else if(cob2 == null){
								cob2 = ep.operators.get(i);
							}else{
								ExceptionMan.die("ERROR 12: " +
										"There must be something wrong with the compiler");
							}
						}
					}
					
					if(cob1==null || cob2==null){
						continue;
					}
					
					String tableName1;
					String tableName2;
					String priorTable1;
					String priorTable2;
					
					tableName1 = cob1.dd_commonOutputPredicate_2_tableName.get(fp);
					tableName2 = cob2.dd_commonOutputPredicate_2_tableName.get(fp);
					
					priorTable1 = cob1.dd_commonOutputPredicate_2_priorTableName.get(fp);
					priorTable2 = cob2.dd_commonOutputPredicate_2_priorTableName.get(fp);
					
					
					String args = StringMan.commaList(fp.getArgs());
                    ArrayList<String> whereargsarray = new ArrayList<String>();
                    for(String a : fp.getArgs()){
                            whereargsarray.add("t0." + a + "=" + "t1." + a);
                    }
                    String whereargs = StringMan.join(" AND ", whereargsarray);
                    ArrayList<String> t0argsArray = new ArrayList<String>();
                    for(String a : fp.getArgs()){
                            t0argsArray.add("t0." + a);
                    }
                    String t0args = StringMan.commaList(t0argsArray);
                    
                    ArrayList<String> t1argsArray = new ArrayList<String>();
                    for(String a : fp.getArgs()){
                            t1argsArray.add("t1." + a);
                    }
                    String t1args = StringMan.commaList(t1argsArray);

					String priorArgs = StringMan.commaList(fp.getArgs());
					priorArgs = priorArgs + "," + "float_" + (fp.arity()+1);
					priorArgs = "truth, club, " + priorArgs;
					
					if(options.marginal == true){

						String deltaCase = "(CASE WHEN abs(t0.prior-t1.prior)<0.1 THEN " + 0 +
								" ELSE 0.9*(t0.prior-t1.prior) END) AS prior11";
						String negDeltaCase = "(CASE WHEN abs(t1.prior-t0.prior)<0.1 THEN " + 0 +
								" ELSE 0.9*(t1.prior-t0.prior) END) AS prior11";
						
						
                         String sql = "(SELECT NULL::Float, TRUE, 2, " + t0args + ", " + deltaCase +" FROM " + tableName1 + " t0, " + tableName2  + " t1 " + " WHERE "
                                 + whereargs+")";
                         
                         sql = sql + " UNION ALL " + "(SELECT NULL::Float, TRUE, 2, " + t0args + ", t0.prior FROM " + tableName1 + " t0 " +
                        		 " WHERE (" + t0args + ") NOT IN (SELECT " + t1args + " FROM " + tableName2 + " t1) AND t0.prior>0.1)";

                         sql =  "INSERT INTO " + priorTable2 + "(prior, " + priorArgs + ")" + " SELECT * FROM (" + sql + ") nt WHERE nt.prior11<>0";
                         Felix.db.update(sql);
                         
                         
                         UIMan.println(">>> INSERT " + Felix.db.getLastUpdateRowCount() + " NEW TUPLES FOR " + fp + " OF " + cob2);

                         sql = "(SELECT NULL::Float, TRUE, 2, " + t0args + "," + negDeltaCase + " FROM " + tableName1 + " t0, " + tableName2  + " t1 " + " WHERE "
                         + whereargs+")";
                         
                         sql = sql + " UNION ALL " + "(SELECT NULL::Float, TRUE, 2, " + t0args + ", -t0.prior FROM " + tableName2 + " t0 " +
                        		 " WHERE (" + t0args + ") NOT IN (SELECT " + t1args + " FROM " + tableName1 + " t1) AND t0.prior>0.1)";
                         
                         sql = "INSERT INTO " + priorTable1 + "(prior, " + priorArgs + ")" + " SELECT * FROM (" + sql + ") nt WHERE nt.prior11<>0";
                         Felix.db.update(sql);
                         UIMan.println(">>> INSERT " + Felix.db.getLastUpdateRowCount() + " NEW TUPLES FOR " + fp + " OF " + cob1);

                         sum += Felix.db.getLastUpdateRowCount();

					}else{
						String sql = "SELECT NULL::Float, TRUE, 2, " + args + ", " + (0.9) + " FROM " + tableName1 + " WHERE ("
							+ args + ") NOT IN ( SELECT " + args + " FROM " + tableName2 + ")"
							+ " UNION ALL " + 
							"SELECT NULL::Float, TRUE, 2, " + args + ", " + (-0.9) + " FROM " + tableName2 + " WHERE ("
							+ args + ") NOT IN ( SELECT " + args + " FROM " + tableName1 + ")";
						sql = "INSERT INTO " + priorTable2 + "(prior, " + priorArgs + ")" + " SELECT * FROM (" + sql + ") nt";
						Felix.db.update(sql);
						UIMan.println(">>> INSERT " + Felix.db.getLastUpdateRowCount() + " NEW TUPLES FOR " + fp + " OF " + cob2);
						
						sql = "SELECT NULL::Float, TRUE, 2, " + args + ", " + (-0.9) + " FROM " + tableName1 + " WHERE ("
						+ args + ") NOT IN ( SELECT " + args + " FROM " + tableName2 + ")"
						+ " UNION ALL " + 
						"SELECT NULL::Float, TRUE, 2, " + args + ", " + (0.9) + " FROM " + tableName2 + " WHERE ("
						+ args + ") NOT IN ( SELECT " + args + " FROM " + tableName1 + ")";
						sql = "INSERT INTO " + priorTable1 + "(prior, " + priorArgs + ")" + " SELECT * FROM (" + sql + ") nt";
						Felix.db.update(sql);
						UIMan.println(">>> INSERT " + Felix.db.getLastUpdateRowCount() + " NEW TUPLES FOR " + fp + " OF " + cob1);
						
                        sum += Felix.db.getLastUpdateRowCount();
						
                     }
				}
				
				Felix.db.commit();
				
			}
			
		
		}catch(Exception e){
			e.printStackTrace();
		}
		
			
	}
	
	/**
	 * The constructor.
	 * @param _ep
	 */
	public DDExecutor(ExecutionPlan _ep, FelixQuery _fq, FelixCommandOptions _options){
		ep = _ep;
		dmoo = new DMOOptimizer(ep.getCostModel());
		fq = _fq;
		options = _options;
	}

	/**
	 * Output the results of this bucket.
	 * @param db
	 * @param fout
	 * @param p
	 */
	public void dumpMapAnswerForPredicate(RDB db, FelixPredicate p, BufferedWriter bufferedWriter) {
		//	spreadTruth();
			HashMap<Long,String> cmap = db.loadIdSymbolMapFromTable();
			try {
				
				int digits = 4;
				
				String sql;
				String tableName = p.getRelName();
				String predName = p.getName();
				
				if(options.useDualDecomposition){
					//sql = "SELECT * FROM " + tableName + " WHERE truth=TRUE ORDER BY prior DESC";
					if(!p.isCorefMapPredicate){
						if(p.belongsTo != null && p.belongsTo.dd_commonOutputPredicate_2_tableName.containsKey(p)){
							sql = "SELECT * FROM " + p.belongsTo.dd_commonOutputPredicate_2_tableName.get(p) + " WHERE truth=TRUE ORDER BY prior DESC";
						}else{
							sql = "SELECT * FROM " + p.getRelName() + " WHERE truth=TRUE ORDER BY prior DESC";
						}
					}else{
						sql = "SELECT * FROM " + tableName + " ORDER BY prior DESC";
					}
				}else{
					sql = "SELECT * FROM " + tableName + " WHERE truth=TRUE ORDER BY prior DESC";
				}
				
				ResultSet rs = db.query(sql);
				while(rs == null){
					rs = db.query(sql);
				}
				while(rs.next()) {
					String line = predName + "(";
					ArrayList<String> cs = new ArrayList<String>();
					int ct = 0;
					for(String a : p.getArgs()) {
						
						
						if(p.getTypeAt(ct).isProbArg == true || p.getTypeAt(ct).isNonSymbolicType()){
							cs.add(rs.getDouble(a)+"");
						}else{
							long c = rs.getLong(a);
						
						
							String v = StringMan.escapeJavaString(cmap.get(c));
						
						//if(v.matches("^[0-9].*$") && !StringMan.escapeJavaString(v).contains(" ")){
						//	cs.add("" + StringMan.escapeJavaString(v) + "");
						//}else{
							cs.add("\"" + StringMan.escapeJavaString(v) + "\"");
						//}
						}
						ct ++;
					}
					line += StringMan.commaList(cs) + ")";
					
					double prior = 1;
					if(options.marginal){
						
						double prob;
						if(rs.getString("prior") == null){
							prob = 1;
						}else{
							prob = Double.valueOf(rs.getString("prior"));
						}
						
						if(Config.output_prolog_format){
						
							line = "tuffyPrediction(" + UIMan.decimalRound(digits, prob) +
								", " + line + ").";
						}else{
							line = UIMan.decimalRound(digits, prob) + "\t" + line;

						}
						
					}else{
						line =  line;
					}
					
					if(prior >= options.minProb){
						bufferedWriter.append(line + "\n");
					}
					
				}
				rs.close();
				//bufferedWriter.close();
			} catch (Exception e) {
				ExceptionMan.handle(e);
			}
		}
}




