package felix.executor;

import java.io.BufferedWriter;
import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;

import tuffy.db.RDB;
import tuffy.infer.DataMover;
import tuffy.mln.MarkovLogicNetwork;
import tuffy.mln.Predicate;
import tuffy.parse.CommandOptions;
import tuffy.util.Config;
import tuffy.util.ExceptionMan;
import tuffy.util.FileMan;
import tuffy.util.StringMan;
import tuffy.util.UIMan;

import felix.dstruct.ConcurrentOperatorsBucket;
import felix.dstruct.ExecutionPlan;
import felix.dstruct.FelixPredicate;
import felix.dstruct.FelixQuery;
import felix.dstruct.StatOperator.OPType;
import felix.optimizer.DMOOptimizer;
import felix.parser.FelixCommandOptions;
import felix.util.FelixConfig;
import felix.util.FelixUIMan;



/**
 * Class for executing a given physical {@link ExecutionPlan}.
 * 
 * @deprecated
 * 
 * @author Ce Zhang
 *
 */
public class Executor {

	/**
	 * The execution plan to be executed.
	 */
	ExecutionPlan ep;
	
	/**
	 * The FelixQuery used by this Executor.
	 */
	FelixQuery fq;
	
	/**
	 * The FelixCommandOptions used by this Executor.
	 */
	FelixCommandOptions options;
	
	/**
	 * The DMOOptimizer used by this Executor.
	 */
	public DMOOptimizer dmoo;
		
	/**
	 * Execute this plan. If Felix runs in
	 * explain mode, just prints out the physical plan.
	 * 
	 * <p> TODO: Find a better way to the explain mode (e.g., a graph).
	 */
	@SuppressWarnings("unchecked")
	public void run(){
		try{
			
			if(FelixConfig.explainMode){
				FelixUIMan.println(0, 0, "\nExecution Plan:\n");
			}
			
			ArrayList<ConcurrentOperatorsBucket> toBeJoined = new ArrayList<ConcurrentOperatorsBucket>();
			
			HashSet<FelixPredicate> finishedPredicates = new HashSet<FelixPredicate>();
			
			
			for(int i = ep.operators.size()-1; i >=0; i --){
									
				if(!FelixConfig.explainMode){
					//model for bucket core allocator
					ep.operators.get(i).setNCore(Config.getNumThreads());
					ep.operators.get(i).pushPredicateScopes(finishedPredicates);
					dmoo.optimizeDMO(ep.operators.get(i));
					ep.operators.get(i).start();
					
				}
				
				HashSet<Predicate> currentOutput = (HashSet<Predicate>) 
										ep.operators.get(i).outputPredicates.clone();
				
				HashSet<Predicate> nextInput = new HashSet<Predicate>();
				if(i>0){
					nextInput = (HashSet<Predicate>) 
										ep.operators.get(i-1).inputPredicates.clone();
				}
				
				currentOutput.retainAll(nextInput);
				
				//TODO:
				if(currentOutput.size() != 0 || i == 0 || toBeJoined.size() == FelixConfig.nCores - 1){
					toBeJoined.add(ep.operators.get(i));
					
					if(FelixConfig.explainMode){
						FelixUIMan.println("Concurrently:");
					}
					
					for(ConcurrentOperatorsBucket o : toBeJoined){
						if(FelixConfig.explainMode){
							FelixUIMan.println(0,0, o.toString());
						}else{
							o.myJoin();
							finishedPredicates.addAll(o.outputPredicates);
						}
					}
					
					toBeJoined.clear();

				}else{
					toBeJoined.add(ep.operators.get(i));
				}
				
			}
			
			
			if(!FelixConfig.explainMode){
				BufferedWriter bw = FileMan.getBufferedWriterMaybeGZ(options.fout);
				RDB newDB = RDB.getRDBbyConfig(Config.db_schema);
				for(FelixPredicate fp : fq.getPredicates()){
					
					if(fp.hasQuery()){
						
						UIMan.println(">>> Dumping results for " + fp + "\n");
						this.dumpMapAnswerForPredicate(newDB, fp, bw);
						
					}
				}
				newDB.close();
				bw.close();
			}
		
		}catch(Exception e){
			e.printStackTrace();
		}
			
	}
	
	/**
	 * The constructor.
	 * @param _ep
	 */
	public Executor(ExecutionPlan _ep, FelixQuery _fq, FelixCommandOptions _options){
		ep = _ep;
		dmoo = new DMOOptimizer(ep.getCostModel());
		fq = _fq;
		options = _options;
	}

	/**
	 * Output the results of this bucket.
	 * @param db
	 * @param fout
	 * @param p
	 */
	public void dumpMapAnswerForPredicate(RDB db, FelixPredicate p, BufferedWriter bufferedWriter) {
		
		//	spreadTruth();
			HashMap<Long,String> cmap = db.loadIdSymbolMapFromTable();
			try {
				
				int digits = 4;
				
				String sql;
				String tableName = p.getRelName();
				String predName = p.getName();
				
				sql = "SELECT * FROM " + tableName + " WHERE truth=TRUE ORDER BY prior DESC";
				
				ResultSet rs = db.query(sql);
				while(rs == null){
					rs = db.query(sql);
				}
				while(rs.next()) {
					String line = predName + "(";
					ArrayList<String> cs = new ArrayList<String>();
					int ct = 0;
					for(String a : p.getArgs()) {
						
						
						if(p.getTypeAt(ct).isProbArg == true || p.getTypeAt(ct).isNonSymbolicType()){
							cs.add(rs.getDouble(a)+"");
						}else{
							long c = rs.getLong(a);
						
						
							String v = StringMan.escapeJavaString(cmap.get(c));
						
						//if(v.matches("^[0-9].*$") && !StringMan.escapeJavaString(v).contains(" ")){
						//	cs.add("" + StringMan.escapeJavaString(v) + "");
						//}else{
							cs.add("\"" + StringMan.escapeJavaString(v) + "\"");
						//}
						}
						ct ++;
					}
					line += StringMan.commaList(cs) + ")";
					
					double prior = 1;
					if(options.marginal){
						
						double prob;
						if(rs.getString("prior") == null){
							prob = 1;
						}else{
							prob = Double.valueOf(rs.getString("prior"));
						}
						
						if(Config.output_prolog_format){
						
							line = "tuffyPrediction(" + UIMan.decimalRound(digits, prob) +
								", " + line + ").";
						}else{
							line = UIMan.decimalRound(digits, prob) + "\t" + line;

						}
						
					}else{
						line =  line;
					}
					
					if(prior >= options.minProb){
						bufferedWriter.append(line + "\n");
					}
					
				}
				rs.close();
				//bufferedWriter.close();
			} catch (Exception e) {
				ExceptionMan.handle(e);
			}
		}
}




