package felix.io;

import java.io.BufferedWriter;
import java.io.OutputStreamWriter;
import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FSDataInputStream;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.postgresql.PGConnection;

import felix.util.FelixConfig;
import tuffy.db.RDB;
import tuffy.util.StringMan;
import tuffy.util.UIMan;

/**
 * Class for populating tables from 
 * files/directories generated with hadoop.
 * @author Ce Zhang
 *
 */
public class HadoopPostgreSQLPopulator{
	
	/**
	 * Dumps given schema.table in to the given file.
	 * @param schema
	 * @param tableName
	 * @param nColumn
	 * @param filePath
	 */
	public static void dumpTableToHDFS(
										String           schema,
										String           tableName,
										int              nColumn,
										String			 filePath){
		try{
			Path pt=new Path(filePath);
			Configuration conf = new Configuration();
			
			conf.set("fs.default.name", FelixConfig.hdfsServer);
			conf.set("fs.defaultFS", FelixConfig.hdfsServer);
			
			FileSystem fs = FileSystem.get(conf);
			BufferedWriter br=new BufferedWriter(new OutputStreamWriter(fs.create(pt,true)));
			
			RDB db = RDB.getRDBbyConfig(schema);
			db.disableAutoCommitForNow();
			
			ResultSet rs = db.query("SELECT * FROM " + tableName);
			while(rs.next()){
				ArrayList<String> tmps = new ArrayList<String>();
				for(int i=0;i<nColumn;i++){
					tmps.add(rs.getString(i+1).replaceAll("\t", "\\\\t").replaceAll("\n", "\\\\n").replaceAll("\r", "\\\\r"));
				}
				br.write(StringMan.join("\t", tmps) + "\n");
			}
			
			db.close();
			br.close();
		}catch(Exception e){
			e.printStackTrace();
		}
		
		
	}
	
	
	
	/**
	 * Creates and populates the given schema.table from the given file.
	 * @param schema
	 * @param tableName
	 * @param columnNames
	 * @param columnTypes
	 * @param filePath
	 * @throws Exception
	 */
	public static void createAndPopulateTable(	
										String            schema,
										String            tableName,
										List<String> 	  columnNames,
										List<String> 	  columnTypes,
										String		      filePath      ) throws Exception{
		
		RDB db;
		if(schema == null){
			db = RDB.getRDBbyConfig();
		}else{
			db = RDB.getRDBbyConfig(schema);
		}
		db.dropTable(tableName);
		
		ArrayList<String> schemas = new ArrayList<String>();
		for(int i=0;i<columnNames.size();i++){
			schemas.add(columnNames.get(i) + " " + columnTypes.get(i));
		}
		
		db.dropTable(tableName);
		
		String sql = "CREATE TABLE " + tableName + " ( truth BOOL, prior FLOAT, " + StringMan.commaList(schemas) + " )";
		db.execute(sql);
		db.close();
		
		HadoopPostgreSQLPopulator.populateTable(schema, tableName, columnNames, columnTypes, filePath);
		
		
	}
	
	//TODO: CHECK HDFS API AND MOVE TO MORE ELEGANT WAY
	/**
	 * Populates the given schema.table from the given directory.
	 * @param schema
	 * @param tableName
	 * @param columnNames
	 * @param columnTypes
	 * @param dirPath
	 * @param nReducer
	 * @throws Exception
	 */
	/*
	public static void populateTableFromDir(	
									String            schema,
									String            tableName,
									List<String> 	  columnNames,
									List<String> 	  columnTypes,
									String		      dirPath,
									int 			  nReducer) throws Exception{
		
		for(int i=0;i<nReducer;i++){
			String fileName = "" + nReducer;
			while(fileName.length() < 5){
				fileName = "0" + fileName;
			}
			fileName = "part-" + fileName;
			HadoopPostgreSQLPopulator.populateTable(schema, tableName, columnNames, columnTypes, fileName);
		}
	}
	*/
	
	//TODO: CHECK HDFS API AND MOVE TO MORE ELEGANT WAY
	/**
	 * Creates and populates the given schema.table from the given directory.
	 * @param schema
	 * @param tableName
	 * @param columnNames
	 * @param columnTypes
	 * @param dirPath
	 * @param nReducer
	 * @throws Exception
	 */
	public static void createAndPopulateTableFromDir(	
									String            schema,
									String            tableName,
									List<String> 	  columnNames,
									List<String> 	  columnTypes,
									String		      dirPath,
									int 			  nReducer) throws Exception{

		for(int i=0;i<nReducer;i++){
			String fileName = "" + i;
			while(fileName.length() < 5){
				fileName = "0" + fileName;
			}
			fileName = dirPath + "/" + "part-r-" + fileName;
			if(i==0){
				HadoopPostgreSQLPopulator.createAndPopulateTable(schema, tableName, columnNames, columnTypes, fileName);
			}else{
				HadoopPostgreSQLPopulator.populateTable(schema, tableName, columnNames, columnTypes, fileName);
			}
		}
	}
	
	/**
	 * Populates the given schema.table from the given file.
	 * @param schema
	 * @param tableName
	 * @param columnNames
	 * @param columnTypes
	 * @param filePath
	 * @throws Exception
	 */
	public static void populateTable(	
										String            schema,
										String            tableName,
										List<String> 	  columnNames,
										List<String> 	  columnTypes,
										String		      filePath      ) throws Exception{

		RDB db;
		if(schema == null){
			db = RDB.getRDBbyConfig();
		}else{
			db = RDB.getRDBbyConfig(schema);
		}
		
		ArrayList<String> schemas = new ArrayList<String>();
		for(int i=0;i<columnNames.size();i++){
			schemas.add(columnNames.get(i) + " " + columnTypes.get(i));
		}
		String sql;
		
		Configuration conf = new Configuration();
		//TOOD: HARD-WIRED ADDRESS IS NTO GOOD
		conf.set("fs.default.name", FelixConfig.hdfsServer);
		conf.set("fs.defaultFS", FelixConfig.hdfsServer);
		
        FileSystem fileSystem = FileSystem.get(conf);

        Path path = new Path(filePath);
        if (!fileSystem.exists(path)) {
            UIMan.verbose(2, "File " + filePath + " does not exists");
            return;
        }

        FSDataInputStream in = fileSystem.open(path);
		PGConnection con = (PGConnection)db.getConnection();
		sql = "COPY " + tableName + "( truth , prior , " + StringMan.commaList(columnNames) + ") " +
				" FROM STDIN WITH DELIMITER '\t'";
		con.getCopyAPI().copyIn(sql, in);
		
		in.close();
        fileSystem.close();
		
        db.commit();
        db.close();
	
	
	}
	
	/**
	 * Main test method.
	 * @param args
	 * @throws Exception
	 */
	/*
	public static void main(String[] args) throws Exception{
	
		
		//HadoopPostgreSQLPopulator.populateTable(
		//						null, 
		//						"test", 
		//						Arrays.asList("c1", "c2"), 
		//						Arrays.asList("TEXT", "TEXT"),
		//						"hdfs://d-02.cs.wisc.edu:9000/firstMapReduceOut38/part-00000");
		
		
		HadoopPostgreSQLPopulator.createAndPopulateTable(
					null, 
					"test", 
					Arrays.asList("c1", "c2"), 
					Arrays.asList("TEXT", "TEXT"),
					"hdfs://d-02.cs.wisc.edu:9000/firstMapReduceOut42/part-00000");
	}
	*/
	
}
