package felix.main;

import java.io.File;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;



import tuffy.db.RDB;
import tuffy.db.SQLMan;
import tuffy.mln.Clause;
import tuffy.mln.Predicate;
import tuffy.mln.Type;
import tuffy.ra.ConjunctiveQuery;
import tuffy.util.Config;
import tuffy.util.ExceptionMan;
import tuffy.util.FileMan;
import tuffy.util.StringMan;
import tuffy.util.Timer;
import tuffy.util.UIMan;
import felix.compiler.StaticAnalyzer;
import felix.dstruct.ExecutionPlan;
import felix.dstruct.FelixPredicate;
import felix.dstruct.FelixQuery;
import felix.executor.DDExecutor;
import felix.executor.Executor;
import felix.io.HadoopPostgreSQLPopulator;
import felix.io.TestHadoop;
import felix.optimizer.Scheduler;
import felix.parser.FelixCommandOptions;
import felix.util.FelixConfig;
import felix.util.FelixUIMan;

/**
 * Felix, an operator-based statistical inference system.
 * @author Ce Zhang
 *
 */
public class Felix {

	/**
	 * Felix's query, which consists of program, query and evidence.
	 */
	public FelixQuery fq;

	/**
	 * Static Analyzer, which analyzes the property of each predicates according to rules.
	 */
	protected StaticAnalyzer sa;

	/**
	 * Database connection.
	 */
	public static RDB db;

	/**
	 * Command line option.
	 */
	public FelixCommandOptions options;

	/**
	 * Whether this Felix object run has loaded evidences.
	 */
	public boolean hasLoadedEvidence = false;

	/**
	 * Return Felix's query.
	 * @return
	 */
	public FelixQuery getFelixQuery(){
		return fq;
	}

	/**
	 * Clear static parameters and initialize variables.
	 */
	public void resetACoupleAuxDataStructures(){

		ConjunctiveQuery.clearIndexHistory();

		Clause.mappingFromID2Const = new HashMap<Integer, String>();
		Clause.mappingFromID2Desc = new HashMap<String, String>();

	}

	/**
	 * Load Felix query from program and query file. If -useEvid
	 * is used, evidence file will also be loaded by this function.
	 * @throws InterruptedException 
	 */
	public FelixQuery parseFelixQuery() throws InterruptedException{

		FelixUIMan.println(">>> Connecting to RDBMS at " + FelixConfig.db_url);

		if(FelixConfig.evidDBSchema != null){
			FelixConfig.db_schema = FelixConfig.evidDBSchema;
		}

		db = RDB.getRDBbyConfig();
		db.resetSchema(FelixConfig.db_schema);
		db.schema = FelixConfig.db_schema;
		
		FelixQuery fq = new FelixQuery();

		String[] progFiles = options.fprog.split(",");
		fq.loadPrograms(progFiles);

		if(options.fquery != null){
			String[] queryFiles = options.fquery.split(",");
			fq.loadQueries(queryFiles);
		}

		if(options.queryAtoms != null){
			FelixUIMan.println(">>> Parsing query atoms in command line");
			fq.parseQueryCommaList(options.queryAtoms);
		}

		if(options.cwaPreds != null){
			String[] preds = options.cwaPreds.split(",");
			for(String ps : preds){
				Predicate p = fq.getPredByName(ps);
				if(p == null){
					fq.closeFiles();
					ExceptionMan.die("COMMAND LINE: Unknown predicate name -- " + ps);
				}else{
					p.setClosedWorld(true);
				}
			}
		}

		for(Predicate p : fq.getAllPred()){
			p.prepareDB(db);
		}

		// whether we use evidence for static analyzer.
		// if we do not load evidence here, it will be
		// loaded in the Scheduler.schedule(). 
		if(options.useEvid == true){
			this.loadEvidence();
		}
		return fq;
	}

	/**
	 * Normalizes string by replacing whitespace
	 * @param script
	 * @return normalized string
	 */
	public String normalizeScript(String script){
		String rs = "";
		rs = script.replaceAll("^\t", "");
		rs = rs.replaceAll("^(\\t|\\s)*<~hdfs~(\\t|\\s)*", "");
		rs = rs.replaceAll("\n\t", "\n");
		return rs;
	}
	
	/**
	 * Returns array list of TEXT
	 * @param n size of ArrayList
	 * @return ArrayList
	 */
	public ArrayList<String> getAllTextArray(int n){
		ArrayList<String> rs = new ArrayList<String>();
		for(int i=0;i<n;i++){
			rs.add("TEXT");
		}
		return rs;
	}
	
	/**
	 * Returns table name for given signature.
	 * @param metaTable
	 * @param signature
	 * @return
	 */
	public String giveMeTableNameIfExist(String metaTable, String signature){
		String ret = null;
		
		RDB db = RDB.getRDBbyConfig(FelixConfig.auxSchema);
		
		ResultSet rs = db.query("SELECT * FROM " + metaTable + " WHERE signature='" + signature + "'");
		
		try {
			while(rs.next()){
				ret = rs.getString(2);
				return ret;
			}
		} catch (SQLException e) {
			e.printStackTrace();
		}
		
		db.close();
		return ret;
	}
	
	/**
	 * @deprecated
	 * @param p
	 * @return
	 */
	public String generateFESignature(FelixPredicate p){
		
		
		System.err.println(p.mapinputvar);
		System.err.println(p.reduceinputkeyvar);
		System.err.println(p.reduceinputvaluesvar);
		
		System.err.println(p.dependencyFile);
		System.err.println(p.mapScript);
		System.err.println(p.reduceScript);
		System.err.println(p.mapinitScript);
		System.err.println(p.xmltag);
		

		
		String ret = "";
		ret += p + "|||||" + StringMan.commaList(p.getArgs()) + "|||||";
		ret += p.mapinputvar + "|||||";
		ret += p.mapinitScript + "|||||";
		ret += p.reduceinputkeyvar + "|||||";
		ret += p.reduceinputvaluesvar + "|||||";
		ret += p.dependencyFile + "|||||";
		ret += p.mapScript + "|||||";
		ret += p.reduceScript + "|||||";
		ret += p.xmltag + "|||||";
		
		ret = ret.replaceAll("\\s|\\t|\n|\r|\\n|\\r|'", "\"");
		
		return ret;
	}
	
	/**
	 * Executor for the Blah Blah feature extraction language.
	 */
	public void blahblahExecutor(){

		// TODO: current assumption is not cross-dependency, i.e.,
		// one jdbc predicate only relies on one hadoop predicates
		ArrayList<FelixPredicate> trivialPlanner = new ArrayList<FelixPredicate>();
		for(FelixPredicate p : fq.getAllPred()){
			if(p.needExtractFeatures){
				if(p.dependencyFile != null){
					trivialPlanner.add(0, p);
				}else if(p.jdbcdep != null){
					trivialPlanner.add(trivialPlanner.size(), p);
				}
			}
		}
		
		// first extract features
		for(FelixPredicate p : trivialPlanner){
			
			if(p.needExtractFeatures){
				
				if(FelixConfig.auxSchema == null){
					ExceptionMan.die("You must provide a schema to save result files from hadoop using -auxSchema option!");
				}
				
				String signature = this.generateFESignature(p);
				//String isInTable = this.giveMeTableNameIfExist(metaTable, signature);
				
				String rsTableName = "_fe_raw_" + p.getName();
				
				//disable re-use
				FelixConfig.forceExt = true;
				String isInTable = null;
				
				if(isInTable != null && !FelixConfig.forceExt){
					UIMan.println(">>> The feature extraction code for " + p + 
							" is cached and does not need to be extracted agian!");
					rsTableName = isInTable;
					
				}else{
					
					RDB db = RDB.getRDBbyConfig();
					db.dropTable(FelixConfig.auxSchema + "." + rsTableName);
					db.close();
					
					//db.execute("SET search_path = " + FelixConfig.auxSchema);
					//db.execute("DELETE FROM " + metaTable + " WHERE tableName='" + rsTableName + "'");
					//db.commit();
					//db.close();
					
					UIMan.println(">>> Extracting Features for predicate " + p + " using " + p.dependencyFile);
				
					String tmpFileName = "rsFileOnHadoop_" + 
									(new Date()).getTime() + 
									"_pred_" + p.getName();
					
					if(p.jdbcdep != null){
					
						String relyOn = normalizeScript(p.jdbcdep).replace("jdbc://", "");
						Pattern pp = Pattern.compile("(.*?)\\((.*?)\\)");
						Matcher m = pp.matcher(relyOn);
						m.find();
						String pred = m.group(1);
						String[] vairableName = m.group(2).split(",");
							
						String tableName = "_fe_raw_" + pred;
						String fileName = FelixConfig.hdfsServer + tmpFileName + "_ori_" + pred + "_for_" + p.getName();
						
						UIMan.println(">>> Dumping database table " + tableName + " to HDFS...");
						HadoopPostgreSQLPopulator.dumpTableToHDFS(FelixConfig.auxSchema, 
								tableName, 
								vairableName.length, 
								fileName);
						
						String[] toPass = {
								
								"standard",
								
								normalizeScript(fileName),
								
								FelixConfig.hdfsServer + tmpFileName,
								
								normalizeScript(p.mapScript),
								
								normalizeScript(p.reduceScript),
								
								p.mapinputvar,
								
								p.reduceinputkeyvar,
								
								p.reduceinputvaluesvar,
								
								normalizeScript(p.mapinitScript),
								
								normalizeScript(p.reduceinitScript)
								
						};
						
						try {
							TestHadoop.executeHadoopProgram(toPass);
							//TestHadoop.post(toPass);
						} catch (Exception e) {
							e.printStackTrace();
							ExceptionMan.die("Hadoop Error!");
						}
						
					}else if(p.xmltag != null){
						
						String[] toPass = {
								
								"xml",
								
								normalizeScript(p.dependencyFile),
								
								FelixConfig.hdfsServer + tmpFileName,
								
								normalizeScript(p.mapScript),
								
								normalizeScript(p.reduceScript),
								
								p.mapinputvar,
								
								p.reduceinputkeyvar,
								
								p.reduceinputvaluesvar,
								
								normalizeScript(p.mapinitScript),
								
								normalizeScript(p.reduceinitScript),
								
								"<" + p.xmltag + "",
								
								"</" + p.xmltag + ">"
								
						};
						
						try {
							TestHadoop.executeHadoopProgram(toPass);
							//TestHadoop.post(toPass);
						} catch (Exception e) {
							e.printStackTrace();
							ExceptionMan.die("Hadoop Error!");
						}
						
					}else{
						
						String[] toPass = {
								
								"standard",
								
								normalizeScript(p.dependencyFile),
								
								FelixConfig.hdfsServer + tmpFileName,
								
								normalizeScript(p.mapScript),
								
								normalizeScript(p.reduceScript),
								
								p.mapinputvar,
								
								p.reduceinputkeyvar,
								
								p.reduceinputvaluesvar,
								
								normalizeScript(p.mapinitScript),

								normalizeScript(p.reduceinitScript)
								
						};
						
						try {
							TestHadoop.executeHadoopProgram(toPass);
							//TestHadoop.post(toPass);
						} catch (Exception e) {
							e.printStackTrace();
							ExceptionMan.die("Hadoop Error!");
						}
						
					}
					
					try {
						
						ArrayList<String> tableArgs = (ArrayList<String>) p.getArgs().clone();
						ArrayList<String> tableTypes = getAllTextArray(p.arity());
						
						if(p.getArgs().size() == 1){
							tableArgs.add("I_am_just_a_sad_meaningless_value");
							tableTypes.add("TEXT");
						}
						
						UIMan.verbose(0, ">>> Dump to database table for relation " + p.toString());
						HadoopPostgreSQLPopulator.createAndPopulateTableFromDir
							(FelixConfig.auxSchema, rsTableName, 
									tableArgs, tableTypes, 
									FelixConfig.hdfsServer + tmpFileName + "_dir", FelixConfig.nReduce);
					} catch (Exception e) {
						e.printStackTrace();
					}
				}
				
			}
		}
		
	}
	
	/**
	 * Load evidence from evidence from file or database table.
	 * @throws InterruptedException 
	 */
	public void loadEvidence() throws InterruptedException{
		
		FelixConfig.mixturedLoading = false;
		
		boolean isOnlyFromFile = true;
		
		for(FelixPredicate fp : fq.getAllPred()){
			if(fp.needExtractFeatures || fp.loadFromDatabase){
				isOnlyFromFile = false;
				FelixConfig.mixturedLoading = true;
			}	
		}
		
		if(options.fevid != null && isOnlyFromFile){
			String[] evidFiles = options.fevid.split(",");
			fq.loadEvidences(evidFiles);
			fq.materializeTables(db);
		}else{
			
			// first, generate pure-string form tables
			FelixConfig.mixturedLoading = true;
			
			HashMap<FelixPredicate, String> strTableName = 
					new HashMap<FelixPredicate, String>();
			
				// first, load from evid file
			if(options.fevid != null){
				String[] evidFiles = options.fevid.split(",");
				fq.loadEvidences(evidFiles);
			}
			for(FelixPredicate fp : fq.getAllPred()){
				if(!fp.needExtractFeatures && !fp.loadFromDatabase){
					fp.flushStrEvidence("_tmp_str_loading_" + fp.getName());
					strTableName.put(fp, "_tmp_str_loading_" + fp.getName());
				}	
			}
			
				// second, from db table
			for(FelixPredicate fp : fq.getAllPred()){
				if(fp.loadFromDatabase){
					String rName = fp.loadingSchema + "." + fp.loadingTable;
					strTableName.put(fp, rName);
				}	
 			}
			
				// third, blahblah program
			this.blahblahExecutor();
			for(FelixPredicate fp : fq.getAllPred()){
				if(fp.needExtractFeatures){
					strTableName.put(fp, FelixConfig.auxSchema + "." + "_fe_raw_" + fp.getName());
				}
			}
			
			RDB db = RDB.getRDBbyConfig(Config.db_schema);
			db.dropSequence("_tmp_seq_id");
			db.execute("CREATE SEQUENCE _tmp_seq_id START " + (fq.mapIDConstant.size() + 1) + ";");
			//second, build constant table
			HashMap<String, ArrayList<String>> typeDomain = new HashMap<String, ArrayList<String>>();
				// first, flush constants in program
			for(FelixPredicate fp : fq.getAllPred()){
				for(int i=0;i<fp.arity();i++){
					Type type = fp.getTypeAt(i);
					String tName = type.name;
					if(type.isNonSymbolicType()){
						continue;
					}
					if(!typeDomain.containsKey(tName)){
						typeDomain.put(tName, new ArrayList<String>());
						
						String sql = "CREATE TABLE type_" + tName + " (constantid bigint, constantvalue TEXT); ";
						db.execute(sql);
						db.commit();
						
						for(Integer constant : type.getDomain()){
							
							sql = "INSERT INTO type_" + tName + " VALUES (" 
									+ constant + ", " + SQLMan.escapeString(fq.mapIDConstant.get(constant)) 
									+ ")"; 
							db.execute(sql);		
							
						}
						
						
					}
				}
			}
			
			
			for(FelixPredicate fp : fq.getAllPred()){
				
				if(!strTableName.containsKey(fp)){
					continue;
				}
				
				for(int i=0;i<fp.arity();i++){
					Type type = fp.getTypeAt(i);
					if(type.isNonSymbolicType()){
						continue;
					}
					String tName = type.name;
					if(!typeDomain.containsKey(tName)){
						typeDomain.put(tName, new ArrayList<String>());
					}
					typeDomain.get(tName).add("SELECT " + fp.getArgs().get(i) 
							+ " constantVALUE FROM " + strTableName.get(fp));
				}
 			}
			

			ArrayList<String> allTypeSQLs = new ArrayList<String>();
			for(String tName : typeDomain.keySet()){
				if(typeDomain.get(tName).size() == 0){
					continue;
				}
				String sql = "INSERT INTO type_" + tName + " " + 
						"SELECT nextval('_tmp_seq_id') constantID, constantVALUE FROM " + 
						"(SELECT DISTINCT constantVALUE FROM " + 
						"(" + StringMan.join(" UNION ALL ", typeDomain.get(tName)) + " ) nt " +
						" WHERE constantVALUE NOT IN (SELECT constantvalue FROM type_" + tName + ")" +
						") wt";	
				allTypeSQLs.add("SELECT * FROM type_" + tName);
				UIMan.print(">>> Create table for type " + tName + "...");
				db.execute(sql);
				
				UIMan.print("*");
				sql = "CREATE INDEX _idx_type_id" + tName + " on type_" + tName + " (constantID) ";
				db.dropIndex("_idx_type_id" + tName);
				db.execute(sql);
				db.analyze("type_" + tName);
				UIMan.println("");
			}
			
			String crel = Config.relConstants;
			db.dropView(crel);
			db.dropTable(crel);
			String sql = "CREATE VIEW " + crel +
					" AS SELECT constantID::bigint id, constantVALUE string FROM (" + 
					StringMan.join(" UNION ALL ", allTypeSQLs) + " ) nt";
			db.execute(sql);
			
			//third, flush to Predicate table.
			for(FelixPredicate fp : fq.getAllPred()){
				String rawTable = strTableName.get(fp);
				String currentTable = strTableName.get(fp);
				
				UIMan.print(">>> Create ID table for " + fp);
				
				// first join the string table with constant map
				for(int i=0 ; i<fp.arity(); i++){
					Type type = fp.getTypeAt(i);
					
					if(type.isNonSymbolicType()){
						continue;
					}
					
					String typeTable = "type_" + type.name;
					
					String column = fp.getArgs().get(i);
					
					String newTable = rawTable + "_ser_" + column;
					
					ArrayList<String> selList = new ArrayList<String>();
					selList.add("t0.truth");
					selList.add("t0.prior");
					for(int j=0 ; j<fp.arity(); j++){
						if(j==i){
							selList.add("t1.constantid::bigint " + fp.getArgs().get(j));
						}else{
							selList.add("t0." + fp.getArgs().get(j));
						}
					}
					
					db.dropTable(newTable);
					db.dropView(newTable);
					sql = "CREATE TABLE " + newTable + " AS SELECT " + StringMan.commaList(selList) + " FROM " + 
							currentTable + " t0, " + typeTable + " t1" + " WHERE " +
							"t0." + column + "=" + "t1.constantvalue";
					
					UIMan.print(".");
					db.execute(sql);
					currentTable = newTable;
				}
				
				
				
				// second, copy to pred_* table
				ArrayList<String> insertList = new ArrayList<String>();
				ArrayList<String> selList = new ArrayList<String>();
				
				//insertList.add("truth");
				//insertList.add("prior");
				//insertList.add("club");
				
				db.dropSequence("_tmp_seq_id_" + fp.getName());
				db.execute("CREATE SEQUENCE _tmp_seq_id_" + fp.getName() + " START 1;");
				
				//selList.add("nextval('_tmp_seq_id_" + fp.getName()+"')");
				//selList.add("NULL::bigint");
				selList.add("truth::Bool");
				selList.add("prior::Float");
				selList.add("2::INT");
				selList.add("NULL::INT");
				for(int i=0 ; i<fp.arity(); i++){
					
					Type type = fp.getTypeAt(i);
					if(type.isNonSymbolicType()){
						insertList.add(fp.getArgs().get(i) + "");
						selList.add(fp.getArgs().get(i) + "::" + type.getNonSymbolicTypeInSQL());
					}else{
						insertList.add(fp.getArgs().get(i) + "");
						selList.add(fp.getArgs().get(i) + "::bigint");
					}
				}
				
				/*
				sql = "INSERT INTO " + fp.getRelName() + 
						" ( " + StringMan.commaList(insertList) + " ) " + 
						" SELECT " + StringMan.commaList(selList) + " FROM " + currentTable; 
				*/
				
				
				//sql += "id SERIAL PRIMARY KEY,\n";
				//sql += "truth BOOL,\n";
				//sql += "prior FLOAT,\n";
				//sql += "club INT DEFAULT 0,\n";
				//sql += "atomID INT DEFAULT NULL,\n";
				
				//db.dropTable(fp.getRelName());
				
				sql = "INSERT INTO " + fp.getRelName() + "(truth, " +
					"prior, club, atomID, " +  
					StringMan.commaList(insertList) + " ) " + 
					" SELECT " + StringMan.commaList(selList) 
					+ " FROM " + currentTable;

				
				UIMan.print("*");
				db.execute(sql);
				db.commit();
				
				UIMan.println("");
			}
			
			db.close();
			
		}
		
		fq.executeAllDatalogRules(db);
		
		hasLoadedEvidence = true;

	}

	/**
	 * Clean up temporary data: the schema in PostgreSQL and the working directory.
	 */
	protected void cleanUp(){		
		Config.exiting_mode = true;
		UIMan.println(">>> Cleaning up temporary data");
		if(!Config.keep_db_data){
			UIMan.print("    Removing database schema '" + Config.db_schema + "'...");
			UIMan.println(db.dropSchema(Config.db_schema)?"OK" : "FAILED");
		}else{
			UIMan.println("    Data remains in schema '" + Config.db_schema + "'.");
		}
		db.close();

		UIMan.print("    Removing temporary dir '" + Config.getWorkingDir() + "'...");
		UIMan.println(FileMan.removeDirectory(new File(Config.getWorkingDir()))?"OK" : "FAILED");

		UIMan.println("*** Felix exited at " + Timer.getDateTime() + " after running for " + Timer.elapsed());
		UIMan.closeDribbleFile();
		
		
		for(RDB db : RDB.historyInstances){
			db.close();
		}
		
		System.exit(0);
				
	}
	
	/**
	 * Run Felix!
	 * @param args Command line options.
	 * @throws InterruptedException 
	 */
	public void run(FelixCommandOptions opt) throws InterruptedException{

		Timer.start("Felix-Timer");
		this.options = opt;
		resetACoupleAuxDataStructures();

		FelixConfig.allRuleAsMLN = this.options.allRuleAsMLN;
		FelixConfig.allView = this.options.allView;
		FelixConfig.allMat = this.options.allMat;
		FelixConfig.hadoopLocal = this.options.local;
		FelixConfig.auxSchema = this.options.auxSchema;
		FelixConfig.forceExt = this.options.forceExt;
		FelixConfig.nReduce = this.options.nReduce;
		
		FelixConfig.hdfsServer = this.options.hdfs;
		FelixConfig.mrServer = this.options.mapreduce;
		
		FelixConfig.nDDIT = this.options.nDD;
		
		fq = this.parseFelixQuery();

		sa = new StaticAnalyzer(this.fq, options);
		sa.parse();

		Scheduler sc = new Scheduler(this, this.fq, options);
		ExecutionPlan ep = sc.schedule();

		if(options.useDualDecomposition){
			DDExecutor ec = new DDExecutor(ep, fq, options);
			ec.run();
		}else{
			Executor ec = new Executor(ep, fq, options);
			ec.run();
		}
		
		cleanUp();
	}

}




