package felix.operator;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;

import org.postgresql.PGConnection;


import tuffy.db.RDB;
import tuffy.mln.Predicate;
import tuffy.mln.Type;
import tuffy.ra.ConjunctiveQuery;
import tuffy.ra.Expression;
import tuffy.ra.Function;
import tuffy.util.Config;
import tuffy.util.ExceptionMan;
import tuffy.util.StringMan;
import tuffy.util.Timer;
import tuffy.util.UIMan;
import tuffy.util.UnionFind;
import felix.dstruct.DataMovementOperator;
import felix.dstruct.FelixPredicate;
import felix.dstruct.FelixQuery;
import felix.dstruct.StatOperator;
import felix.dstruct.FelixPredicate.FPProperty;
import felix.parser.FelixCommandOptions;
import felix.util.FelixConfig;
import felix.util.FelixUIMan;

/**
 * A COREF operator in Felix.
 * @author Ce Zhang
 *
 */
public class COREFOperator extends StatOperator{
	
	/**
	 * DMOs for soft positive edges.
	 */
	ArrayList<DataMovementOperator> softPosDMOs = new ArrayList<DataMovementOperator>();
	
	/**
	 * DMOs for soft negative edges.
	 */
	ArrayList<DataMovementOperator> softNegDMOs = new ArrayList<DataMovementOperator>();
	
	/**
	 * DMOs for hard positive edges.
	 */
	ArrayList<DataMovementOperator> hardPosDMOs = new ArrayList<DataMovementOperator>();
	
	/**
	 * DMOs for hard negative edges.
	 */
	ArrayList<DataMovementOperator> hardNegDMOs = new ArrayList<DataMovementOperator>();
	
	/**
	 * DMO for efficient representation of hard negative rules.
	 */
	DataMovementOperator nodeClassRule = null;

	/**
	 * DMO for efficient representation of hard negative rules.
	 */
	DataMovementOperator classTagsRule = null;
	
	/**
	 * DMO for retrieving the node's domain.
	 */
	DataMovementOperator nodeListDMO = null;
	
	/**
	 * The DataMovementOperator which is the union of all hard-neg DataMovementOperators.
	 */
	DataMovementOperator hardNegDMO1 = null;
	
	/**
	 * The DataMovementOperator which is the union of all hard-neg DataMovementOperators.
	 */
	DataMovementOperator hardNegDMO2 = null;
	
	/**
	 * The DataMovementOperator which is the union of all hard-pos DataMovementOperators.
	 */
	DataMovementOperator hardPosDMO = null;
	
	/**
	 * The DataMovementOperator which is the union of all soft-pos DataMovementOperators.
	 */
	DataMovementOperator softPosDMO1 = null;
	
	/**
	 * The DataMovementOperator which is the union of all soft-pos DataMovementOperators.
	 */
	DataMovementOperator softPosDMO2 = null;
	
	/**
	 * Whether represent clusterings results using pairwise representation.
	 * Note that, setting this parameter to true will cause quadratic numbers of
	 * result tuples.
	 */
	public boolean usePairwiseRepresentation = false;
	
	/**
	 * Target predicate of this Coref operator.
	 */
	FelixPredicate corefHead;
	
	/**
	 * Special syntax sugar for coref.
	 */
	HashMap<Integer, Integer> nodeClass = null;
	
	/**
	 * Special syntax sugar for coref.
	 */
	HashMap<Integer, HashSet<Integer>> classTags = null;
	
	/**
	 * The constructor of COREFOperator.
	 * @param _fq Felix query.
	 * @param _goalPredicates target predicates of this coref operator.
	 * @param _opt Command line options of this Felix run.
	 */
	public COREFOperator(FelixQuery _fq, HashSet<FelixPredicate> _goalPredicates,
			FelixCommandOptions _opt) {
		super(_fq, _goalPredicates, _opt);
		for(FelixPredicate p : _goalPredicates){
			p.isCorefPredicate = true;
		}
		this.type = OPType.COREF;
		this.precedence = 10;
	}
	
	/**
	 * Get the size of domain on which clustering is conducted.
	 * @param p clustering predicate
	 * @return
	 */
	public int getDomainSize(Predicate p){
		return p.getTypeAt(0).size()/this.partitionedInto;
	}
	
	
	/**
	 * Generate Data Movement Operator used by this Coref Operator.
	 * @param rules rules defining this operator.
	 */
	public void prepareDMO(HashSet<ConjunctiveQuery> rules){
		
		// build linear representation table for this operator
		if(fq.getPredByName(corefHead.getName() + "_map") == null){
			FelixPredicate pmap = new FelixPredicate(corefHead.getName() + "_map", true);
			pmap.appendArgument(corefHead.getTypeAt(0));
			pmap.appendArgument(corefHead.getTypeAt(1));
			pmap.prepareDB(db);
			fq.addFelixPredicate(pmap);
		}
		
		
		// generate Data Movement Operator according to conjunctive queries
		for(ConjunctiveQuery cq : rules){
			
			/*
			if(cq.sourceClause.hasEmbeddedWeight()){
				Expression e = new Expression(Function.GreaterThan);
				Expression tmpe1 = Expression.exprVariableBinding(cq.sourceClause.getVarWeight());
				Expression tmpe = Expression.exprConstInteger(0);
				e.addArgument(tmpe1);
				e.addArgument(tmpe);
				e.changeName = false;
					
				cq.addConstraint(e);
			}*/
			
			DataMovementOperator dmo = new DataMovementOperator(db, this);
			dmo.logicQueryPlan.addQuery(cq, cq.head.getPred().getArgs(),
					new ArrayList<String>(Arrays.asList("weight")) );
			dmo.whichToBound.add(cq.head.getTerms().get(0).toString());
			
			// hard rule
			if(cq.sourceClause.isHardClause() && cq.getWeight() > 0){
				dmo.predictedBB = 0;
				dmo.PredictedFF = 1;
				dmo.PredictedBF = 0;
				hardPosDMOs.add(dmo);
				allDMOs.add(dmo);
			}
			// hard negative rule
			else if(cq.sourceClause.isHardClause() && cq.getWeight() < 0){
				dmo.predictedBB = 0;
				dmo.PredictedFF = 1;
				dmo.PredictedBF = this.getDomainSize(corefHead);
				hardNegDMOs.add(dmo);
				allDMOs.add(dmo);
			}
			// soft incomplete rule
//TODO
//			else if( (!cq.sourceClause.isHardClause() && cq.getWeight() > 0) ||
//					cq.sourceClause.hasEmbeddedWeight()
//			){
			else if( (!cq.sourceClause.isHardClause() )){
				dmo.predictedBB = 0;
				dmo.PredictedFF = 0;
				dmo.PredictedBF = this.getDomainSize(corefHead);
				softPosDMOs.add(dmo);
				allDMOs.add(dmo);
			}else{
				UIMan.warn("The following rule is ignored in the COREFOperator!\n" + cq);
			}
			
		}
		
		// generate Data Movement Operator used in this operator
		
			//first, the DMO used to fetch the node domain 
		DataMovementOperator dmo = new DataMovementOperator(db, this);
		dmo.predictedBB = 0; dmo.PredictedFF = 1; dmo.PredictedBF = 0;
		dmo.logicQueryPlan.addQuery(db.getPrepareStatement(
				"SELECT DISTINCT constantID FROM " + corefHead.getTypeAt(0).getRelName()),
				new ArrayList<String>(Arrays.asList("constantID")), new ArrayList<String>());
		dmo.allowOptimization = false;
		nodeListDMO = dmo;
		allDMOs.add(dmo);
		
			//second, the DMO for the union of all hard-pos DMO
		if(this.hardPosDMOs.size() > 0){
			this.hardPosDMO = DataMovementOperator.UnionAll(db, this, 
					this.hardPosDMOs, "000", new ArrayList<Integer>());
			this.hardPosDMO.isIntermediaDMO = true;
			allDMOs.add(hardPosDMO);
		}
		
			//third, the DMO for the union of all soft-pos DMO
		if(this.softPosDMOs.size() > 0){
			
			this.softPosDMO1 = DataMovementOperator.UnionAll(db, this, 
				this.softPosDMOs, "100", new ArrayList<Integer>());
			this.softPosDMO1.isIntermediaDMO = true;
			allDMOs.add(softPosDMO1);
			
			/*
			DataMovementOperator groupedSoftPosDMO1 = new DataMovementOperator(db, this);
			groupedSoftPosDMO1.predictedBB = 0; groupedSoftPosDMO1.PredictedFF = 1; groupedSoftPosDMO1.PredictedBF = 0;
			groupedSoftPosDMO1.logicQueryPlan.addQuery(db.getPrepareStatement(
					"SELECT " + softPosDMO1.finalSelList.get(0) + ","
							  +	softPosDMO1.finalSelList.get(1) + ","
							  + "sum(" + softPosDMO1.finalSelList.get(0) + ") as sumweight "
				  + "FROM "   + softPosDMO1.getAllFreeViewName() + " "
				  + "GROUP BY " + softPosDMO1.finalSelList.get(0) + ","
				                + softPosDMO1.finalSelList.get(1) + " "
				  + "WHERE "  + "sumweight > 0"
				              + " AND " + softPosDMO1.finalSelList.get(0) + " = ?"),
				  softPosDMO1.finalSelList, new ArrayList<String>());
			groupedSoftPosDMO1.allowOptimization = false;
			allDMOs.add(groupedSoftPosDMO1);
			*/
			
			
			this.softPosDMO2 = DataMovementOperator.UnionAll(db, this, 
					this.softPosDMOs, "010", new ArrayList<Integer>());
			this.softPosDMO2.isIntermediaDMO = true;
			allDMOs.add(softPosDMO2);
			
			/*
			DataMovementOperator groupedSoftPosDMO2 = new DataMovementOperator(db, this);
			groupedSoftPosDMO2.predictedBB = 0; groupedSoftPosDMO2.PredictedFF = 1; groupedSoftPosDMO2.PredictedBF = 0;
			groupedSoftPosDMO2.logicQueryPlan.addQuery(db.getPrepareStatement(
					"SELECT " + softPosDMO2.finalSelList.get(0) + ","
							  +	softPosDMO2.finalSelList.get(1) + ","
							  + "sum(" + softPosDMO2.finalSelList.get(0) + ") as sumweight "
				  + "FROM "   + softPosDMO2.getAllFreeViewName() + " "
				  + "GROUP BY " + softPosDMO2.finalSelList.get(0) + ","
				                + softPosDMO2.finalSelList.get(1) + " "
				  + "WHERE "  + "sumweight > 0"
				              + " AND " + softPosDMO2.finalSelList.get(0) + " = ?"),
				  softPosDMO1.finalSelList, new ArrayList<String>());
			groupedSoftPosDMO2.allowOptimization = false;
			allDMOs.add(groupedSoftPosDMO2);
			*/
		}
		
			//forth, the DMO for the union of all hard-negative DMO
		if(this.hardNegDMOs.size() > 0){
			this.hardNegDMO1 = DataMovementOperator.UnionAll(db, this, 
				this.hardNegDMOs, "100", new ArrayList<Integer>());
			this.hardNegDMO1.isIntermediaDMO = true;
			allDMOs.add(hardNegDMO1);
			
			this.hardNegDMO2 = DataMovementOperator.UnionAll(db, this, 
					this.hardNegDMOs, "010", new ArrayList<Integer>());
			this.hardNegDMO2.isIntermediaDMO = true;
			allDMOs.add(hardNegDMO2);
		}
		
			//then process special rules
		for(ConjunctiveQuery cq : fq.getSpecialClusteringRules(this.corefHead.getName())){
		
			if(cq.type == ConjunctiveQuery.CLUSTERING_RULE_TYPE.NODE_CLASS){
				
				nodeClassRule = new DataMovementOperator(db, this);
				nodeClassRule.logicQueryPlan.addQuery(cq, cq.head.getPred().getArgs(),
						new ArrayList<String>() );
				nodeClassRule.whichToBound.add(cq.head.getTerms().get(0).toString());
				nodeClassRule.allowOptimization = false;
				allDMOs.add(nodeClassRule);
				
			}else if(cq.type == ConjunctiveQuery.CLUSTERING_RULE_TYPE.CLASS_TAGS){
				classTagsRule = new DataMovementOperator(db, this);
				classTagsRule.logicQueryPlan.addQuery(cq, cq.head.getPred().getArgs(),
						new ArrayList<String>() );
				classTagsRule.whichToBound.add(cq.head.getTerms().get(0).toString());
				classTagsRule.allowOptimization = false;
				allDMOs.add(classTagsRule);
				
			}else{
				ExceptionMan.die("No special rules other than NODE_CLASS " +
						"and CLASS_TAGS are supported in Felix!");
			}
		
		}
		
		
	}
	
	boolean prepared = false;
	
	/**
	 * Prepares operator for execution.
	 */
	@Override
	public void prepare() {
		
		softPosDMOs.clear();
		
		softNegDMOs.clear();
		
		hardPosDMOs.clear();
		
		hardNegDMOs.clear();
		
		allDMOs.clear();
		//if(!prepared){
			
			db = RDB.getRDBbyConfig(Config.db_schema);
			
			corefHead = this.getTargetPredicateIfHasOnlyOne();
			HashSet<ConjunctiveQuery> rules = 
				this.translateFelixClasesIntoFactorGraphEdgeQueries(corefHead, true, 
													this.inputPredicateScope,
													FPProperty.NON_RECUR,
													FPProperty.CHAIN_RECUR,
													FPProperty.OTHER_RECUR);
			
			this.prepareDMO(rules);
			prepared = true;
		//}
	}

	/**
	 * Executes operator.
	 */
	@Override
	public void run() {
		UIMan.print(">>> Start Running " + this);
		
		try{
			
			this.isMarginal = belongsToBucket.isMarginal();
			
			Timer.start("Coref-Op" + this.getId());
			
			if(corefHead == null){
				throw new Exception("The head of this Coref operator is NULL.");
			}
		
			db.disableAutoCommitForNow();
			
			cluster();

			db.setAutoCommit(true);
			
			//this.oriMLN.dumpMapAnswerForPredicate(options.fout+"_coref_" + headPredicate.getName() + "_op" + id, 
			//		headPredicate, false);
			
			FelixUIMan.println(0,0, "\n>>> {" + this + "} uses " + Timer.elapsed("Coref-Op" + this.getId()));
			
			if(!options.useDualDecomposition){
				this.belongsToBucket.runNextOperatorInBucket();
			}
			
			db.commit();
			//db.close();

			// TODO!!!!!!!!!!!!!!!
			//			db.close();
			
		} catch (Exception e) {
			e.printStackTrace();
		}
		
	}

	@Override
	public String explain() {
		// TODO Auto-generated method stub
		return null;
	}
	
	/**
	 * Clustering worker.
	 */
	public void cluster() throws Exception{

		procSepcialRules();
		
		// get domain for the clustering predicate
		// We assume P(type,type)
		ArrayList<Integer> nodes = new ArrayList<Integer>();		
		this.nodeListDMO.execute(null, new ArrayList<Integer>());
		while(this.nodeListDMO.next()){
			nodes.add(this.nodeListDMO.getNext(1));
		}		
		
		FelixUIMan.println(2,0,"#nodes = " + nodes.size());

		Collections.shuffle(nodes);
		HashMap<Integer, Integer> rankMap = new HashMap<Integer, Integer>();

		ArrayList<Integer> ranks = new ArrayList<Integer>();
		for(int i=0; i<nodes.size(); i++){
			rankMap.put(nodes.get(i), i);
			ranks.add(i);
		}

		UnionFind<Integer> clusters = new UnionFind<Integer>();
		clusters.makeUnionFind(ranks);
		ranks = null;

		Timer.start("clustering");
		int ct = 0;
		int edges = 0;

		HashMap<Integer, HashSet<Integer>> hardClusters = new HashMap<Integer, HashSet<Integer>>();
		
		for(int i=0; i<nodes.size(); i++){
			HashSet<Integer> s = new HashSet<Integer>();
			s.add(nodes.get(i));
			hardClusters.put(i, s);
		}

		if(this.hardPosDMO != null){
			//UIMan.println(">>> Processing hard positive edges...");
			Timer.start("hardpos");
			db.disableAutoCommitForNow();

			this.hardPosDMO.execute(null, new ArrayList<Integer>());
			int cnt = 0;
			while(this.hardPosDMO.next()){
				cnt ++;
				if(cnt % 100000000 == 0){
				//	UIMan.print("*");
					FelixUIMan.println(2,0, "# hard edges: " + cnt);
				}
				Integer i = rankMap.get(this.hardPosDMO.getNext(1));
				Integer j = rankMap.get(this.hardPosDMO.getNext(2));
				i = clusters.getRoot(i);
				j = clusters.getRoot(j);
				if(i == j) continue;
				if(i < j){
					hardClusters.get(i).addAll(hardClusters.get(j));
					hardClusters.remove(j);
				}else{
					hardClusters.get(j).addAll(hardClusters.get(i));
					hardClusters.remove(i);
				}
				clusters.unionByValue(i, j);
			}

			int x = 0;
			for(int y : hardClusters.keySet()){
				x += hardClusters.get(y).size();
			}

			db.restoreAutoCommitState();
			//Timer.printElapsed("hardpos");
		}


		for(int i=0; i<nodes.size(); i++){
			
			int s = nodes.get(i);
			if(ct % 100000 == 0){
				//UIMan.print(".");
				FelixUIMan.println(2,0,ct + "/" + nodes.size() +" : " + 
						Timer.elapsed("clustering") + "  edges : " + edges);
				int nc = clusters.getNumClusters();
				FelixUIMan.println(2,0,"#clusters = " + nc);
			}
			ct ++;

			Integer root = clusters.getRoot(i);

			if(root != i) continue;
			HashSet<Integer> classesInCluster = null;
			
			if(nodeClassRule != null && classTagsRule != null){
				classesInCluster = new HashSet<Integer>();
				for(int n : hardClusters.get(root)){
					classesInCluster.add(nodeClass.get(n));
				}
			}
			
			List<Integer> wl = this.retrieveNeighbors(s);
			HashSet<Integer> brokenByHardNeg = new HashSet<Integer>();
			brokenByHardNeg.addAll(this.retrieveHardNegEdges(s));

			edges += wl.size();
			
			//if(couldLinkPairwiseDMO != null){
			//	group.addAll(hardClusters.get(i));
			//}
			HashSet<Integer> merged = new HashSet<Integer>();
			merged.add(root);

			for(int t : wl){
				
				int j = rankMap.get(t);
				Integer r2 = clusters.getRoot(j);
				if(j <= i || j != r2 || r2 == root) continue;
				if(brokenByHardNeg.contains(t)) continue;
				
				if(nodeClassRule != null && classTagsRule != null){
					HashSet<Integer> classesInOtherCluster = new HashSet<Integer>();
					for(int n : hardClusters.get(r2)){
						classesInOtherCluster.add(nodeClass.get(n));
					}
					classesInOtherCluster.removeAll(classesInCluster);
					boolean compatible = true;
					if(!classesInOtherCluster.isEmpty()){
						labf:
							for(int x : classesInCluster){
								for(int y : classesInOtherCluster){
									HashSet<Integer> xt = classTags.get(x);
									HashSet<Integer> yt = classTags.get(y);
									xt.retainAll(yt);
									if(xt.isEmpty()){
										compatible = false;
										break labf;
									}
								}
							}
					}
					if(compatible){
						classesInCluster.addAll(classesInOtherCluster);
					}else{
						continue;
					}
				} 
				
				clusters.unionByValue(root, r2);
				brokenByHardNeg.addAll(this.retrieveHardNegEdges(t));
				merged.add(r2);
			}
		}

		int nc = clusters.getNumClusters();
		FelixUIMan.println(2,0,"# clusters = " + nc);
		FelixUIMan.println(2,0,"# edges : " + edges);
		
		this.dumpAnswerToDBTable(corefHead, clusters, nodes);
	}
	
	/**
	 * Get soft-pos neighbors of a given node.
	 * @param m1
	 * @return
	 * @throws SQLException
	 */
	public List<Integer> retrieveNeighbors(Integer m1) throws SQLException{
		
		ArrayList<Integer> ret = new ArrayList<Integer>();
		HashSet<Integer> ns = new HashSet<Integer>();
		
		if(this.softPosDMO1 != null){
			
			this.softPosDMO1.execute(null , new ArrayList<Integer>(Arrays.asList(m1)));
			
			while(this.softPosDMO1.next()){
				Integer i = softPosDMO1.getNext(1) + softPosDMO1.getNext(2) - m1;
				if(!ns.contains(i)){
					ns.add(i);
					ret.add(i);
				}
			}
		}
		
		if(this.softPosDMO2 != null){
			
			this.softPosDMO2.execute(null , new ArrayList<Integer>(Arrays.asList(m1)));
			
			while(this.softPosDMO2.next()){
				Integer i = softPosDMO2.getNext(1) + softPosDMO2.getNext(2) - m1;
				if(!ns.contains(i)){
					ns.add(i);
					ret.add(i);
				}
			}
		}
		
		return ret;
	}
	
	/**
	 * Get hard-neg neighbors of a given node.
	 * @param m1
	 * @return
	 * @throws SQLException
	 */
	public List<Integer> retrieveHardNegEdges(Integer m1) throws SQLException{
		
		ArrayList<Integer> ret = new ArrayList<Integer>();
		HashSet<Integer> ns = new HashSet<Integer>();
		
		if(this.hardNegDMO1 != null){
			
			this.hardNegDMO1.execute(null , new ArrayList<Integer>(Arrays.asList(m1)));
			
			while(this.hardNegDMO1.next()){
				Integer i = hardNegDMO1.getNext(1) + hardNegDMO1.getNext(2) - m1;
				if(!ns.contains(i)){
					ns.add(i);
					ret.add(i);
				}
			}
		}
		
		if(this.hardNegDMO2 != null){
			
			this.hardNegDMO2.execute(null , new ArrayList<Integer>(Arrays.asList(m1)));
			
			while(this.hardNegDMO2.next()){
				Integer i = hardNegDMO2.getNext(1) + hardNegDMO2.getNext(2) - m1;
				if(!ns.contains(i)){
					ns.add(i);
					ret.add(i);
				}
			}
		}
		
		return ret;
	}
	
	/**
	 * Process Tag and Class rules.
	 */
	private void procSepcialRules(){
		//System.out.println(">>> Processing special rules...");
		if(nodeClassRule != null){
			nodeClass = new HashMap<Integer, Integer>();
			nodeClassRule.execute(null, new ArrayList<Integer>());
			while(nodeClassRule.next()){
				int a = nodeClassRule.getNext(1);
				int b = nodeClassRule.getNext(2);
				nodeClass.put(a, b);
			}
		}

		if(classTagsRule != null){
			classTags = new HashMap<Integer, HashSet<Integer>>();
			classTagsRule.execute(null, new ArrayList<Integer>());
			while(classTagsRule.next()){
				int a = classTagsRule.getNext(1);
				int b = classTagsRule.getNext(2);
				if(!classTags.containsKey(a)){
					classTags.put(a, new HashSet<Integer>());
				}
				classTags.get(a).add(b);
			}

		}

	}
	
	/**
	 * Dump answers to a database table (or create view for it).
	 * @param p clustering predicate.
	 * @param clusters clustering result.
	 * @param nodes domain on which clustering is conducted.
	 */
	public void dumpAnswerToDBTable(Predicate p, UnionFind<Integer> clusters, ArrayList<Integer> nodes){

		File loadingFile = new File(Config.getLoadingDir(), "loading_cg_" + p.getRelName() + "_op" + this.getId());
		//p.nextTupleID = 0;

		Predicate pmap = fq.getPredByName(p.getName() + "_map");
		//pmap.nextTupleID = 0;
		
		String relLinear = pmap.getRelName();

		try {
			BufferedWriter loadingFileWriter = new BufferedWriter(new OutputStreamWriter
					(new FileOutputStream(loadingFile),"UTF8"));		

			HashMap<Integer, Integer> map = clusters.getPartitionMap();
			HashMap<Integer, HashSet<Integer>> label2mention = new HashMap<Integer, HashSet<Integer>>();
			for(int k : map.keySet()){
				int c = map.get(k);
				k = nodes.get(k);
				c = nodes.get(c);
				HashSet<Integer> set = label2mention.get(c);
				if(set == null){
					set = new HashSet<Integer>();
					label2mention.put(c, set);
				}
				set.add(k);
			}

			if(this.usePairwiseRepresentation){

				for(Integer clusterID : label2mention.keySet()){
					if(label2mention.get(clusterID).size() == 0){
						continue;
					}

					for(Integer node1 : label2mention.get(clusterID)){
						for(Integer node2 : label2mention.get(clusterID)){

							ArrayList<String> parts = new ArrayList<String>();
							//parts.add(Integer.toString(p.nextTupleID(p.nextTupleID++)));
							parts.add("TRUE");
							if(options.useDualDecomposition){
								parts.add(Integer.toString(1));
							}else{
								parts.add(Integer.toString(2));
							}


							parts.add(Integer.toString(node1));
							parts.add(Integer.toString(node2));

							loadingFileWriter.append(StringMan.join(",", parts) + "\n");

						}
					}
				}
			}else{
				for(Integer clusterID : label2mention.keySet()){
					if(label2mention.get(clusterID).size() == 0){
						continue;
					}
					
					int newClusterID = clusterID;
					for(Integer node1 : label2mention.get(clusterID)){
						if(node1 < newClusterID){
							newClusterID = node1;
						}
					}

					for(Integer node1 : label2mention.get(clusterID)){

						ArrayList<String> parts = new ArrayList<String>();
						//parts.add(Integer.toString(p.nextTupleID(pmap.nextTupleID++)));
						//parts.add("TRUE");
						
						if(options.useDualDecomposition){
							parts.add("TRUE");
							parts.add(Integer.toString(2));
						}else{
							parts.add("TRUE");
							parts.add(Integer.toString(2));
						}

						parts.add(Integer.toString(node1));
						parts.add(Integer.toString(newClusterID));

						loadingFileWriter.append(StringMan.join(",", parts) + "\n");

					}
				}
			}
			loadingFileWriter.close();

			if(this.usePairwiseRepresentation){

				FileInputStream in = new FileInputStream(loadingFile);
				PGConnection con = (PGConnection)db.getConnection();

				String sql;
				//String sql = "DELETE FROM " + p.getRelName();
				//db.update(sql);
				//db.vacuum(p.getRelName());

				sql = "COPY " + p.getRelName() + "(truth, club, " + StringMan.commaList(p.getArgs())  + " ) FROM STDIN CSV";
				con.getCopyAPI().copyIn(sql, in);
				in.close();
				p.isCurrentlyView = false;

			}else{

				FileInputStream in = new FileInputStream(loadingFile);
				PGConnection con = (PGConnection)db.getConnection();

				String sql;
				db.dropView(p.getRelName());
				db.dropTable(p.getRelName());

				//sql = "DELETE FROM " + relLinear;
				//db.update(sql);
				//db.vacuum(relLinear);

				if(options.useDualDecomposition){

					if(FelixConfig.isFirstRunOfDD){
						sql = "COPY " + relLinear + "(truth, club," + StringMan.commaList(pmap.getArgs())  + " ) FROM STDIN CSV";
						con.getCopyAPI().copyIn(sql, in);
						in.close();
						//p.setHasSoftEvidence(true);
					}
				}else{
					sql = "COPY " + relLinear + "(truth, club, " + StringMan.commaList(pmap.getArgs())  + " ) FROM STDIN CSV";
					con.getCopyAPI().copyIn(sql, in);
					in.close();
				}

				db.dropIndex(relLinear + "_label_idx");
				sql = "CREATE INDEX " + relLinear + "_label_idx on " +
				relLinear + "(" + pmap.getArgs().get(1) + ")";
				db.update(sql);

				db.dropIndex(relLinear + "_node_idx");
				sql = "CREATE INDEX " + relLinear + "_node_idx on " +
				relLinear + "(" + pmap.getArgs().get(0) + ")";
				db.update(sql);

				db.analyze(relLinear);
				
				db.dropSequence(p.getRelName()+"_seq");
				sql = "CREATE SEQUENCE " + p.getRelName()+"_seq" + 
				" START WITH 1";
				db.update(sql);

				sql = "CREATE VIEW " + p.getRelName() + " AS SELECT nextval('" + p.getRelName()+"_seq')" +  
				"::integer AS id," +
				"TRUE::boolean AS truth, NULL::float as prior, " +
				"2::integer as club, NULL::integer as atomID, " +
				"t1."+pmap.getArgs().get(0)+"::integer as " + p.getArgs().get(0) + 
				", t2."+pmap.getArgs().get(0)+"::integer as " + p.getArgs().get(1) + 
				" FROM " + relLinear + " t1, " + relLinear + " t2" +
				" WHERE t1."+pmap.getArgs().get(1)+"=t2."+pmap.getArgs().get(1)+"";
				p.isCurrentlyView = true;
				//p.hasSoftEvidence = false;
				p.setHasSoftEvidence(false);
				db.update(sql);
				
			    ((FelixPredicate)p).viewDef = sql + "";
				
				if(options.useDualDecomposition){
					for(FelixPredicate fp : this.dd_CommonOutput){
						if(!fp.getName().equals(this.corefHead.getName())
								&& !fp.getName().equals(this.corefHead.corefMAPPredicate.getName())){
							ExceptionMan.die("COREF 868: There must be something wrong with the parser!");
						}
						
						in = new FileInputStream(loadingFile);
						String tableName;
						String viewName;
						
						if(fp.isCorefMapPredicate){
							tableName = this.dd_commonOutputPredicate_2_tableName.get(fp);
						}else{
							tableName = this.dd_commonOutputPredicate_2_tableName.get(fp.corefMAPPredicate);
						}
						
						if(fp.isCorefMapPredicate){
							viewName = this.dd_commonOutputPredicate_2_tableName.get(fp.oriCorefPredicate);
						}else{
							viewName = this.dd_commonOutputPredicate_2_tableName.get(fp);
						}

						if(viewName != null)
							db.dropView(viewName);
						//db.dropTable(tableName);
						
						sql = "DELETE from " + tableName;
						db.execute(sql);
						
						sql = "COPY " + tableName + "(truth, club, " + StringMan.commaList(pmap.getArgs())  + " ) FROM STDIN CSV";
						con.getCopyAPI().copyIn(sql, in);
						in.close();
						
						db.execute("UPDATE " + tableName + " SET prior = 1");

						db.dropIndex(tableName + "_label_idx");
						sql = "CREATE INDEX " + tableName + "_label_idx on " +
						tableName + "(" + pmap.getArgs().get(1) + ")";
						db.update(sql);

						db.dropIndex(tableName + "_node_idx");
						sql = "CREATE INDEX " + tableName + "_node_idx on " +
						tableName + "(" + pmap.getArgs().get(0) + ")";
						db.update(sql);

						db.analyze(tableName);
						
						if(viewName != null){
						
							db.dropSequence(viewName+"_seq");
							sql = "CREATE SEQUENCE " + viewName+"_seq" + 
							" START WITH 1";
							db.update(sql);
	
							sql = "CREATE VIEW " + viewName + " AS SELECT nextval('" + viewName+"_seq')" +  
							"::integer AS id," +
							"TRUE::boolean AS truth, 1 as prior, " +
							"2::integer as club, NULL::integer as atomID, " +
							"t1."+pmap.getArgs().get(0)+"::integer as " + p.getArgs().get(0) + 
							", t2."+pmap.getArgs().get(0)+"::integer as " + p.getArgs().get(1) + 
							" FROM " + tableName + " t1, " + tableName + " t2" +
							" WHERE t1."+pmap.getArgs().get(1)+"=t2."+pmap.getArgs().get(1)+"";
							db.update(sql);
							
							
							
						}
						
					}
				}
				
			}

		} catch (Exception e) {
			ExceptionMan.handle(e);
		}

	}


	@Override
	public void learn() {
		
	}

}
