package felix.operator;


import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import org.postgresql.PGConnection;

import tuffy.db.RDB;
import tuffy.mln.Predicate;
import tuffy.ra.ConjunctiveQuery;
import tuffy.util.Config;
import tuffy.util.ExceptionMan;
import tuffy.util.StringMan;
import tuffy.util.Timer;
import tuffy.util.UIMan;
import felix.dstruct.DataMovementOperator;
import felix.dstruct.FelixPredicate;
import felix.dstruct.FelixQuery;
import felix.dstruct.StatOperator;
import felix.dstruct.FelixPredicate.FPProperty;
import felix.parser.FelixCommandOptions;
import felix.util.FelixConfig;
import felix.util.FelixStringMan;
import felix.util.FelixUIMan;

/**
 * A CRF operator in Felix.
 * @author Ce Zhang
 *
 */
public class CRFOperator extends StatOperator{

	/**
	 * Target predicate of this CRF operator.
	 */
	FelixPredicate crfHead;
	
	/**
	 * Mapping from label's constant ID to a new label ID. This new label ID is from 0, which 
	 * is used in the inference of CRF (s.t., we can use array to represent
	 * labels).
	 */
	HashMap<String, Integer> label2ID = new HashMap<String, Integer>();
	
	/**
	 * The inverse map of {@link CRFOperator#label2ID}. Here String[] means
	 * there can be multiple fields corresponding to CRF's labeling field.
	 */
	HashMap<Integer, String[]> id2Label = new HashMap<Integer, String[]>();
	
	/**
	 * All DataMovementOperators used as LR rules.
	 */
	ArrayList<DataMovementOperator> lrDMOs = new ArrayList<DataMovementOperator>();
	
	/**
	 * The DataMovementOperator used as CRF Chain rule.
	 */
	DataMovementOperator crfDMO = null;
	
	/**
	 * The DataMovementOperator which is the union of all LR rules.
	 */
	DataMovementOperator lrDMO = null;
	
	/**
	 * The DataMovementOperator representing the table/view for all unigram features.
	 */
	DataMovementOperator unigramDMO = null;
	
	/**
	 * The DataMovementOperator representing the table/view for all bigram features.
	 */
	DataMovementOperator bigramDMO = null;
	
	/**
	 * The DataMovementOperator representing the table/view for all possible labels.
	 */
	DataMovementOperator labelDomainDMO = null;
	
	/**
	 * The DataMovementOperator fetching a partition of grounded literals (which may 
	 * be a sequence or multiples sequences)
	 */
	DataMovementOperator getAllPossiblePartitioningDMO = null;
	
	/**
	 * The DataMovementOperator fetching all bigram features for a given sequence.
	 */
	DataMovementOperator getBigramFeaturesForPartitioningDMO = null;
	
	/**
	 * The DataMovementOperator fetching all unigram features for a given sequence
	 */
	DataMovementOperator getUnigramFeaturesForPartitioningDMO = null;
	
	/**
	 * The DataMovementOperator fetching all bigram features.
	 */
	DataMovementOperator getAllBigramFeaturesDMO = null;
	
	/**
	 * The DataMovementOperator fetching all unigram features.
	 */
	DataMovementOperator getAllUnigramFeaturesDMO = null;
	
	public int nRuns = 0;

	/**
	 * The constructor of CRFOperator.
	 * @param _fq Felix query.
	 * @param _goalPredicates target predicates of this coref operator.
	 * @param _opt Command line options of this Felix run.
	 */
	public CRFOperator(FelixQuery _fq, HashSet<FelixPredicate> _goalPredicates,
			FelixCommandOptions _opt) {
		super(_fq, _goalPredicates, _opt);
		this.type = OPType.CRF;
		this.precedence = 5;
	}
	
	boolean prepared = false;
	
	/**
	 * Prepares operator for execution.
	 */
	@Override
	public void prepare() {
		
		this.lrDMOs.clear();
		allDMOs.clear();
		
		//if(!prepared){
			
			db = RDB.getRDBbyConfig(Config.db_schema);
		
			crfHead = this.getTargetPredicateIfHasOnlyOne();
			HashSet<ConjunctiveQuery> chainQueries = 
				this.translateFelixClasesIntoFactorGraphEdgeQueries(crfHead, false, this.inputPredicateScope, FPProperty.CHAIN_RECUR);
			HashSet<ConjunctiveQuery> lrQueries = 
				this.translateFelixClasesIntoFactorGraphEdgeQueries(crfHead, false, this.inputPredicateScope, FPProperty.NON_RECUR);
			
			this.prepareDMO(lrQueries, chainQueries);
			
			prepared = true;
		//}
		
	}
	
	/**
	 * Return the signature of a string array as a string.
	 * @param _array
	 * @return
	 */
	String array2str(String[] _array){
		String ret = "";
		for(String s : _array){
			ret = ret + ":" + s;
		}
		return ret;
	}

	/**
	 * Executes operator.
	 */
	@Override
	public void run() {

		nRuns ++;
		
		if(!this.options.useDualDecomposition){
			crfHead.setHasSoftEvidence(true);
		}
		
		UIMan.println(">>> Start Running " + this);
		
		Timer.start("CRF-Operator-" + crfHead.getName() + "-" + this.getId());

		try{
			BufferedWriter bw = new BufferedWriter(
					new FileWriter(Config.getLoadingDir() + "/_loading_crf_" + 
							crfHead.getName() + "_op" + this.getId()));

			this.labelDomainDMO.execute(null, new ArrayList<Integer>());
			int labelID = 0;
			int nLabelFileds = crfHead.getLabelFieldsArgs().size();
			
			while(this.labelDomainDMO.next()){
				String[] currLabel = new String[nLabelFileds];
				
				for(int i=0;i<nLabelFileds;i++){
					currLabel[i] = this.labelDomainDMO.getNext(i+1).toString();
				}
				
				this.id2Label.put(labelID, currLabel);
				this.label2ID.put(array2str(currLabel), labelID);
				labelID++;
			}
			
			if(id2Label.size() == 0){
				return;
			}			
			
			
			if(crfHead.getCRFPartitionFields() != null){
				
				db.disableAutoCommitForNow();
				this.fastInfer(bw);
				db.commit();
				db.restoreAutoCommitState();

			}else{				
				UIMan.warn("Your CRF Rule Cannot be partitioned into different components...");	
				
				ExceptionMan.die("Rewritting your rules may be a better idea... " +
						"Or simply use -noCRF option");
				
				/**
				db.disableAutoCommitForNow();
				this.slowInfer(bw);
				db.commit();
				db.restoreAutoCommitState();
				**/
			}
			
			bw.close();
			
			
			FileInputStream in = new FileInputStream(Config.getLoadingDir() + 
					"/_loading_crf_" + crfHead.getName() + "_op" + this.getId());
			
			PGConnection con = (PGConnection) db.getConnection();
			
			String sql;
			
			if(options.useDualDecomposition){
				for(FelixPredicate fp : this.dd_CommonOutput){
					if(!fp.getName().equals(this.crfHead.getName())){
						ExceptionMan.die("ERROR: I am not fuzzy-LR/CRF/COREF!!! Contact us!!!");
						continue;
					}
				
					
					in = new FileInputStream(Config.getLoadingDir() + 
							"/_loading_crf_" + crfHead.getName() + "_op" + this.getId());
					String tableName = this.dd_commonOutputPredicate_2_tableName.get(fp);
					
					
					sql = "COPY " + tableName + "(truth, prior, club, " + StringMan.commaList(crfHead.getArgs())  + " ) FROM STDIN CSV";
					con.getCopyAPI().copyIn(sql, in);
					in.close();
					
				}
				
				if(FelixConfig.isFirstRunOfDD){
					in = new FileInputStream(Config.getLoadingDir() + 
							"/_loading_crf_" + crfHead.getName() + "_op" + this.getId());
					
					sql = "COPY " + crfHead.getRelName() + "(truth, prior, club, " + StringMan.commaList(crfHead.getArgs())  + " ) FROM STDIN CSV";
					con.getCopyAPI().copyIn(sql, in);
					in.close();
				}
				crfHead.isCurrentlyView = false;
				
			}else{

				sql = "COPY " + crfHead.getRelName() + "(truth, prior, club, " + StringMan.commaList(crfHead.getArgs())  + " ) FROM STDIN CSV";
				con.getCopyAPI().copyIn(sql, in);
				in.close();
				crfHead.isCurrentlyView = false;

			}
			
			FelixUIMan.println(0,0,"\n>>> {" + this + "} uses " + Timer.elapsed("CRF-Operator-" + crfHead.getName() + "-" + this.getId()));
			
			db.close();
			
			if(!options.useDualDecomposition){
				this.belongsToBucket.runNextOperatorInBucket();
			}

		}catch(Exception e){
			e.printStackTrace();
		}
		
	}

	@Override
	public String explain() {
		//TODO:
		return null;
	}
	
	/**
	 * Conduct CRF infer WITH knowledge about partitioning which is parsed statically 
	 * from the input program.
	 * 
	 * @param bw Buffered writer to dump results.
	 */
	public void fastInfer(BufferedWriter bw){
		
		//get the whole sequence in db based on provided seqHead

		try {
			
			if(this.getAllUnigramFeaturesDMO != null){
				this.getAllUnigramFeaturesDMO.execute(null, new ArrayList<Integer>());		
				this.getAllUnigramFeaturesDMO.next();
			}else{
				//TODO:
				return;
			}
			
			boolean reachUnigramEnd = false;
			
			this.getAllPossiblePartitioningDMO.execute(null, new ArrayList<Integer>());
			int ctt = 0;
			while(this.getAllPossiblePartitioningDMO.next()){
				
				//System.out.println(ctt++);
				
				String partSignature = "";
				ArrayList<Integer> bindings = new ArrayList<Integer>();
				ArrayList<String> toSig = new ArrayList<String>();
				for(String s : crfHead.getCRFPartitionFields()){
					toSig.add(this.getAllPossiblePartitioningDMO.getNext(s) + "");
					bindings.add(this.getAllPossiblePartitioningDMO.getNext(s));
				}
				partSignature = StringMan.commaList(toSig);
				
				Sequence imSeq = new Sequence(crfHead, null, id2Label, label2ID);
				
				String[] currLabel = new String[crfHead.getLabelFieldsArgs().size()];
				String[] prevLabel = new String[crfHead.getLabelFieldsArgs().size()];
				
				while(true){
					
					if(reachUnigramEnd == true){
						break;
					}
					
					String curPartSignature = "";
					toSig = new ArrayList<String>();
					for(String s : crfHead.getCRFPartitionFields()){
						toSig.add(this.getAllUnigramFeaturesDMO.getNext(s) + "");
					}
					curPartSignature = StringMan.commaList(toSig);
					
					if(curPartSignature.equals(partSignature)){
					
						String currSignature = "";
						int lct = 0;
						toSig = new ArrayList<String>();
						for(int i=0;i<crfHead.arity();i++){
							if(crfHead.getLabelPositions().contains(i)){
								currLabel[lct++] = this.getAllUnigramFeaturesDMO.getNext(i+1).toString();
								toSig.add("%s");
							}else{
								toSig.add(this.getAllUnigramFeaturesDMO.getNext(i+1).toString());
							}
						}
						currSignature = StringMan.commaList(toSig);
						
						Double weight = this.getAllUnigramFeaturesDMO.getNextDouble(crfHead.arity()+1);
						
						imSeq.registerNodeIfNotExist(currSignature);
						imSeq.registerUnigramFeatures(currSignature, this.label2ID.get(array2str(currLabel)), weight);
						
						if(this.getAllUnigramFeaturesDMO.next() == null){
							reachUnigramEnd = true;
							break;
						}
						
					}else{
						break;
					}
				}

				this.getBigramFeaturesForPartitioningDMO.execute(null, bindings);

				/*
				System.err.println();
				System.err.println(this.crfDMO.getAllFreeViewName());
				System.err.println(this.crfDMO.physicalQueryPlan.objectConjunctiveQuery);
				System.err.println(this.crfDMO.physicalQueryPlan.objectPreparedStatement);
				System.err.println(this.getBigramFeaturesForPartitioningDMO.physicalQueryPlan.objectPreparedStatement);
				*/
				
				while(getBigramFeaturesForPartitioningDMO.next()){
					
					String prevSignature = "";

					toSig = new ArrayList<String>();
					int lct = 0;
					for(int i=0;i<crfHead.arity();i++){
						if(crfHead.getLabelPositions().contains(i)){
							prevLabel[lct++] = this.getBigramFeaturesForPartitioningDMO.getNext(i+1).toString();
							toSig.add("%s");
						}else{
							toSig.add(this.getBigramFeaturesForPartitioningDMO.getNext(i+1).toString());
						}
					}
					prevSignature = StringMan.commaList(toSig);
							
					String currSignature = "";
					lct = 0;
					toSig = new ArrayList<String>();
					for(int i=crfHead.arity();i< 2* crfHead.arity();i++){
						if(crfHead.getLabelPositions().contains(i - crfHead.arity())){
							currLabel[lct++] = this.getBigramFeaturesForPartitioningDMO.getNext(i+1).toString();
							toSig.add("%s");
						}else{
							toSig.add(this.getBigramFeaturesForPartitioningDMO.getNext(i+1).toString());
						}
					}
					currSignature = StringMan.commaList(toSig);
					
					Double weight = getBigramFeaturesForPartitioningDMO.getNextDouble(2*crfHead.arity()+1);
					
					imSeq.registerNodeIfNotExist(prevSignature);
					imSeq.registerNodeIfNotExist(currSignature);
					imSeq.registerBigramFeatures(prevSignature, currSignature, label2ID.get(array2str(prevLabel)), 
							label2ID.get(array2str(currLabel)), weight);
				}
				

				
				imSeq.infer();
				imSeq.dumpAnswers(bw);
			
			}
			
		} catch (Exception e) {
			ExceptionMan.die("unconsistent value!");
			e.printStackTrace();
		}
	}
	
	/**
	 * Conduct CRF infer WITHOUT knowledge about partitioning which is parsed statically 
	 * from the input program.
	 * 
	 * @deprecated
	 * 
	 * @param bw Buffered writer to dump results.
	 */
	public void slowInfer(BufferedWriter bw){
		
		//get the whole sequence in db based on provided seqHead

		try {
			
			if(this.getAllUnigramFeaturesDMO != null){
				this.getAllUnigramFeaturesDMO.execute(null, new ArrayList<Integer>());		
			}
			if(this.getAllBigramFeaturesDMO != null){
				this.getAllBigramFeaturesDMO.execute(null, new ArrayList<Integer>());		
			}
			
			Sequence imSeq = new Sequence(crfHead, null, id2Label, label2ID);
			String[] currLabel = new String[crfHead.getLabelFieldsArgs().size()];
			String[] prevLabel = new String[crfHead.getLabelFieldsArgs().size()];
			ArrayList<String> toSig = new ArrayList<String>();
			
			if(this.getAllUnigramFeaturesDMO != null){
				while(this.getAllUnigramFeaturesDMO.next()){
					
					String currSignature = "";
					int lct = 0;
					toSig = new ArrayList<String>();
					for(int i=0;i<crfHead.arity();i++){
						if(crfHead.getLabelPositions().contains(i)){
							currLabel[lct++] = this.getAllUnigramFeaturesDMO.getNext(i+1).toString();
							toSig.add("%s");
						}else{
							toSig.add(this.getAllUnigramFeaturesDMO.getNext(i+1).toString());
						}
					}
					currSignature = StringMan.commaList(toSig);
						
					Double weight = this.getAllUnigramFeaturesDMO.getNextDouble(crfHead.arity()+1);
						
					imSeq.registerNodeIfNotExist(currSignature);
					imSeq.registerUnigramFeatures(currSignature, this.label2ID.get(array2str(currLabel)), weight);
				}
			}
			
			
			
			if(this.getAllBigramFeaturesDMO != null){
				while(this.getAllBigramFeaturesDMO.next()){
					
					String prevSignature = "";

					toSig = new ArrayList<String>();
					int lct = 0;
					for(int i=0;i<crfHead.arity();i++){
						if(crfHead.getLabelPositions().contains(i)){
							prevLabel[lct++] = this.getAllBigramFeaturesDMO.getNext(i+1).toString();
							toSig.add("%s");
						}else{
							toSig.add(this.getAllBigramFeaturesDMO.getNext(i+1).toString());
						}
					}
					prevSignature = StringMan.commaList(toSig);
							
					String currSignature = "";
					lct = 0;
					toSig = new ArrayList<String>();
					for(int i=crfHead.arity();i< 2* crfHead.arity();i++){
						if(crfHead.getLabelPositions().contains(i - crfHead.arity())){
							currLabel[lct++] = this.getAllBigramFeaturesDMO.getNext(i+1).toString();
							toSig.add("%s");
						}else{
							toSig.add(this.getAllBigramFeaturesDMO.getNext(i+1).toString());
						}
					}
					currSignature = StringMan.commaList(toSig);
					
					Double weight = getAllBigramFeaturesDMO.getNextDouble(2*crfHead.arity()+1);
					
					imSeq.registerNodeIfNotExist(prevSignature);
					imSeq.registerNodeIfNotExist(currSignature);
					imSeq.registerBigramFeatures(prevSignature, currSignature, label2ID.get(array2str(prevLabel)), 
							label2ID.get(array2str(currLabel)), weight);
					
				}
			}
	
			imSeq.infer();
			imSeq.dumpAnswers(bw);
			
			
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Generate Data Movement Operator used by this CRF Operator.
	 * @param rules rules defining this operator.
	 */
	public void prepareDMO(HashSet<ConjunctiveQuery> lrQueries, HashSet<ConjunctiveQuery> chainQueries){
		
		try {
			
			// DMO for LR rules
			for(ConjunctiveQuery cq : lrQueries){
				
				DataMovementOperator dmo = new DataMovementOperator(db, this);
				dmo.logicQueryPlan.addQuery(cq, cq.head.getPred().getArgs(),
						new ArrayList<String>(Arrays.asList("weight")) );
				
				dmo.predictedBB = 0;
				dmo.PredictedFF = 1;
				dmo.PredictedBF = 0;
				
				dmo.allowOptimization = false;
				
				allDMOs.add(dmo);
				lrDMOs.add(dmo);
			}
				
			// DMOs for CRF chain rules
			for(ConjunctiveQuery cq : chainQueries){
					
				DataMovementOperator dmo = new DataMovementOperator(db, this);
				dmo.logicQueryPlan.addQuery(cq, cq.head.getPred().getArgs(),
						new ArrayList<String>(Arrays.asList("weight")) );
				
				dmo.predictedBB = 0;
				dmo.PredictedFF = 0;
				dmo.PredictedBF = this.getPartitionSize();
				
				dmo.allowOptimization = true;
				
				allDMOs.add(dmo);
				crfDMO = dmo;
			}
			
			
			//the DMO for the union of all LR DMOs
			if(this.lrDMOs.size() > 0){
				this.lrDMO = DataMovementOperator.UnionAll(db, this, 
						this.lrDMOs, StringMan.zeros(this.crfHead.arity() + 1), new ArrayList<Integer>());
				allDMOs.add(lrDMO);
			
				//the DMO for all unigram features.
				this.unigramDMO = new DataMovementOperator(db, this);
				this.unigramDMO.allowOptimization = false;
				this.unigramDMO.asView = false;
				this.unigramDMO.logicQueryPlan.addQuery(db.getPrepareStatement(
						"SELECT " + StringMan.commaList(lrDMO.selListFromRule) + ", sum(weight) AS sumweight " +
								" FROM " + lrDMO.getAllFreeViewName() + " GROUP BY " + StringMan.commaList(lrDMO.selListFromRule) + 
								(crfHead.getCRFPartitionFields() == null ? "" :
										" ORDER BY " + StringMan.commaList(crfHead.getCRFPartitionFields())) ),
								lrDMO.selListFromRule, new ArrayList<String>(Arrays.asList("sumweight")));
				allDMOs.add(this.unigramDMO);
			}		
			
			//the DMO for the label domain
			this.labelDomainDMO = new DataMovementOperator(db, this);
			this.labelDomainDMO.allowOptimization = false;
			ArrayList<String> fields = crfHead.getLabelFieldsTypeTable();
			ArrayList<String> fieldsViewName = new ArrayList<String>();
			for(int i=0;i<fields.size();i++){
				fieldsViewName.add("t" + i + "." + "constantid AS l" + i);
				fields.set(i, fields.get(i) + " t" + i);
			}
			this.labelDomainDMO.logicQueryPlan.addQuery(db.getPrepareStatement(
					"SELECT " + StringMan.commaList(fieldsViewName) +" FROM " + FelixStringMan.commaList(fields)),
							crfHead.getLabelFieldsArgs(), 
							new ArrayList<String>());
			allDMOs.add(this.labelDomainDMO);
			
			// check whether we can use partitioning, or we must use a slow
			// version which detects the head for each sequence using recursive SQL.
			if(crfHead.getCRFPartitionFields() == null){
				
			}else{
				
				if(this.unigramDMO != null){
					this.getAllPossiblePartitioningDMO = new DataMovementOperator(db, this);
					this.getAllPossiblePartitioningDMO.allowOptimization = false;
					this.getAllPossiblePartitioningDMO.logicQueryPlan.addQuery(db.getPrepareStatement(
							"SELECT DISTINCT * FROM ( " +
							//"SELECT " + StringMan.commaList(crfHead.getCRFPartitionFields()) 
							//	+ " FROM " + crfDMO.getAllFreeViewName() + " UNION " + 
								" SELECT "  + FelixStringMan.commaList(crfHead.getCRFPartitionFields()) 
								+ " FROM " + unigramDMO.getAllFreeViewName() + ") nt ORDER BY " + 
								FelixStringMan.commaList(crfHead.getCRFPartitionFields())), crfHead.getCRFPartitionFields(),
								new ArrayList<String>());
					allDMOs.add(this.getAllPossiblePartitioningDMO);
				}else{
					this.getAllPossiblePartitioningDMO = new DataMovementOperator(db, this);
					this.getAllPossiblePartitioningDMO.allowOptimization = false;
					this.getAllPossiblePartitioningDMO.logicQueryPlan.addQuery(db.getPrepareStatement(
							"SELECT DISTINCT * FROM ( SELECT " + FelixStringMan.commaList(crfHead.getCRFPartitionFields()) 
								+ " FROM " + crfDMO.getAllFreeViewName() + ") nt ORDER BY " + 
								FelixStringMan.commaList(crfHead.getCRFPartitionFields())), crfHead.getCRFPartitionFields(),
								new ArrayList<String>());
					allDMOs.add(this.getAllPossiblePartitioningDMO);
				}
			
								
				this.getBigramFeaturesForPartitioningDMO = DataMovementOperator.Select(db, this, 
						this.crfDMO, crfHead.getCRFPartitionFields());
				this.getBigramFeaturesForPartitioningDMO.isIntermediaDMO = true;
				this.getBigramFeaturesForPartitioningDMO.hasKnownFetchingOrder = true;
				allDMOs.add(this.getBigramFeaturesForPartitioningDMO);
				
			}
			
			this.getAllBigramFeaturesDMO = DataMovementOperator.Select(db, this, 
					this.crfDMO, new ArrayList<String>());
			this.getAllBigramFeaturesDMO.isIntermediaDMO = true;
			this.getAllBigramFeaturesDMO.hasKnownFetchingOrder = true;
			allDMOs.add(this.getAllBigramFeaturesDMO);
				
			if(this.unigramDMO != null){
				this.getAllUnigramFeaturesDMO = DataMovementOperator.SelectOrderBy(db, this, 
						this.unigramDMO, new ArrayList<String>(), 
							" ORDER BY " + StringMan.commaList(crfHead.getCRFPartitionFields()));
				this.getAllUnigramFeaturesDMO.isIntermediaDMO = true;
				this.getAllUnigramFeaturesDMO.hasKnownFetchingOrder = true;
				allDMOs.add(this.getAllUnigramFeaturesDMO);
			}
		
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Estimate the number of sequences.
	 * @return
	 */
	public int getPartitionSize(){
		if(crfHead.getCRFPartitionFields() == null){
			return 1;
		}
		
		int maxSingleField = -1;
		for(int i=0;i<crfHead.getArgs().size();i++){
			if(crfHead.getCRFPartitionFields().contains(crfHead.getArgs().get(i))){
				// just an estimate
				if(maxSingleField < crfHead.getTypeAt(i).size()){
					maxSingleField = crfHead.getTypeAt(i).size();
				}
			}
		}
		if(maxSingleField > 0){
			return maxSingleField/this.partitionedInto;
		}
		
		return Integer.MAX_VALUE;
	}

	
	/**
	 * Returns sum of given log numbers.
	 * @param logX
	 * @param logY
	 * @return
	 */
	//ACKNOWLEDGE: FROM https://facwiki.cs.byu.edu/nlp/index.php/Log_Domain_Computations
	//COPYRIGHT OF THIS FUNCTION BELONGS TO ITS ORIGINAL AUTHOR
	public static double logAdd(double logX, double logY) {

	       if (logY > logX) {
	           double temp = logX;
	           logX = logY;
	           logY = temp;
	       }

	       if (logX == Double.NEGATIVE_INFINITY) {
	           return logX;
	       }
	       
	       double negDiff = logY - logX;
	       //if (negDiff < -1000000) {
	       //    return logX;
	       //}
	       
	       return logX + java.lang.Math.log(1.0 + java.lang.Math.exp(negDiff)); 
	 }
	
	
	/**
	 * Class for a node in the {@link Sequence}.
	 */
	class Node{
		public double[][] prev2currBigram;
		public double[] currUnigram;
		
		public double[] forwardSum;
		public double[] backwardSum;
		public double[] currentMax;
		public int[] prevArgMax;
		public int[] nextArgMax;
		
		public Node(int nOfLabel){
			prev2currBigram = new double[nOfLabel][nOfLabel];
			for(int i=0;i<nOfLabel;i++){
				for(int j=0;j<nOfLabel;j++){
					prev2currBigram[i][j] = 0;
				}
			}
			
			currUnigram = new double[nOfLabel];
			forwardSum = new double[nOfLabel];
			backwardSum = new double[nOfLabel];
			currentMax = new double[nOfLabel];
			prevArgMax = new int[nOfLabel];
			nextArgMax = new int[nOfLabel];
			for(int i=0;i<nOfLabel;i++){
				currUnigram[i] = 0;
				forwardSum[i] = Double.NEGATIVE_INFINITY;
				backwardSum[i] = Double.NEGATIVE_INFINITY;
				currentMax[i] = Double.NEGATIVE_INFINITY;
				prevArgMax[i] = -1;
				nextArgMax[i] = -1;
			}
			
			
		}
	}
	
	/**
	 * In-memory representation of a CRF chain. This class
	 * supports infer (both marginal and MAP) and dumps results to file.
	 * 
	 * @author Ce Zhang
	 *
	 */
	class Sequence{
		
		// although I personally think using string instead of integer id will not
		// be so slow (because the sequence is normally short), we may like to try a 
		// pure-integer version if we think current Viterbi is slow...
		
		/**
		 * The predicate to be labeled.
		 */
		Predicate pred = null;
		
		/**
		 * The signature of the root node.
		 */
		String rootSignature = "";
		
		/**
		 * The signature of the last node
		 */
		String lastSignature = "";
				
		/**
		 * Map from signature to Node object.
		 */
		HashMap<String, Node> signature2Node = new HashMap<String, Node>();
		
		/**
		 * Set of all roots nodes in this sequence.
		 */
		HashSet<String> roots = new HashSet<String>();
		
		/**
		 * Set of all last nodes in this sequence.
		 */
		HashSet<String> lasts = new HashSet<String>();
		
		/**
		 * The optimal labels for the last nodes, which are used in MAP inference.
		 */
		HashMap<String, Integer> last2maxArg = new HashMap<String, Integer>();
		
		/**
		 * Map from one node to the next node in the chain.
		 */
		HashMap<String, String> next = new HashMap<String, String>();
		
		/**
		 * Map from one node to the previous node in the chain.
		 */
		HashMap<String, String> prev = new HashMap<String, String>();
		
		/**
		 * See {@link CRFOperator#label2ID}.
		 */
		HashMap<Integer, String[]> id2Label = new HashMap<Integer, String[]>();
		
		/**
		 * See {@link CRFOperator#id2Label}.
		 */
		HashMap<String, Integer> label2ID = new HashMap<String, Integer>();

		/**
		 * the constructor.
		 * @param _p the predicate to be labeled.
		 * @param _rootSignature the root of this sequence, which can be null (in this case,
		 * this class will find roots before infer).
		 * @param _id2Label
		 * @param _label2ID
		 */
		public Sequence(Predicate _p , String _rootSignature, HashMap<Integer, String[]> _id2Label, HashMap<String, Integer> _label2ID){
			pred = _p;
			rootSignature = _rootSignature;
			if(rootSignature != null){
				signature2Node.put(rootSignature, new Node(_id2Label.size()));
			}
			id2Label = _id2Label;
			label2ID = _label2ID;
		}
		
		
		/**
		 * Add a node in this sequence with a given signature. 
		 * @param _signature
		 */
		public void registerNodeIfNotExist(String _signature){
			
			if(signature2Node.containsKey(_signature)){
				return;
			}
			
			signature2Node.put(_signature, new Node(this.id2Label.size()));
		}
		
		/**
		 * Add a bigram feature for a node with signature _currSignature and label _currLabel.
		 * @param _prevSignature
		 * @param _currSignature
		 * @param _prevLabel
		 * @param _currLabel
		 * @param _weight
		 */
		public void registerBigramFeatures(String _prevSignature, String _currSignature, int _prevLabel, int _currLabel, Double _weight){	
			
			assert this.signature2Node.containsKey(_prevSignature);
			assert this.signature2Node.containsKey(_currSignature);
			
			if(next.containsKey(_prevSignature)){
				assert next.get(_prevSignature).equals(_currSignature);
			}else{
				next.put(_prevSignature, _currSignature);
			}
			
			if(prev.containsKey(_currSignature)){
				assert prev.get(_currSignature).equals(_prevSignature);
			}else{
				prev.put(_currSignature, _prevSignature);
			}
			
			Node tmpNode = this.signature2Node.get(_currSignature);
			tmpNode.prev2currBigram[_prevLabel][_currLabel] += _weight;
			
		}
		
		/**
		 * Add a unigram feature for a node with signature _currSignature and label _currLabel.
		 * @param _currSignature
		 * @param _currLabel
		 * @param _weight
		 */
		public void registerUnigramFeatures(String _currSignature, int _currLabel, Double _weight){	
			assert this.signature2Node.containsKey(_currSignature);
			
			Node tmpNode = this.signature2Node.get(_currSignature);
			tmpNode.currUnigram[_currLabel] = _weight;
		}
		
		/**
		 * Infer on this sequence.
		 */
		public void infer(){	
			
			//find root
			if(rootSignature == null){
				Set<String> nodes = new HashSet<String>();
				nodes.addAll(this.signature2Node.keySet());
				
				while(true){
					if(nodes.size() == 0){
						break;
					}
					
					String rootCandidate = nodes.iterator().next();
					while(true){
						if(this.next.get(rootCandidate) != null){
							rootCandidate = this.next.get(rootCandidate);
						}else{
							break;
						}
					}
					
					while(true){
						nodes.remove(rootCandidate);
						if(this.prev.get(rootCandidate) != null){
							rootCandidate = this.prev.get(rootCandidate);
						}else{
							rootSignature = rootCandidate;
							roots.add(rootSignature);
							break;
						}
					}
					
				}
			}else{
				roots.add(rootSignature);
			}
			
			
			//forward
			for(String sssss : roots){
				this.rootSignature = sssss;
				String current = this.rootSignature;
				while(true){
					if(current.equals(rootSignature)){
						Node n = this.signature2Node.get(current);
						// foreach label
						for(int i = 0; i < n.forwardSum.length; i++){
							n.forwardSum[i] = n.currUnigram[i];
							n.currentMax[i] = n.currUnigram[i];
						}
					}else{
						Node n = this.signature2Node.get(current);
						Node p = this.signature2Node.get(this.prev.get(current));
						
						for(int i = 0; i < n.forwardSum.length; i++){
							int maxArg = -1;
							double maxValue = Double.NEGATIVE_INFINITY;
							for(int j = 0; j< p.forwardSum.length; j++){
								n.forwardSum[i] = logAdd(n.prev2currBigram[j][i] + n.currUnigram[i] + p.forwardSum[j], n.forwardSum[i]);
								double tmp = n.prev2currBigram[j][i] + n.currUnigram[i] + p.currentMax[j];
								if( tmp > maxValue ){
									maxValue = tmp;
									maxArg = j;
								}
							}
							n.prevArgMax[i] = maxArg;
							n.currentMax[i] = maxValue;
						}
					}
					if(!this.next.containsKey(current)){
						
						Node n = this.signature2Node.get(current);
						this.lastSignature = current;
						
						int maxArg = -1;
						double maxValue = Double.NEGATIVE_INFINITY;
						for(int i = 0; i < n.forwardSum.length; i++){
							if(n.currentMax[i] > maxValue){
								maxArg = i;
								maxValue = n.currentMax[i];
							}
						}
						
						this.lasts.add(current);
						this.last2maxArg.put(current, maxArg);
						break;
					}
					current = this.next.get(current);
				}
					
				//backward
				String last = current;
				while(true){
					if(current.equals(last)){
						Node n = this.signature2Node.get(current);
						// foreach label
						for(int i = 0; i < n.backwardSum.length; i++){
							n.backwardSum[i] = n.currUnigram[i];
						}
					}else{
						Node p = this.signature2Node.get(current);
						Node n = this.signature2Node.get(this.next.get(current));
						
						for(int i = 0; i < p.backwardSum.length; i++){
							for(int j = 0; j< n.backwardSum.length; j++){
								p.backwardSum[i] = logAdd(n.prev2currBigram[i][j] + n.currUnigram[j] + n.backwardSum[j], p.backwardSum[i]);
							}
						}
					}				
					
					if(!this.prev.containsKey(current)){
						break;
					}
					current = this.prev.get(current);
				}
			}
		}
		
		/**
		 * Dump answers to the given buffered writer. These answers are in a format that can be
		 * COPY into postgres table directly.
		 * @param bw
		 */
		public void dumpAnswers(BufferedWriter bw){
			try{
				
				if(isMarginal || (FelixConfig.isFirstRunOfDD && dd_commonOutputPredicate_2_tableName.containsKey(crfHead) )){
					for(String sssss : roots){
						this.rootSignature = sssss;
						String current = this.rootSignature;
						while(true){
		
							if(current.equals(rootSignature)){
								Node n = this.signature2Node.get(current);
								double sum = Double.NEGATIVE_INFINITY;
								for(int i = 0; i < n.forwardSum.length; i++){
									sum = logAdd(n.currUnigram[i] + n.backwardSum[i], sum);
								}
								for(int i = 0; i < n.forwardSum.length; i++){
									double prob = n.currUnigram[i] + n.backwardSum[i] - sum;
									prob = Math.exp(prob);
									if(prob > Config.soft_evidence_activation_threshold){
										ArrayList<String> parts = new ArrayList<String>();
									//	parts.add(Integer.toString(pred.nextTupleIDAndUpdate()));
										parts.add("TRUE");
										parts.add(Double.toString(prob));
										
										if(options.useDualDecomposition){
											parts.add(Integer.toString(2));
										}else{
											parts.add(Integer.toString(2));
										}
									//	parts.add("1");//this is for vote
										
										//String tmp = current.replaceAll("\\?", id2Label.get(i));
										String tmp = String.format(current, (Object[]) id2Label.get(i));
										
										bw.append(FelixStringMan.commaListNoSpace(parts) + "," + tmp + "\n");
									}
								}
								
							}else{
								Node n = this.signature2Node.get(current);
								Node p = this.signature2Node.get(this.prev.get(current));
								
								double sum = Double.NEGATIVE_INFINITY;
								for(int i = 0; i < n.forwardSum.length; i++){
									for(int j = 0; j< p.forwardSum.length; j++){
										sum = logAdd(p.forwardSum[j] + n.currUnigram[i] + n.prev2currBigram[j][i]
										             + n.backwardSum[i], sum);
									}
								}
								
								for(int i = 0; i < n.forwardSum.length; i++){
									double marginal = 0;
									
									for(int j = 0; j< p.forwardSum.length; j++){
										marginal = logAdd( p.forwardSum[j] + n.currUnigram[i] + n.prev2currBigram[j][i] + n.backwardSum[i], 
												marginal );
									}
									
									double prob = Math.exp(marginal - sum);
									
									if(prob > Config.soft_evidence_activation_threshold){
										ArrayList<String> parts = new ArrayList<String>();
								//		parts.add(Integer.toString(pred.nextTupleIDAndUpdate()));
										parts.add("TRUE");
										parts.add(Double.toString(prob));
										
										if(options.useDualDecomposition){
											parts.add(Integer.toString(2));
										}else{
											parts.add(Integer.toString(2));
										}
									//	parts.add("1");//this is for vote
										
										String tmp = String.format(current, (Object[]) id2Label.get(i));
										bw.append(FelixStringMan.commaListNoSpace(parts) + "," + tmp + "\n");
									}
									
								}
								
							}
							
							if(!this.next.containsKey(current)){
								break;
							}
							current = this.next.get(current);
						}
					}
				}else{
					
					for(String sssss : lasts){
						
						String current = sssss;
						
						int toDump = this.last2maxArg.get(sssss);
						
						while(true){
													
							Node n = this.signature2Node.get(current);
							
							ArrayList<String> parts = new ArrayList<String>();
						//	parts.add(Integer.toString(pred.nextTupleIDAndUpdate()));
							parts.add("TRUE");
							parts.add("");
							
							if(options.useDualDecomposition){
								parts.add(Integer.toString(2));
							}else{
								parts.add(Integer.toString(2));
							}
						//	parts.add("1");//this is for vote
							
							String tmp = String.format(current, (Object[]) id2Label.get(toDump));
							bw.append(FelixStringMan.commaListNoSpace(parts) + "," + tmp + "\n");
											
							if(!this.prev.containsKey(current)){
								break;
							}
							
							toDump = n.prevArgMax[toDump];
							current = this.prev.get(current);
						}
					}
					
				}
			}catch(Exception e){
				e.printStackTrace();
			}
			
		}
	}

	@Override
	public void learn() {
		
	}
	
}
