package felix.optimizer;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import tuffy.db.RDB;
import tuffy.mln.Literal;
import tuffy.ra.ConjunctiveQuery;
import tuffy.util.Config;
import felix.dstruct.FelixPredicate;
import felix.main.Felix;
import felix.util.FelixConfig;
import felix.util.FelixUIMan;

/**
 * An object of the cost model used by the optimizer. This
 * class works as a bridge from PostgreSQL's optimizer to
 * Felix's optimizer. 
 * 
 * @author Ce Zhang
 *
 */
public class CostModel {

	/**
	 * Felix object in which this CostModel is used.
	 */
	Felix parentFelix;
	
	/**
	 * Database connection.
	 */
	RDB db;
	
	/**
	 * The ratio of postgreSQL's unit cost to page IO cost.
	 * This is estimated by some sample queries.
	 */
	public double postgreUnit = -1;
	public double memoryTradeOff = 200;
	
	/**
	 * Return the number of rows of the given table.
	 * @param tableName
	 * @return
	 */
	public double getTableSize(String tableName){
		String sql = "SELECT * FROM " + tableName;
		return getRows(sql);
	}
	
	/**
	 * Return the width (in bits) of the given table.
	 * @param tableName
	 * @return
	 */
	public double getTableWidth(String tableName){
		String sql = "SELECT * FROM " + tableName;
		return getWidth(sql);
	}
	
	/**
	 * Return a new {@link resultTuple}.
	 * @return
	 */
	public resultTuple getNewEmptyResultTuple(){
		return new resultTuple();
	}
	
	/**
	 * Return the cost of the given sql query.
	 * @param sql
	 * @return
	 */
	public double getCost(String sql){
	//	System.out.println(sql);
		String explain = db.estimateCost(sql);
		if(explain == null){
			return Double.MAX_VALUE;
		}
		Pattern pa = Pattern.compile("cost=(.*?)\\.\\.(.*?) rows=(.*?) width=(.*?)\\)");
		Matcher m = pa.matcher(explain);
		double cost = -1;
		if(m.find()){
			cost = Double.valueOf(m.group(2));
		}
		return cost;
	}
	
	/**
	 * Return the number of rows returned by the given sql query.
	 * @param sql
	 * @return
	 */
	public double getRows(String sql){
	//	System.out.println(sql);
		String explain = db.estimateCost(sql);
		if(explain == null){
			return Double.MAX_VALUE;
		}
		Pattern pa = Pattern.compile("cost=(.*?)\\.\\.(.*?) rows=(.*?) width=(.*?)\\)");
		Matcher m = pa.matcher(explain);
		double row = -1;
		if(m.find()){
			row = Double.valueOf(m.group(3));
		}
		return row;
	}
	
	/**
	 * Return the width (in bits) returned by the given sql query.
	 * @param sql
	 * @return
	 */
	public double getWidth(String sql){
		String explain = db.estimateCost(sql);
		if(explain == null){
			return Double.MAX_VALUE;
		}
		Pattern pa = Pattern.compile("cost=(.*?)\\.\\.(.*?) rows=(.*?) width=(.*?)\\)");
		Matcher m = pa.matcher(explain);
		double width = -1;
		if(m.find()){
			width = Double.valueOf(m.group(4)); 
		}
		return width;
	}
	
	/**
	 * Estimate {@link CostModel#postgreUnit}.
	 */
	public void getPostgreUnit(){
		if(postgreUnit > 0){
			return;
		}
		int ct = 0;
		double sum = 0;
		for(FelixPredicate p : parentFelix.getFelixQuery().getAllPred()){
			String sql = "SELECT * FROM " + p.getRelName();
			double cost = this.getCost(sql);
			double row = this.getRows(sql);
			double width = this.getWidth(sql);
			if(row > 0){
				sum += cost/((row*width)/8192);
				ct ++;
			}
		}
		postgreUnit = sum/ct;
	}
	
	/**
	 * The constructor of CostModel. This function will load the evidence
	 * file if it is not loaded before.
	 * @param _felix
	 */
	public CostModel(Felix _felix){
		
		parentFelix = _felix;
		if(!parentFelix.hasLoadedEvidence){
			try {
				parentFelix.loadEvidence();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		
		db = RDB.getRDBbyConfig(Config.db_schema);
		
		this.getPostgreUnit();
		FelixUIMan.println(2, 0, ">>> PostgreSQL's cost per page is " +
				"esitmated as " + this.postgreUnit + "\n");
	}
	
	/**
	 * Close the database connection used by CostModel.
	 */
	public void close(){
		db.close();
	}
	
	/**
	 * Return the cost of Fully-Materialization plan for q.
	 * @param q
	 * @param numberOfBB
	 * @param numberOfBF
	 * @param numberOfFF
	 * @param whichToBound
	 * @return
	 */
	public double getFullMatCost(ConjunctiveQuery q,  double numberOfBB, double numberOfBF, double numberOfFF, HashSet<String> whichToBound){
		
		double rs = 0.0;
		
		Double fullCost = Double.MAX_VALUE;
		Double incCost = Double.MAX_VALUE;
		Double nOfResultTuples = Double.MAX_VALUE;
		Double nOfTableTuples = Double.MAX_VALUE;
		Double width = 1.0;
		
		if(q.getBoundedSQL(new HashSet<String>()) == null){
			return Double.MAX_VALUE;
		}
		
		fullCost = this.getCost(q.getBoundedSQL(new HashSet<String>()));
		nOfTableTuples = this.getRows(q.getBoundedSQL(new HashSet<String>()));
		width = this.getWidth(q.getBoundedSQL(new HashSet<String>()));
		
		String sql = q.getBoundedSQL(whichToBound);
		if(sql == null){
			return Double.MAX_VALUE;
		}
		
		nOfResultTuples = this.getRows(sql);
		
		double wholeTableFitInMemory = 1.0;
		
		if(nOfTableTuples * width < 10000000){
			wholeTableFitInMemory = memoryTradeOff;
		}
		
		incCost = (1 + (Math.log(  nOfTableTuples)/Math.log(512) + 1)/wholeTableFitInMemory + nOfResultTuples/wholeTableFitInMemory) * postgreUnit;
		rs = nOfTableTuples/512 * postgreUnit + fullCost +  numberOfBF * incCost;
		
		return rs;
	}
	

	/**
	 * Return the cost of Fully-View plan for q.
	 * @param q
	 * @param numberOfBB
	 * @param numberOfBF
	 * @param numberOfFF
	 * @param whichToBound
	 * @return
	 */
	public double getFullViewCost(ConjunctiveQuery q, double numberOfBB, double numberOfBF, double numberOfFF, HashSet<String> whichToBound){
		
		q.buildIndexes(db, null, null, null, false, new ArrayList<String>());
		
		double rs = Double.MAX_VALUE;
		
		// if there are no BF and BB invocations at all, this query should not
		// be regarded as view.
		if(numberOfBF == 0 && numberOfBB == 0){
			if(numberOfFF == 1){
				return getFullMatCost(q, numberOfBB, numberOfBF, numberOfFF, whichToBound);
			}
			return rs;
		}
		
		// current setting is not fair to pure view mat.
		double maxTableMemorySize = Double.MIN_VALUE;
		double minTableMemorySize = Double.MAX_VALUE;
		double isEveryTableFitInMemory = 1;
		for(Literal l : q.body){
			double v = this.getTableSize(l.getPred().getRelName()) * 
				this.getTableWidth(l.getPred().getRelName());
			if(v > maxTableMemorySize){
				maxTableMemorySize = v;
			}
			if(v < minTableMemorySize){
				minTableMemorySize = v;
			}
		}
		
		if(maxTableMemorySize < 10000000){
			isEveryTableFitInMemory = memoryTradeOff;
		}
				
		String sql = q.getBoundedSQL(whichToBound);
		
		if(sql == null){
			return Double.MAX_VALUE;
		}
		
		Double incCost = Double.MAX_VALUE;
		incCost = this.getCost(sql);
		Double nResultTuples = this.getRows(sql);
		
		//TODO
		
		if(minTableMemorySize < 10000000){
			incCost -= postgreUnit * nResultTuples;
			if(incCost < 0){
				incCost = postgreUnit;
			}
		}
		
		
		rs = numberOfBF * incCost / isEveryTableFitInMemory;
		
		return rs;
	}
	
	/**
	 * Return the cost of Hybrid-Materialization plan for query {head :- q1.head, q2.head}.
	 * @param q
	 * @param numberOfBB
	 * @param numberOfBF
	 * @param numberOfFF
	 * @param whichToBound
	 * @return
	 */
	public resultTuple getJoinCostBetweenTwoMaterializedTable(Literal head, ConjunctiveQuery q1, 
			ConjunctiveQuery q2, double numberOfBB, double numberOfBF, double numberOfFF, HashSet<String> whichToBound ){
			
		resultTuple rsTuple = new resultTuple();
		rsTuple.totalCost = Double.MAX_VALUE;
		
		// if there are no BF and BB invocations at all, this query should not
		// be regarded as view.
		if(numberOfBF == 0 && numberOfBB == 0){
			return rsTuple;
		}
		
		ConjunctiveQuery q = new ConjunctiveQuery();
		q.addBodyLit(q1.head);
		q.addBodyLit(q2.head);
		q.head = head;
		
		//Direct Join (BF)
		String sql = q.getJoinSQL(whichToBound);
		if(sql == null || q1.getBoundedSQL(new HashSet<String>()) == null 
				|| q2.getBoundedSQL(new HashSet<String>()) == null){
			return null;
		}
		sql = sql.replaceAll(q1.head.getPred().getRelName(), "\n(" + q1.getBoundedSQL(new HashSet<String>()) + ")");
		sql = sql.replaceAll(q2.head.getPred().getRelName(), "\n(" + q2.getBoundedSQL(new HashSet<String>()) + ")");
		
		double n1 = this.getRows(sql);
		
		//Join with Index
		double costOfQ2 = this.getCost("select * from ( " + q2.getBoundedSQL(new HashSet<String>()) + " ) ppppp;");
		double rowOfQ2 = this.getRows("select * from ( " + q2.getBoundedSQL(new HashSet<String>()) + " ) ppppp;");
		double width3 = this.getWidth("select * from ( " + q2.getBoundedSQL(new HashSet<String>()) + " ) ppppp;");
		
		double costOfQ1 = this.getCost("select * from ( " + q1.getBoundedSQL(new HashSet<String>()) + " ) ppppp;");
		double rowOfQ1 = this.getRows("select * from ( " + q1.getBoundedSQL(new HashSet<String>()) + " ) ppppp;");
		double width2 = this.getWidth("select * from ( " + q1.getBoundedSQL(new HashSet<String>()) + " ) ppppp;");
		double n2 = this.getRows("select * from ( " + q1.getBoundedSQL(whichToBound) + " ) ppppp;");

		double leftTableFitInMemory = 1;
		double rightTableFitInMemory = 1;
		
		if(rowOfQ1 * width2 < 10000000){
			leftTableFitInMemory = memoryTradeOff;
		}
		
		if(rowOfQ2 * width3 < 10000000){
			rightTableFitInMemory = memoryTradeOff;
		}
					
		double metaCost = postgreUnit;
		
		//newCost = n2 * (n1/n2+1) / rightTableFitInMemory * metaCost  ;
		double newCost = n2 * (n1/n2/(8192/width3)+1) / rightTableFitInMemory * metaCost  ;
		double cost = newCost + n2 * (Math.log(rowOfQ2)/Math.log(512) + 1)/rightTableFitInMemory * metaCost;
		cost += (Math.log(rowOfQ1)/Math.log(512) + 1) / leftTableFitInMemory * metaCost;
		//cost /= 1.5;	//cache
		
		rsTuple.eachCostInAir = cost;
		
		cost = cost*numberOfBF;
		
		rsTuple.totalCostInAir = cost;
		
		rsTuple.costOfQ1 = costOfQ1;
		rsTuple.costOfQ2 = costOfQ2;
		rsTuple.sizeOfQ1 = rowOfQ1;
		rsTuple.sizeOfQ2 = rowOfQ2;
		rsTuple.totalCost = cost + costOfQ2 + costOfQ1;

		return rsTuple;
		
	}
	
	
	/**
	 * Class for a set of different costs.
	 * @author Ce Zhang
	 *
	 */
	public class resultTuple{
		
		double costOfQ1;
		double costOfQ2;
		double totalCost;
		double totalCostInAir;
		double eachCostInAir;
		double oriCostInAir;
		double sizeOfQ1;
		double sizeOfQ2;
		
		public String toString(){
			String content = "";
			content += "/////////////////////////////////\n";
			content += "Total Cost      = " + totalCost + "\n";
			content += "Mat. Cost Of Q1 = " + costOfQ1 + "\n";
			content += "Mat. Cost Of Q2 = " + costOfQ2 + "\n";
			content += "Size Of Q1      = " + sizeOfQ1 + "\n";
			content += "Size Of Q2      = " + sizeOfQ2 + "\n";
			content += "Total Inc. Cost = " + totalCostInAir + "\n";
			content += "Each Inc. Cost  = " + eachCostInAir + "\n";
			content += "/////////////////////////////////\n";
			return content;
		}
		
	}
	
	
}
