package felix.optimizer;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;


import tuffy.db.RDB;
import tuffy.mln.Literal;
import tuffy.mln.Term;
import tuffy.mln.Type;
import tuffy.ra.ConjunctiveQuery;
import tuffy.ra.Expression;
import tuffy.util.Config;
import tuffy.util.Timer;
import tuffy.util.UIMan;

import felix.dstruct.ConcurrentOperatorsBucket;
import felix.dstruct.DataMovementOperator;
import felix.dstruct.FelixPredicate;
import felix.dstruct.StatOperator;
import felix.optimizer.CostModel;
import felix.optimizer.CostModel.resultTuple;
import felix.society.TaskList;
import felix.society.TaskSet;
import felix.task.OptimizeDMOTask;
import felix.util.FelixConfig;
import felix.util.FelixUIMan;

/**
 * An object of DMOOptimer takes inputs as a DMO, analyzes its
 * logic plan, and fill in its physical plan. Current version of the
 * DMOOptimizer will only optimize the materialization trade-off.
 * Note that, this class does not touch the DB directly, instead
 * it uses an instance of {@link CostModel}.
 * 
 * @author Ce Zhang
 *
 */
public class DMOOptimizer {

	/**
	 * Cost model used to optimize the DMO.
	 */
	public CostModel cm;
	
	/**
	 * Database connection.
	 */
	RDB db;
	
	/**
	 * The constructor.
	 * @param _cm
	 */
	public DMOOptimizer(CostModel _cm){
		cm = _cm;
		db = RDB.getRDBbyConfig(Config.db_schema);
	}
	
	/**
	 * Close the database connection used in this DMOOptimizer.
	 */
	public void close(){
		db.close();
	}
	
	boolean picasso = false;
	
	
	/**
	 * Optimize all DMOs appearing in the given {@link ConcurrentOperatorsBucket}.
	 * @param cob
	 */
	public void optimizeDMO(ConcurrentOperatorsBucket cob){
		
		// optimize their DMOs
		ExecutorService pool = Executors.newFixedThreadPool(Config.getNumThreads());
		
		TaskList tasks = new TaskList();
		
		TaskSet taskset1 = new TaskSet();
			
		for(StatOperator op : cob.getOperators()){
			taskset1.addSubTask(new OptimizeDMOTask(op, this));
		}	
		
		tasks.addSubTask(taskset1);
		try {
			tasks.execute(pool);
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		pool.shutdown();
	}
	
	/**
	 * Returns signature of query set.
	 * @param sets
	 * @return
	 */
	public String giveMeTheSignatureOfTwoQuery(HashMap<Literal, ConjunctiveQuery> sets){
		
		String signature = "";
		
		String[] qs = new String[sets.values().size()];
		int ct = 0;
		for(ConjunctiveQuery query : sets.values()){
			
			String[] l = new String[query.body.size()];
			
			for(int i=0;i<l.length;i++){
				l[i] = query.body.get(i).toString();
			}
			
			Arrays.sort(l);
			String ss = "";
			for(String s : l){
				ss = ss + s + ",";
			}
			
			qs[ct++] = ss;
			
		}
		
		Arrays.sort(qs);
		for(String s : qs){
			signature = signature + s + "  |  ";
		}
			
		return signature;
	}
	
	/**
	 * Optimize the materialization trade-off of the given DMO.
	 * @param dmo
	 */
	public void optimizeMateralization(DataMovementOperator dmo){
		try{
			
			ConjunctiveQuery rule = dmo.logicQueryPlan.objectConjunctiveQuery;
						
			double bb = dmo.predictedBB;
			double bf = dmo.PredictedBF;
			double ff = dmo.PredictedFF;
			
			// cost for full view.
			double view = cm.getFullViewCost(rule, bb, bf, ff, dmo.whichToBound);
			
			// cost for full materalization.
			double full = cm.getFullMatCost(rule, bb, bf, ff, dmo.whichToBound);
										
			HashMap<ConjunctiveQuery, HashMap<Literal, ConjunctiveQuery>> comb = 
				new HashMap<ConjunctiveQuery, HashMap<Literal, ConjunctiveQuery>>();
			ArrayList<String> additionalSelList = new ArrayList<String>();
			HashMap<String, String> typeMapping = new HashMap<String, String>();
			
			for(String s : dmo.additionalSelList){
				if(s.equals("weight")){
					if(rule.sourceClause.hasEmbeddedWeight()){
						additionalSelList.add(rule.sourceClause.getVarWeight());
						typeMapping.put(rule.sourceClause.getVarWeight(), "double_");
					}
				}
			}
			
			// all binary decomposition of this conjunctive query.
			comb = this.generateAllPossiblePlans(rule.head, rule.body, additionalSelList, typeMapping);
			
			ConjunctiveQuery bestQuery = null;
			HashMap<Literal, ConjunctiveQuery> fullplan = null;
			ConjunctiveQuery fullQuery = null;
			
			resultTuple smallestCost = cm.getNewEmptyResultTuple();
			smallestCost.totalCost = Double.MAX_VALUE;
			
			if(comb != null){
				for(ConjunctiveQuery cq : comb.keySet()){
		
					ConjunctiveQuery q1 = null;
					ConjunctiveQuery q2 = null;
					
					for(Literal sub : cq.body){
						if(q1 == null)
							q1 = comb.get(cq).get(sub);
						else
							q2 = comb.get(cq).get(sub);
					}
					
					if(q1 != null){
						q1.addConstraintAll(rule.getConstraint(q1.allVariable));
					}
					
					if(q2 != null){
						q2.addConstraintAll(rule.getConstraint(q2.allVariable));
					}
					
					if(q2 == null){
						fullQuery = cq;
						fullplan = comb.get(cq);
						continue;
					}
					
					//TODO:
					HashSet<Expression> cons = new HashSet<Expression>();
					cons.addAll(q1.getConstraint());
					cons.addAll(q2.getConstraint());
					if(rule.getConstraint().size() > cons.size()){
						continue;
					}
					
					resultTuple cost = cm.getJoinCostBetweenTwoMaterializedTable(
							cq.head, q1, q2, bb, bf, ff, dmo.whichToBound);
					
					if(cost == null){
						continue;
					}
					
					if(FelixConfig.pickRandom== false && (smallestCost.totalCost == Double.MAX_VALUE || 
						(cost.totalCost <= smallestCost.totalCost/1)) ){
										
						bestQuery = cq;
						smallestCost = cost;
					}					
					
					if(FelixConfig.pickRandom==true && Math.random()<0.8 && cost.totalCost < 300000){
						
						bestQuery = cq;
						smallestCost = cost;
					}
					
				}
			}else{
				fullplan = new HashMap<Literal, ConjunctiveQuery>();
				fullQuery = new ConjunctiveQuery();
				fullQuery.setHead(rule.head);
			}
						
			if(picasso == true){
				
				System.out.print("postgreUnit=" + cm.postgreUnit + "\tmemoryTradeOff=" + cm.memoryTradeOff + "\t");
				
				if(smallestCost.totalCost <= view && smallestCost.totalCost <= full){
					
					System.out.println(this.giveMeTheSignatureOfTwoQuery(comb.get(bestQuery)));
					
				}else if (view <= smallestCost.totalCost && view <= full){
					System.out.println("VIEW");
				}else{
					System.out.println("FULL");
				}
				
				
			}else{
				
				//TODO:
				//if(Config.gp == true){
				//	FelixConfig.allMat = true;
				//}
				
				
				// select the best one from hybrid/full-mat/full-view
				FelixUIMan.printobj(2,0,rule);
				if(comb != null){
					FelixUIMan.printobj(2,0,smallestCost);
					if(bestQuery != null){
						FelixUIMan.printobj(2,0,comb.get(bestQuery));
					}
				}
				FelixUIMan.println(2,0,"full = " + full);
				FelixUIMan.println(2,0,"view = " + view);
				
				if(FelixConfig.allView){
					
					FelixUIMan.println(1,0,">>> Regard the following query as view: \n" + "\t" + rule);
					
				//	SchedulerTest.planCosts.add(view);
					rule.isView = true;
					dmo.physicalQueryPlan.objectConjunctiveQuery = rule;
				}else{
				
					if(FelixConfig.allMat){
						
						FelixUIMan.println(1,0,">>> Regard the following query as materialized table: \n" + "\t" + rule);
						
						//	SchedulerTest.planCosts.add(full);
							
						rule.isView = false;
						
						fullQuery.isView = true;
						fullQuery.type = rule.type;
						fullQuery.setWeight(rule.getWeight());
						fullQuery.sourceClause = rule.sourceClause;
						
						fullQuery.inverseEmbededWeight = rule.inverseEmbededWeight;
						
						dmo.physicalQueryPlan.objectConjunctiveQuery = fullQuery;
						if(fullplan.size() > 0){
							dmo.physicalQueryPlan.datalogQueries.add(fullplan.get(fullQuery.body.get(0)));
							fullplan.get(fullQuery.body.get(0)).head.getPred().prepareDB(db);
						}
						
					}else{
					
						if(smallestCost.totalCost <= view && smallestCost.totalCost <= full){
							
					//		SchedulerTest.planCosts.add(smallestCost.totalCost);
							
							ConjunctiveQuery q1 = null;
							ConjunctiveQuery q2 = null;
							
							if(bestQuery == null){
								throw new Exception("Errors!");
							}
							
							for(Literal sub : comb.get(bestQuery).keySet()){
								if(q1 == null) q1 = comb.get(bestQuery).get(sub);
								else q2 = comb.get(bestQuery).get(sub);
							}
							
							ConjunctiveQuery rsCQ = new ConjunctiveQuery();
							rsCQ.isView = true;
							rsCQ.type = rule.type;
							rsCQ.setWeight(rule.getWeight());
							rsCQ.sourceClause = rule.sourceClause;
							rsCQ.inverseEmbededWeight = rule.inverseEmbededWeight;
							
							rsCQ.setHead(bestQuery.head);
							
							if(q1.body.size() == 1){
								rsCQ.addBodyLit(q1.body.get(0));
							}else{
								rsCQ.addBodyLit(q1.head);
								dmo.physicalQueryPlan.datalogQueries.add(q1);
								//q1.head.getPred().createTable();
								q1.head.getPred().prepareDB(db);
							}
							
							if(q2.body.size() == 1){
								rsCQ.addBodyLit(q2.body.get(0));
							}else{
								rsCQ.addBodyLit(q2.head);
								dmo.physicalQueryPlan.datalogQueries.add(q2);
								q2.head.getPred().prepareDB(db);
							}
							
							dmo.physicalQueryPlan.objectConjunctiveQuery = rsCQ;
							
							q1.sourceClause = rule.sourceClause;
							q2.sourceClause = rule.sourceClause;
							
							FelixUIMan.println(1,0,">>> Regard the following query as partially materialized view: \n" 
									+ "\t" + rule + "\n" + "\t" + q1 + "\n" + "\t" + q2);
							
						}else if (view <= smallestCost.totalCost && view <= full){
							
							FelixUIMan.println(1,0,">>> Regard the following query as view: \n" + "\t" + rule);
							
						//	SchedulerTest.planCosts.add(view);
							
							rule.isView = true;
							dmo.physicalQueryPlan.objectConjunctiveQuery = rule;
							
						}else{
							
							FelixUIMan.println(1,0,">>> Regard the following query as materialized table: \n" + "\t" + rule);
							
						//	SchedulerTest.planCosts.add(full);
							
							rule.isView = false;
							
							fullQuery.isView = true;
							fullQuery.type = rule.type;
							fullQuery.setWeight(rule.getWeight());
							fullQuery.sourceClause = rule.sourceClause;
							
							fullQuery.inverseEmbededWeight = rule.inverseEmbededWeight;
							
							dmo.physicalQueryPlan.objectConjunctiveQuery = fullQuery;
							if(fullplan.size() > 0){
								dmo.physicalQueryPlan.datalogQueries.add(fullplan.get(fullQuery.body.get(0)));
								fullplan.get(fullQuery.body.get(0)).head.getPred().prepareDB(db);
							}
						}
					}
				}
			}
			
			
			
		}catch(Exception e){
			e.printStackTrace();
		}
	}
	
	
	/**
	 * Given a literal as goal, a set of literals as subgoals, generate all possible binary decompositions
	 * of this set of subgoals. The decomposition looks like:
	 * 
	 * <br/>
	 * 
	 * Q(...) :- Q1(...), Q2(...) <br/>
	 * Q1(...) :- g1, g2, ... <br/>
	 * Q2(...) :- g1', g2', ... <br/>
	 * 
	 * <br/>
	 * 
	 * Q1 and Q2 are materialized, and Q will be regarded as a view. If Q2 is null, this
	 * is equivalent to the fully-materialized case.
	 * 
	 * @param head goal
	 * @param subgoals set of subgoals
	 * @param additionalSelList terms needed to be maintained in the variable list of Q1 and Q2.
	 * By default all variables needed to compute Q from Q1 and Q2 will be automatically maintained,
	 * however, if you want to maintain others, add them in this list.
	 * @param typeMapping If you want some variables in Q1 and Q2 have special types (e.g., double instead
	 * of constant ID), put them in this map.
	 * @return Mappings from Q to Q1 and Q2.
	 */
	public HashMap<ConjunctiveQuery, HashMap<Literal, ConjunctiveQuery>> generateAllPossiblePlans(Literal head, 
			ArrayList<Literal> subgoals, ArrayList<String> additionalSelList, HashMap<String, String> typeMapping){
		
		HashMap<ConjunctiveQuery, HashMap<Literal, ConjunctiveQuery>> forReturn = 
			new HashMap<ConjunctiveQuery, HashMap<Literal, ConjunctiveQuery>>();
		
		//ONLY CONSIDERING BI-PARTITIONING NOW
		Integer[] biPar = new Integer[subgoals.size()];
		for(int i=0;i<biPar.length;i++) biPar[i] = 0;
		
		if(subgoals.size() == 0){
			return null;
		}
		
		Integer ct = (int) Math.pow(2, subgoals.size());
		while(--ct >= 0){

						
			HashMap<Literal, ConjunctiveQuery> list = new HashMap<Literal, ConjunctiveQuery>();
			
			ConjunctiveQuery q1 = new ConjunctiveQuery();
			ConjunctiveQuery q2 = new ConjunctiveQuery();
			ConjunctiveQuery q = new ConjunctiveQuery();
			q.setHead(head);//q.head = head;
			
			HashSet<Term> varSet1 = new HashSet<Term>();
			HashSet<Term> varSet2 = new HashSet<Term>();
			HashSet<String> termName1 = new HashSet<String>();
			HashSet<String> termName2 = new HashSet<String>();
			FelixPredicate p1 = new FelixPredicate(FelixPredicate.getNextTmpPredicateName(), false);
			FelixPredicate p2 = new FelixPredicate(FelixPredicate.getNextTmpPredicateName(), false);
			
			HashMap<Term, Type> term2TypeMapping = new HashMap<Term, Type>();
			
			for(int i=0;i<biPar.length;i++){
				if(biPar[i] == 0){
					//q1.body.add(subgoals.get(i));
					q1.addBodyLit(subgoals.get(i));
					varSet1.addAll(subgoals.get(i).getTerms());
					for(int j=0;j<subgoals.get(i).getTerms().size();j++){
						term2TypeMapping.put(subgoals.get(i).getTerms().get(j), subgoals.get(i).getPred().getTypeAt(j));
					}
					
				}else{
					//q2.body.add(subgoals.get(i));
					q2.addBodyLit(subgoals.get(i));
					varSet2.addAll(subgoals.get(i).getTerms());
					for(int j=0;j<subgoals.get(i).getTerms().size();j++){
						term2TypeMapping.put(subgoals.get(i).getTerms().get(j), subgoals.get(i).getPred().getTypeAt(j));
					}
				}
			}
			
			// generate next 0-1 combination.
			biPar[0] ++;
			for(int i=0;i<biPar.length-1;i++){
				if(biPar[i] == 2){
					biPar[i] = 0;
					biPar[i+1] ++;
				}
			}
			
			if(q1.body.size() == 0){
				continue;
			}
			
			
			HashSet<String> allIntermediateTerms = new HashSet<String>();
			HashSet<String> allIntermediateTerms2 = new HashSet<String>();
			
			for(Term t : varSet1) allIntermediateTerms.add(t.var());
			for(Term t : varSet2) allIntermediateTerms2.add(t.var());
			allIntermediateTerms.retainAll(allIntermediateTerms2);
			for(Term t : head.getTerms()) allIntermediateTerms.add(t.var());
			allIntermediateTerms.addAll(additionalSelList);
			
			Literal q1Head = new Literal(p1, true);
			for(Term t : varSet1){
				if(t.isVariable() == true && allIntermediateTerms.contains(t.var())){
					if(!termName1.contains(t.var())){
						q1Head.appendTerm(t);
						termName1.add(t.var());
						if(typeMapping.containsKey(t.var())){
							p1.appendArgument(new Type(typeMapping.get(t.var())));
						}else{
							p1.appendArgument(term2TypeMapping.get(t));
						}
					}
				}
			}
			q1.setHead(q1Head);
			
			
			Literal q2Head =  new Literal(p2, true);
			for(Term t : varSet2){
				if(t.isVariable() == true && allIntermediateTerms.contains(t.var())){
					if(!termName2.contains(t.var())){
						q2Head.appendTerm(t);
						termName2.add(t.var());
						if(typeMapping.containsKey(t.var())){
							p2.appendArgument(new Type(typeMapping.get(t.var())));
						}else{
							p2.appendArgument(term2TypeMapping.get(t));
						}
					}
					
				}
			}
			q2.setHead(q2Head);
			
			if(q2.head.getTerms().size() == 0 && q1.head.getTerms().size() == 0){
				continue;
			}
			
			// full
			if(q2.head.getTerms().size() == 0){
				list.put(q1.head, q1);
				q.addBodyLit(q1.head);
				forReturn.put(q, list);
				continue;
			}
			
			list.put(q1.head, q1);
			list.put(q2.head, q2);
			
			q.addBodyLit(q1.head);
			q.addBodyLit(q2.head);
			
			forReturn.put(q, list);
			

		}
		
		
		return forReturn;
	}
}
