package felix.optimizer;

import java.util.HashSet;


import tuffy.mln.Literal;
import tuffy.util.ExceptionMan;
import tuffy.util.StringMan;

import felix.dstruct.FelixClause;
import felix.dstruct.FelixPredicate;
import felix.dstruct.FelixQuery;
import felix.dstruct.StatOperator;
import felix.dstruct.FelixPredicate.FPProperty;
import felix.dstruct.StatOperator.OPType;
import felix.operator.COREFOperator;
import felix.operator.CRFOperator;
import felix.operator.LROperator;
import felix.operator.TUFFYOperator;
import felix.parser.FelixCommandOptions;
import felix.util.FelixConfig;
import felix.util.FelixStringMan;
import felix.util.FelixUIMan;

/**
 * An object of OperatorSelector will partition the FelixQuery into
 * different StatOperators. Current version does not rely on the
 * cost model, however, we may need to do that in the future.
 * 
 * @author Ce
 *
 */
public class OperatorSelector {

	/**
	 * Cost model used to partition operators.
	 */
	CostModel cm;
	
	/**
	 * FelixQuery which will be partitioned.
	 */
	FelixQuery fq;
	
	/**
	 * Command line options.
	 */
	FelixCommandOptions options;
	
	/**
	 * The constructor.
	 * @param _fq
	 * @param _cm
	 * @param _options
	 */
	public OperatorSelector(FelixQuery _fq, CostModel _cm, FelixCommandOptions _options){
		cm = _cm;
		fq = _fq;
		options = _options;
	}
	
	/**
	 * Get partitioned operators.
	 * @return
	 */
	public HashSet<StatOperator> getOperators(){

		HashSet<StatOperator> opts = new HashSet<StatOperator>();
		
		FelixUIMan.println(0, 0, ">>> Parsing Operators...");
				
		int ctCoref = 0;
		int ctMLN = 0;
		int ctCRF = 0;
		int ctLR = 0;
		
		HashSet<FelixPredicate> processedPredicates = new HashSet<FelixPredicate>();
		HashSet<FelixPredicate> openPredicates = fq.getAllOpenPred();
		
		HashSet<FelixClause> consideredClauses = new HashSet<FelixClause>();
		
		if(FelixConfig.allRuleAsMLN == false){
		
			for(FelixPredicate fp : openPredicates){
				
				// pick coref rules
				if( options.noCOREF == false &&
					fp.hasProperty(FPProperty.SYMM) &&
					fp.hasProperty(FPProperty.REFLEX) &&
					fp.hasProperty(FPProperty.TRANS) && 
					//!fp.hasProperty(FPProperty.EMBED_WEIGHT_RULE) &&
					(fp.mustbe == null || fp.mustbe.equals(OPType.COREF)) ){
					
					FelixUIMan.println(0, 0, "    Add coref operator for {" + fp.getName() + "}");
					COREFOperator op = new COREFOperator(fq, FelixStringMan.toHashSet(fp), options);
					opts.add(op);
					
					for(FelixClause fc : fp.getRelevantClauses()){
						
						
						if(fp.getPropertyClauses(FPProperty.CHAIN_RECUR).contains(fc)){
							continue;
						}
						
						if(fp.getPropertyClauses(FPProperty.KEY_CONSTRAINT).contains(fc)){
							continue;
						}
						
						if(fp.getPropertyClauses(FPProperty.OTHER_RECUR).contains(fc)){
							continue;
						}

						if(fp.getPropertyClauses(FPProperty.OTHER_RECUR_WITHOTHER_OPENPRED).contains(fc)){
							continue;
						}
						
						
						// pure-evidence op assumption.
						// we need a better optimizer and scheduler for
						// other non-trival cases.
						boolean hasOtherOpen = false;
						for(Literal l : fc.getRegLiterals()){
							if(l.getPred().isClosedWorld() == false && 
									!l.getPred().getName().equals(fp.getName())){
								hasOtherOpen = true;
							}
						}
						if(hasOtherOpen){
							continue;
						}
						
						op.registerRelevantClause(fc);
						consideredClauses.add(fc);
					}
					op.sealDefinition();
					
					processedPredicates.add(fp);
					ctCoref ++;
				}
				
				// pick CRF and LR rules
				else if(fp.hasProperty(FPProperty.KEY_CONSTRAINT)){
					
					//crf
					if( options.noCRF == false &&
																				//TODO: CHECK CONSISTENT CHAIN RULE
						fp.getPropertyClauses(FPProperty.CHAIN_RECUR).size() >= 1 &&
						fp.hasProperty(FPProperty.OTHER_RECUR) == false &&
						(fp.mustbe == null || fp.mustbe.equals(OPType.CRF)) ){
						
						FelixUIMan.println(0, 0, "    Add crf operator for {" + fp.getName() + "}");
						CRFOperator op = new CRFOperator(fq, FelixStringMan.toHashSet(fp), options);
						opts.add(op);
						
						for(FelixClause fc : fp.getRelevantClauses()){
									
							if(fp.getPropertyClauses(FPProperty.KEY_CONSTRAINT).contains(fc) ||
									fp.getPropertyClauses(FPProperty.CHAIN_RECUR).contains(fc) ||
									fp.getPropertyClauses(FPProperty.NON_RECUR).contains(fc)){
								
								// pure-evidence op assumption.
								// we need a better optimizer and scheduler for
								// other non-trival cases.
								boolean hasOtherOpen = false;
								for(Literal l : fc.getRegLiterals()){
									if(l.getPred().isClosedWorld() == false && 
											!l.getPred().getName().equals(fp.getName())){
										hasOtherOpen = true;
									}
								}
								if(hasOtherOpen){
									continue;
								}
								
								op.registerRelevantClause(fc);
								consideredClauses.add(fc);
							}
							
						}
						op.sealDefinition();
						
						processedPredicates.add(fp);
						ctCRF ++;
					}
					//lr
					else if(options.noLR == false &&
							fp.hasProperty(FPProperty.CHAIN_RECUR) == false &&
							fp.hasProperty(FPProperty.OTHER_RECUR) == false &&
							(fp.mustbe == null || fp.mustbe.equals(OPType.LR)) ){
						
						FelixUIMan.println(0, 0, "    Add lr operator for {" + fp.getName() + "}");
						LROperator op = new LROperator(fq, FelixStringMan.toHashSet(fp), options);
						opts.add(op);
						
						for(FelixClause fc : fp.getRelevantClauses()){
							

							if(fp.getPropertyClauses(FPProperty.KEY_CONSTRAINT).contains(fc) ||
									fp.getPropertyClauses(FPProperty.NON_RECUR).contains(fc)){
								
								// pure-evidence op assumption.
								// we need a better optimizer and scheduler for
								// other non-trival cases.
								boolean hasOtherOpen = false;
								for(Literal l : fc.getRegLiterals()){
									if(l.getPred().isClosedWorld() == false && 
											!l.getPred().getName().equals(fp.getName())){
										hasOtherOpen = true;
									}
								}
								if(hasOtherOpen){
									continue;
								}
								
								op.registerRelevantClause(fc);
								consideredClauses.add(fc);
							}
							
						}
						op.sealDefinition();
						
						processedPredicates.add(fp);
						ctLR ++;
					}	
				}	
			}
		}
		
		HashSet<FelixPredicate> alsoInput = new HashSet<FelixPredicate>();
		// other predicates as MLN
		HashSet<FelixPredicate> mlnPredicates = new HashSet<FelixPredicate>();
		
		HashSet<String> openPredicatesInLastMLN = new HashSet<String>();
		
		for(FelixPredicate fp : openPredicates){
			
			// do not include predicate as output of other operator
			if(processedPredicates.contains(fp)){
				continue;
			}
			
			openPredicatesInLastMLN.add(fp.getName());	
		}
		
		
		for(FelixPredicate fp : openPredicates){
			
			// do not include predicate as output of other operator
			if(processedPredicates.contains(fp)){
				continue;
			}
						
			if(fp.mustbe != null && !fp.mustbe.equals(OPType.TUFFY)){
				continue;
			}
			
			processedPredicates.add(fp);
			
			// do not include predicate w/o any relevant clauses
			if(fp.getRelevantClauses().size() == 0){
				continue;
			}
			
			ctMLN ++;
			mlnPredicates.add(fp);
			
		}
		if(mlnPredicates.size() != 0 || fq.getAllClause().size() != consideredClauses.size()){
			TUFFYOperator op = new TUFFYOperator(fq, mlnPredicates, options);
			op.inputPredicates.addAll(alsoInput);
			op.setPrecedence(2);
			FelixUIMan.println(0, 0, "    Add mln operator for {" + 
					StringMan.join(",", FelixStringMan.colToStringArray(mlnPredicates))+"}");
			opts.add(op);
			
			for(FelixClause fc : fq.getAllClause()){
				if(consideredClauses.contains(fc)){
					continue;
				}
				op.registerRelevantClause(fc);
			}
			
			op.sealDefinition();
							
		}
		
		// if there are remanining predicates, this program is not decomposable.
		if(processedPredicates.size() != openPredicates.size()){
			ExceptionMan.die("The input program is not decomposable. In current version of Felix, please" +
					" check whether you use P_map correctly. This predicate can only be associated" +
					" with Coref predicate P. Please change its name if you use it as a regular predicate");
		}

		FelixUIMan.println(2, 0, "\t # Coref Operators: " + ctCoref + "\n" +
			"\t # MLN Operators: " + ctMLN + "\n" + 
			"\t # LR Operators: " + ctLR + "\n" + 
			"\t # CRF Operators: " + ctCRF + "\n");
		return opts;
	}
	
}







