package felix.optimizer;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;

import tuffy.mln.Literal;
import tuffy.mln.Predicate;
import tuffy.ra.Expression;
import tuffy.util.Config;


import felix.dstruct.ConcurrentOperatorsBucket;
import felix.dstruct.ExecutionPlan;
import felix.dstruct.FelixClause;
import felix.dstruct.FelixPredicate;
import felix.dstruct.FelixQuery;
import felix.dstruct.OperatorBucketGraph;
import felix.dstruct.StatOperator;
import felix.dstruct.StatOperator.OPType;
import felix.main.Felix;
import felix.parser.FelixCommandOptions;
import felix.util.FelixUIMan;

/**
 * The class of a scheduler, which takes as input a FelixQuery, and 
 * outputs an physical execution plan that is ready to be fed into {@link Executor}.
 * The
 * output of a scheduler, i.e., {@link ExecutionPlan}, can be regarded
 * as a description of the physical plan, which contains the execution
 * order of different statistical operators.
 * 
 * <br/>
 * TODO: + better cycle support for operator graph.
 * 
 * 
 * @author Ce Zhang
 *
 */
public class Scheduler {

	/**
	 * Felix object.
	 */
	Felix parentFelix;
	
	/**
	 * Felix query to be scheduled.
	 */
	FelixQuery fq;
	
	/**
	 * Command line options.
	 */
	FelixCommandOptions options;

	/**
	 * Dependency graph between different operator buckets.
	 */
	OperatorBucketGraph obg;
	
	/**
	 * Partition the rules into different operators.
	 * @param cm
	 * @return
	 */
	public HashSet<StatOperator> ruleDecomposition(CostModel cm){
		OperatorSelector os = new OperatorSelector(fq, cm, options);
		HashSet<StatOperator> ops = os.getOperators();
		return ops;
	}
	
	/**
	 * Partition the data into different parts. This method will partition
	 * one operator returned by {@link Scheduler#ruleDecomposition} into different ones,
	 * with each of them deals with different portions of data. These operators
	 * will be put into a ConcurrentOperatorsBucket, which means they can be 
	 * executed in parallel.
	 * @param ops
	 * @param cm
	 * @return
	 */
	public OperatorBucketGraph dataDecomposition(HashSet<StatOperator> ops, CostModel cm){
		
		OperatorBucketGraph obg = new OperatorBucketGraph();
		int dpart = Config.getNumThreads()-1;
		
		if(Config.getNumThreads() > 1){
			for(StatOperator op : ops){
				DataCracker1991 dc = new DataCracker1991();
				dc.decompose(op);
				
				if(dc.isDecomposable == true && (options.decomposeTuffy || op.type != OPType.TUFFY)){
					
					FelixUIMan.println(0, 0, ">>> Decomposing the following operator into " + dpart + " parts:");
					FelixUIMan.printobj(0, 1, op);
					ConcurrentOperatorsBucket bucket = new ConcurrentOperatorsBucket(options.marginal);
					
					for(int i=0;i<dpart;i++){
					
						StatOperator newOp = op.clone();
						newOp.dc = dc;
						
						HashMap<Predicate, Expression> toSig = new HashMap<Predicate, Expression>();
						for(FelixClause fc : op.allRelevantFelixClause){
							
							newOp.dPart = dpart;
							newOp.nPart = i;
							
							HashSet<Expression> e = dc.getExpressions(fc, null, dpart, i, false);
							newOp.clauseConstraints.put(fc, e);
							
							for(FelixPredicate fop : newOp.outputPredicates){
								
								Predicate _p = fop;
								if(_p.isClosedWorld() || toSig.containsKey(_p) ){
									continue;
								}
								
								HashSet<Expression> expForClause = dc.getExpressions(null, _p, dpart, i, true);
								
								if(expForClause.size() != 0){
									toSig.put(_p, expForClause.iterator().next());
								}
								
							}
						}
						newOp.dataCrackerSignature = toSig.toString();
						
						newOp.partitionedInto = dpart;
						
						FelixUIMan.printobj(2, 2, newOp);
						bucket.addOperator(newOp);
					}
					
					obg.addOperator(bucket);
					
				}else{
					
					FelixUIMan.println(0, 0, ">>> The following operator is not decomposable:\n\t" + op.toString() + "\n");
					
					ConcurrentOperatorsBucket bucket = new ConcurrentOperatorsBucket(options.marginal);
					bucket.addOperator(op);
					obg.addOperator(bucket);
				}
				
			}
		}else{
			// do not partition
			for(StatOperator op : ops){

				FelixUIMan.println(0, 0, ">>> Decomposing the following operator into " + dpart + " parts:");
				FelixUIMan.printobj(0, 1, op);
					
				ConcurrentOperatorsBucket bucket = new ConcurrentOperatorsBucket(options.marginal);
				bucket.addOperator(op);
				obg.addOperator(bucket);
			}
		}
		
		obg.parseDependency();
		
		//FelixUIMan.println(2, 0, obg.toString());
		
		return obg;
	}
	
	/**
	 * Returns list of operator buckets sorted by precedence.
	 * @param torank
	 * @return
	 */
	public ArrayList<ConcurrentOperatorsBucket> rank(Collection<ConcurrentOperatorsBucket> torank){
		ArrayList<ConcurrentOperatorsBucket> ret = new ArrayList<ConcurrentOperatorsBucket>();
		
		ret.addAll(torank);
		
		Collections.sort(ret, new Comparator<ConcurrentOperatorsBucket>(){

			@Override
			public int compare(ConcurrentOperatorsBucket o1,
					ConcurrentOperatorsBucket o2) {
				// TODO Auto-generated method stub
				return o1.precedence - o2.precedence;
			}
			
		});
		
		
		return ret;
	}
	
	/**
	 * Schedule the order of running the operators.
	 * @param obg
	 * @return
	 */
	public ExecutionPlan orderOperators(OperatorBucketGraph obg){
		ExecutionPlan ep = new ExecutionPlan();
		
		// generate the order of execution
		ArrayList<ConcurrentOperatorsBucket> HotList = new ArrayList<ConcurrentOperatorsBucket>();
		HashSet<ConcurrentOperatorsBucket> notFinished = new HashSet<ConcurrentOperatorsBucket>();
		notFinished.addAll(obg.getOperators());
		
		while(notFinished.isEmpty() != true){
			ConcurrentOperatorsBucket op = null;
			
			int lowest = 10000;
			int numberOfClause = 0;
			// pick one with lowest precedence and put it in the last
			for(ConcurrentOperatorsBucket aop : notFinished){
				if(lowest > aop.getPrecedence()){
					numberOfClause = aop.nStartingRule;
					lowest = aop.getPrecedence();
					op = aop;
				}
				
				if(lowest == aop.getPrecedence() && 
						aop.nStartingRule > numberOfClause){
					numberOfClause = aop.nStartingRule;
					lowest = aop.getPrecedence();
					op = aop;
				}
			}
			
			HotList = new ArrayList<ConcurrentOperatorsBucket>();
			HotList.add(op);
			
			while(HotList.isEmpty() != true){
			
				// TODO: CHANGE TO A SMARTER VERSION
				ConcurrentOperatorsBucket current = rank(HotList).get(0);
				HotList.remove(0);
				
				//put in the last position
				ep.addOperatorAfter(current);
				notFinished.remove(current);
				
				//put upstream opts in HotList, which will be added after current position
				//in next iterations
				HashSet<ConcurrentOperatorsBucket> upstreams = obg.upStreams.get(current);
				HotList.addAll(upstreams);
				HotList.retainAll(notFinished);
								
				//put downstream operators before current position
				HashSet<ConcurrentOperatorsBucket> downstreams = obg.downStreams.get(current);
				for(ConcurrentOperatorsBucket ooo : rank(downstreams)){
					if(notFinished.contains(ooo) && !HotList.contains(ooo)){
						ep.addOperatorBefore(ooo);
					}
					if(HotList.contains(ooo) && ooo.getPrecedence() == current.getPrecedence()
							&& ooo.nStartingRule < current.nStartingRule){
						ep.addOperatorBefore(ooo);
						HotList.remove(ooo);
						notFinished.remove(ooo);
					}
					if(!HotList.contains(ooo)){
						notFinished.remove(ooo);
					}
				}

			}
		}
		
		return ep;
	}
	
	/**
	 * Returns operator bucket graph.
	 * @return
	 */
	public OperatorBucketGraph getOperatorBucketGraph(){
		return obg;
	}
	
	/**
	 * Parses common predicates among buckets.
	 * @param _ep
	 */
	public void parseCommonPredicates(ExecutionPlan _ep){
		
		HashMap<FelixPredicate, HashSet<ConcurrentOperatorsBucket>>
			commonPredicates = new HashMap<FelixPredicate, HashSet<ConcurrentOperatorsBucket>>();
		
		for(ConcurrentOperatorsBucket cob : _ep.operators){
						
			for(FelixPredicate fp : cob.commonCandidates){
				if(fp.isClosedWorld() && !fp.isCorefMap()) continue;
				FelixPredicate _fp = fp;
				//if(fp.isCorefMap()){
				//	_fp = fp.getOriCorefPredicate();
				//}
				if(!commonPredicates.containsKey(_fp)){
					commonPredicates.put(_fp, new HashSet<ConcurrentOperatorsBucket>());
				}
				commonPredicates.get(_fp).add(cob);
				
				if(_fp.isCorefPredicate && cob.outputPredicates.contains(_fp)){
					_fp = fp.corefMAPPredicate;
					if(!commonPredicates.containsKey(_fp)){
						commonPredicates.put(_fp, new HashSet<ConcurrentOperatorsBucket>());
					}
					commonPredicates.get(_fp).add(cob);
				}
					
			}
		}
		
		for(FelixPredicate fp : commonPredicates.keySet()){
			if(commonPredicates.get(fp).size() > 1){
				_ep.dd_CommonPredicates.add(fp);
			}
		}
		
		_ep.dd_Predicate2OperatorBucket = commonPredicates;
		
		for(FelixPredicate fp : _ep.dd_CommonPredicates){
			for(ConcurrentOperatorsBucket cob : _ep.dd_Predicate2OperatorBucket.get(fp)){
				cob.addCommonOutput(fp);
			}
		}
		
	}
	
	/**
	 * The entry of this optimizer.
	 * @return
	 */
	public ExecutionPlan schedule(){
		
		// decompose the whole MLN into different operators
		HashSet<StatOperator> ops = this.ruleDecomposition(null);

		// decompose each operators into smaller operators dealing
		// with different portion of data.
		obg = this.dataDecomposition(ops, null);
		
		// optimize execution plan.
			//note that, in the future we way want to merge 1) operator ordering and 2) DMO optimization part
			//into an unified optimization model. however, currently, we only worry about them separately.
		// therefore, current assumption is, the ordering of operators does not rely on the cost model.
		ExecutionPlan ep = this.orderOperators(obg);
				
		if(options.useDualDecomposition){
			this.parseCommonPredicates(ep);
		}
		
		FelixUIMan.println(">>> Serialized Execution Plan:");
		FelixUIMan.printobj(0, 1, ep);

		
		CostModel cm = new CostModel(this.parentFelix);
		
		ep.setCostModel(cm);
		
		//DMOOptimizer dmoo = new DMOOptimizer(cm);
		//this.optimizeDMO(ep, dmoo);
		//dmoo.close();
	
		return ep;
		
	}
	
	/**
	 * The constructor.
	 * @param _felix
	 * @param _fq
	 * @param _options
	 */
	public Scheduler(Felix _felix, FelixQuery _fq, FelixCommandOptions _options){
		parentFelix = _felix;
		fq = _fq;
		options = _options;
	}
	
	
	
}



