package felix.parser;

import java.io.FileInputStream;
import java.io.InputStream;
import java.util.zip.GZIPInputStream;

import org.antlr.runtime.ANTLRInputStream;
import org.antlr.runtime.ANTLRStringStream;
import org.antlr.runtime.CharStream;
import org.antlr.runtime.CommonTokenStream;

import felix.dstruct.FelixQuery;

import tuffy.util.ExceptionMan;

/**
 * Class that parses program/evidence files. 
 * @author Ce Zhang
 *
 */
public class FelixInputParser {
	
	FelixQuery mln;
	
	/**
	 * The constructor.
	 * @param amln
	 */
	public FelixInputParser(FelixQuery amln){
		mln = amln;
	}
	
	/**
	 * Parses given program file.
	 * @param fprog
	 */
	public void parseProgramFile(String fprog){
		FELIXParser parser = new FELIXParser(getTokens(fprog));
		parser.ml = this.mln;
		try {
			parser.definitions();
		} catch (Exception e) {
			mln.closeFiles();
			ExceptionMan.handle(e);
		}
	}

	/**
	 * Parses given evidence file.
	 * @param fevid
	 */
	public void parseEvidenceFile(String fevid){
		FELIXParser parser = new FELIXParser(getTokens(fevid));
		parser.ml = this.mln;
		try {
			parser.evidenceList();
		} catch (Exception e) {
			mln.closeFiles();
			ExceptionMan.handle(e);
		}
	}

	/**
	 * Parses given evidence string.
	 * @param chunk
	 * @param lineOffset
	 */
	public void parseEvidenceString(String chunk, long lineOffset){
		ANTLRStringStream input = new ANTLRStringStream(chunk);
		FELIXLexer lexer = new FELIXLexer(input);
		CommonTokenStream tokens = new CommonTokenStream(lexer);
		FELIXParser parser = new FELIXParser(tokens);
		parser.lineOffset = lineOffset;
		parser.ml = this.mln;
		try {
			parser.evidenceList();
			parser.reset();
			tokens.reset();
			lexer.reset();
			input.reset();
			parser.ml = null;
			parser = null;
		} catch (Exception e) {
			mln.closeFiles();
			ExceptionMan.handle(e);
		}
	}

	/**
	 * Parses query file.
	 * @param fquery
	 */
	public void parseQueryFile(String fquery){
		FELIXParser parser = new FELIXParser(getTokens(fquery));
		parser.ml = this.mln;
		try {
			parser.queryList();
		} catch (Exception e) {
			mln.closeFiles();
			ExceptionMan.handle(e);
		}
	}
	
	/**
	 * Parses query atom list.
	 * @param queryAtoms
	 */
	public void parseQueryCommaList(String queryAtoms){
		CharStream input = new ANTLRStringStream(queryAtoms);
		FELIXLexer lexer = new FELIXLexer(input);
		CommonTokenStream tokens = new CommonTokenStream(lexer);
		FELIXParser parser = new FELIXParser(tokens);
		parser.ml = this.mln;
		try {
			parser.queryCommaList();
		} catch (Exception e) {
			mln.closeFiles();
			ExceptionMan.handle(e);
		}
	}
	
	/**
	 * Parses program file(s) from given string.
	 * @param strProgs
	 */
	public void parseProgramFileFromString(String strProgs){
		CharStream input = new ANTLRStringStream(strProgs);
		FELIXLexer lexer = new FELIXLexer(input);
		CommonTokenStream tokens = new CommonTokenStream(lexer);
		FELIXParser parser = new FELIXParser(tokens);
		parser.ml = this.mln;
		try {
			parser.definitions();
		} catch (Exception e) {
			mln.closeFiles();
			ExceptionMan.handle(e);
		}
	}

	/**
	 * Returns token stream for given file.
	 * @param fname
	 * @return
	 */
	private CommonTokenStream getTokens(String fname){
		try {
			InputStream is;
			FileInputStream fis = new FileInputStream(fname);
			if(fname.toLowerCase().endsWith(".gz")){
				is = new GZIPInputStream(fis);
			}else{
				is = fis;
			}
			ANTLRInputStream input = new ANTLRInputStream(is);
			FELIXLexer lexer = new FELIXLexer(input);
			CommonTokenStream tokens = new CommonTokenStream(lexer);
			is.close();
			return tokens;
		} catch (Exception e) {
			ExceptionMan.handle(e);
		}
		return null;
	}

}
