package felix.test;

import static org.junit.Assert.*;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.regex.Pattern;

import org.junit.BeforeClass;
import org.junit.Test;

import tuffy.db.RDB;
import tuffy.util.Config;
import tuffy.util.FileMan;
import tuffy.util.UIMan;

import felix.main.Felix;
import felix.parser.FelixCommandOptions;
import felix.util.FelixConfig;
import felix.util.FelixUIMan;

/**
 * Test using evidence supplied in files.
 * @author Josh Slauson
 *
 */
public class IOFileTest {

	public static RDB db;

	/**
	 * Perform required setup for this test.
	 */
	@BeforeClass
	public static final void setUp() {
		// delete existing output files
		FileMan.removeFile("test/felix/output/out.txt_IOFileTest_simpleTest");
		FileMan.removeFile("test/felix/output/out.txt_IOFileTest_missingEvidence");
		
		FelixConfig.overrideID();
		
		UIMan.parseConfigFile(Config.path_conf);
		
		if(FelixConfig.evidDBSchema != null){
			FelixConfig.db_schema = FelixConfig.evidDBSchema;
		}

		db = RDB.getRDBbyConfig();
		db.schema = FelixConfig.db_schema;
		db.resetSchema(FelixConfig.db_schema);
		
		if(!db.isSchemaExists("felix_test")) {
			String query = "CREATE SCHEMA felix_test";
			db.execute(query);
		}

	}

	/**
	 * Check if evidence was correctly loaded into a database table.
	 * @param expect
	 */
	public void checkEvidence(String[] expect) {
		// Check if evidence table exists
		db = RDB.getRDBbyConfig(FelixConfig.db_schema);
		long count = db.countTuples("pred_evid");
		System.out.println("Checking evidence count: " + count + " = " + expect.length/2 + "?");
		assertTrue(count == expect.length/2);
		
		// Check evidence table contents
		String query = "SELECT c1.string AS a1, c2.string AS b2 FROM pred_evid e, constants c1, constants c2 WHERE e.a1 = c1.id AND e.b2 = c2.id";
		ResultSet rs = db.query(query);
		
		try {
			while(rs != null && rs.next()) {
				String a = rs.getString("a1");
				String b = rs.getString("b2");
				
				System.out.println("Checking evidence contents: " + a + ", " + b);
				
				boolean match = false;
				for(int i = 0; i + 1 < expect.length; i+=2) {
					if(a.equals(expect[i])) {
						assertTrue(b.equals(expect[i+1]));
						match = true;
					}
				}
				if(!match) {
					System.out.println("\tNo match");
				}
				assertTrue(match);
			}
			
			db.close();
		} catch (SQLException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			assertTrue(false);
		}

	}
	
	/**
	 * Checks if output file contains expected results.
	 * @param output
	 * @param expect
	 */
	public void checkOutput(String output, String[] expect) {
		ArrayList<String> lines = FileMan.getLines(output);
		System.out.println("Checking output size: " + lines.size() + " = " + expect.length/2 + "?");
		assertTrue(lines.size() == expect.length/2);
		
		for(int i = 0; i < lines.size(); i++) {
			System.out.println("Checking output: " + lines.get(i));

			boolean match = false;
			for(int j = 0; j + 1 < expect.length; j+=2) {
				if(lines.get(i).equals("fact(\"" + expect[j] + "\", \"" + expect[j+1] + "\")")) {
					match = true;
				}
			}
			if(!match) {
				System.out.println("\tNo match");
			}
			assertTrue(match);
		}

	}

	/**
	 * Simple smoke test.
	 */
	@Test
	public void simpleTest() {
		try {
			String output = "test/felix/output/out.txt_IOFileTest_simpleTest";
			String[] args = {"-e", "test/felix/io_test/evid.db", "-i", "test/felix/io_test/prog.mln",
					"-o", output, "-queryFile", "test/felix/io_test/query.db", "-keepData"};
			
			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
			new Felix().run(options);

			String[] expect = {"1", "2", "4", "5"};
			checkEvidence(expect);
			checkOutput(output, expect);
			
		} catch(Exception e) {
			e.printStackTrace();
			assertTrue(false);
		}
	}


// NOTE: correctly handles error
//	@Test
//	public void extraEvidence() {
//		try {
//			String output = "test/felix/output/out.txt_IOFileTest_extraEvidence";
//			String[] args = {"-e", "test/felix/io_test/evid2.db", "-i", "test/felix/io_test/prog.mln",
//					"-o", output, "-queryFile", "test/felix/io_test/query.db", "-keepData"};
//			
//			FelixConfig.overrideID();
//			FelixCommandOptions options = FelixUIMan.parseCommand(args);
//			new Felix().run(options);
//			
//			String[] expect = {"1", "2", "4", "5"};
//			checkEvidence(expect);
//			checkOutput(output, expect);
//			
//		} catch(Exception e) {
//			e.printStackTrace();
//			assertTrue(false);
//		}
//	}

// NOTE: correctly handles error
//	@Test
//	public void noEvidence() {
//		try {
//			String output = "test/felix/output/out.txt_IOFileTest_noEvidence";
//			String[] args = {"-e", "test/felix/io_test/evid.db", "-i", "test/felix/io_test/prog3.mln",
//					"-o", output, "-queryFile", "test/felix/io_test/query.db", "-keepData"};
//			
//			FelixConfig.overrideID();
//			FelixCommandOptions options = FelixUIMan.parseCommand(args);
//			new Felix().run(options);
//			
//			String[] expect = {};
//			checkEvidence(expect);
//			checkOutput(output, expect);
//			
//		} catch(Exception e) {
//			e.printStackTrace();
//			assertTrue(false);
//		}
//	}
	
	/**
	 * Test using a evidence file that doesn't exist.
	 */
	@Test
	public void missingEvid() {
		try {
			
			
			String output = "test/felix/output/out.txt_IOFileTest_missingEvidence";
			String[] args = {"-e", "test/felix/io_test/evid.db-missing", "-i", "test/felix/io_test/prog.mln",
					"-o", output, "-queryFile", "test/felix/io_test/query.db", "-keepData"};
			
			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);

			new Felix().run(options);
			
			// just assume all goes well if we don't throw an exception
		} catch (Exception e) {
			e.printStackTrace();
			assertTrue(false);
		}
	}

//	@Test
//	public void incorrectEvidence() {
//		try {
//			String output = "test/felix/output/out.txt_IOFileTest_incorrectEvidence";
//			String[] args = {"-e", "test/felix/io_test/evid.db-incorrect", "-i", "test/felix/io_test/prog.mln",
//					"-o", output, "-queryFile", "test/felix/io_test/query.db", "-keepData"};
//			
//			FelixConfig.overrideID();
//			FelixCommandOptions options = FelixUIMan.parseCommand(args);
//
//			new Felix().run(options);
//		} catch (Exception e) {
//			e.printStackTrace();
//			assertTrue(false);
//		}
//	}
}
