package tuffy.ground.partition;



import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;


import org.postgresql.PGConnection;

import tuffy.db.RDB;
import tuffy.ground.Grounding;
import tuffy.mln.MarkovLogicNetwork;
import tuffy.util.Config;
import tuffy.util.ExceptionMan;
import tuffy.util.UIMan;
import tuffy.util.UnionFind;

/**
 * Utilities for partitioning the MRF generated by the grounding process.
 *
 */
public class Partitioning {
	private MarkovLogicNetwork mln = null;
	private RDB db = null;

	/**
	 * Construct a partitioning worker based on the grounding
	 * result.
	 * @param g the grounding worker
	 */
	public Partitioning(Grounding g){
		mln = g.getMLN();
		db = mln.getRDB();
	}

	
	private double perAtomWeight(){
		return 
		16.0 // MRF collections
		+ 48.0; // GAtom
	}
	
	private double perClauseSharedWeight(int nlits){
		return 
		(4.0 // atom-clause index
		+ 4.0 // unsat
		+ 32.0 + nlits * 4 // GClause object
		) / nlits; 
	}
	
	
	/**
	 * Agglomeratively cluster the atoms in the MRF into partitions
	 * while heuristically minimizing the (weighted) cut size. The heuristic
	 * is to scan through all clauses in the descending-abs(weight) order.
	 * 
	 * @param ramBudgetPerPartition size bound of a partition, roughly proportional to
	 * the number of atoms and clauses in each partition
	 * 
	 * @return a partitioning scheme
	 */
	public PartitionScheme partitionAtoms(double ramBudgetPerPartition){
		UIMan.verbose(1, ">>> Partitioning atoms...");
		double maxPartRAM = ramBudgetPerPartition;
		try {
			db.disableAutoCommitForNow();
			// read in set of atoms
			ArrayList<Integer> atoms = new ArrayList<Integer>();
			String sql = "SELECT atomID FROM " + mln.relAtoms;
			ResultSet rs = db.query(sql);
			while(rs.next()){
				atoms.add(rs.getInt(1));
			}
			rs.close();
			
			// the memory usage amortized onto individual atoms
			HashMap<Integer, Double> atomWts = new HashMap<Integer, Double>();
			for(int a : atoms){
				atomWts.put(a, perAtomWeight());
			}
			sql = "SELECT lits FROM " + mln.relClauses;
			rs = db.query(sql);
			while(rs.next()){
				Integer[] lits = (Integer[])rs.getArray("lits").getArray();
				double delta = perClauseSharedWeight(lits.length);
				for(int lit : lits){
					int a = Math.abs(lit);
					double w = atomWts.get(a) + delta;
					atomWts.put(a, w);
				}
			}
			rs.close();

			// first scan, compute components and parts
			UnionFind<Integer> ufpart = new UnionFind<Integer>();
			UnionFind<Integer> ufcomp = new UnionFind<Integer>();
			ufpart.makeUnionFind(atoms, atomWts);
			ufcomp.makeUnionFind(atoms, atomWts);
			
			sql = "SELECT lits FROM " + mln.relClauses +
			" ORDER BY ABS(weight) DESC";
			rs = db.query(sql);
			while(rs.next()){
				Integer[] lits = (Integer[])rs.getArray("lits").getArray();
				if(lits.length <= 1) continue;
				double afterSize = 0;
				HashSet<Integer> roots = new HashSet<Integer>();
				for(int lit : lits){
					Integer root = ufpart.getRoot(Math.abs(lit));
					if(!roots.contains(root)){
						afterSize += ufpart.clusterWeight(root);
						roots.add(root);
					}
				}
				Integer c1 = Math.abs(lits[0]);
				for(int i=1; i<lits.length; i++){
					ufcomp.union(c1, Math.abs(lits[i]));
					//if(ProbMan.testChance(1-afterSize/maxPartSize))
					if(afterSize <= maxPartRAM){
						ufpart.union(c1, Math.abs(lits[i]));
					}
				}
			}
			rs.close();
			
			// allocate result data structures
			HashMap<Integer, Component> compMap = new HashMap<Integer, Component>();
			for(Integer r : ufcomp.getRoots()){
				int size = ufcomp.clusterSize(r);
				double wt = ufcomp.clusterWeight(r);
				Component comp = new Component();
				comp.numAtoms = size;
				comp.ramSize = wt;
				comp.rep = r;
				compMap.put(r, comp);
			}

			HashMap<Integer, Partition> partMap = new HashMap<Integer, Partition>();
			for(Integer r : ufpart.getRoots()){
				int size = ufpart.clusterSize(r);
				double wt = ufpart.clusterWeight(r);
				Partition part = new Partition();
				part.numAtoms = size;
				part.ramSize = wt;
				partMap.put(r, part);
				Component comp = compMap.get(ufcomp.getRoot(r));
				comp.parts.add(part);
				part.parentComponent = comp;
			}
			
			// second scan, aggregate size info
			sql = "SELECT lits, weight FROM " + mln.relClauses;;
			rs = db.query(sql);
			while(rs.next()){
				Integer[] lits = (Integer[])rs.getArray("lits").getArray();
				double weight = Math.abs(rs.getDouble("weight"));
				Component comp = compMap.get(ufcomp.getRoot(Math.abs(lits[0])));
				comp.totalWeight += weight;
				comp.numClauses ++;
				comp.numPins += lits.length;
				HashSet<Partition> pset = new HashSet<Partition>();
				for(int lit : lits){
					int a = Math.abs(lit);
					Partition part = partMap.get(ufpart.getRoot(a));
					if(!pset.contains(part)){
						pset.add(part);
						part.numIncidentClauses ++;
					}
				}
				if(pset.size() > 1){
					comp.totalCutWeight += weight;
					comp.numCutClauses ++;
				}
			}
			rs.close();
			db.restoreAutoCommitState();
			
			// assign IDs to components and parts
			ArrayList<Component> comps = new ArrayList<Component>(compMap.values());
			Collections.sort(comps);
			int partID = 0;
			for(int i=0; i<comps.size(); i++){
				Component comp = comps.get(i);
				comp.id = i+1;
				Collections.sort(comp.parts);
				for(int j=0; j<comp.parts.size(); j++){
					Partition part = comp.parts.get(j);
					part.id = ++partID;
				}
			}
			PartitionScheme pmap = new PartitionScheme(comps);
			
			// output clustering results
			File fbuf = new File(Config.getLoadingDir(), "partition_map" + mln.getID());
			BufferedWriter writer = new BufferedWriter(new OutputStreamWriter
					(new FileOutputStream(fbuf),"UTF8"));	
			for(int a : atoms){
				int compid = compMap.get(ufcomp.getRoot(a)).id;
				int partid = partMap.get(ufpart.getRoot(a)).id;
				writer.append(a + " , " + compid + " , " + partid + "\n");
			}
			writer.close();
			db.dropTable(mln.relAtomPart);
			sql = "CREATE TABLE " + mln.relAtomPart + "(atomID INT, compID INT, partID INT)";
			db.update(sql);
			FileInputStream in = new FileInputStream(fbuf);
			PGConnection con = (PGConnection)db.getConnection();
			sql = "COPY " + mln.relAtomPart + " FROM STDIN CSV";
			con.getCopyAPI().copyIn(sql, in);
			in.close();
			return pmap;
		} catch (Exception e) {
			ExceptionMan.handle(e);
		}
		return null;
	}

	public PartitionScheme partitionMRF(double maxPartitionSize){
		PartitionScheme pmap = partitionAtoms(maxPartitionSize);
		partitionClauses(pmap);
		return pmap;
	}
	
	/**
	 * Given a partitioning scheme, partition the data accordingly.
	 * 
	 */
	public void partitionClauses(PartitionScheme pmap){
		int numParts = pmap.numParts();
		String sql;
		UIMan.verbose(1, ">>> Partitioning the MRF into " + numParts + " parts...");
		db.dropTable(mln.relClausePart);
		sql = "CREATE TABLE " + mln.relClausePart +
		"(cid INT, partID INT, compID INT)";
		db.update(sql);

		sql = "INSERT INTO " + mln.relClausePart + "(cid, partID, compID) " +
		"SELECT rc.cid, ra.partID, ra.compID FROM " + 
		mln.relClauses + " rc, " + mln.relAtomPart + " ra " +
		" WHERE ABS(_random_element(rc.lits))=ra.atomID";
		db.update(sql);

	}

}
