package tuffy.infer;


import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;


import org.postgresql.PGConnection;

import tuffy.db.RDB;
import tuffy.infer.ds.GAtom;
import tuffy.infer.ds.GClause;
import tuffy.mln.MarkovLogicNetwork;
import tuffy.mln.Predicate;
import tuffy.mln.Type;
import tuffy.util.Config;
import tuffy.util.ExceptionMan;
import tuffy.util.FileMan;
import tuffy.util.StringMan;
import tuffy.util.UIMan;

/**
 * Methods for moving data around between the RDBMS and memory,
 * as well as writing inference results to files.
 */
public class DataMover {
	RDB db;
	MarkovLogicNetwork mln;
	
	
	public DataMover(MarkovLogicNetwork mln){
		this.mln = mln;
		this.db = mln.getRDB();
	}
	

	/**
	 * Load the truth table of atoms from the database.
	 */
	public HashMap<Integer, Boolean> loadTruthTable(String relAtoms){
		HashMap<Integer, Boolean> map = new HashMap<Integer, Boolean>();
		String sql = "SELECT atomID, truth FROM " + relAtoms;
		ResultSet rs = db.query(sql);
		try {
			while(rs.next()){
				int aid = rs.getInt(1);
				boolean truth = rs.getBoolean(2);
				map.put(aid, truth);
			}
			rs.close();
		} catch (SQLException e) {
			ExceptionMan.handle(e);
		}
		return map;
	}

	public MRF loadMrfFromDb(String relAtoms, String relClauses){
		MRF mrf = new MRF(mln);
		loadMrfFromDb(mrf, relAtoms, relClauses);
		return mrf;
	}

	
	/**
	 * Load the entire grounding result into memory as an MRF.
	 * Also build the atom-clause index.
	 */
	public MRF loadMrfFromDb(MRF mrf, String relAtoms, String relClauses){
		mrf.ownsAllAtoms = true;
		try {
			db.disableAutoCommitForNow();
			// load atoms
			String sql = "SELECT atomID, truth FROM " + relAtoms;
			ResultSet rs = db.query(sql);
			while(rs.next()){
				GAtom n = new GAtom(rs.getInt("atomID"));
				n.truth = rs.getBoolean("truth");
				/*
				if(rs.getInt("keyID") != -1){
					mrf.keyBlock.pushGAtom(rs.getInt("keyID"), n);
				}
				*/
				mrf.atoms.put(n.id, n);
				mrf.addAtom(n.id);
			}
			rs.close();
			System.gc();
			// load clauses
			sql = "SELECT * FROM " + relClauses;
			rs = db.query(sql);
			while(rs.next()){
				GClause f = new GClause();
				f.parse(rs);
				mrf.clauses.add(f);
			}
			rs.close();

			System.gc();
			db.restoreAutoCommitState();
			mrf.buildIndices();
		} catch (Exception e) {
			ExceptionMan.handle(e);
		}
		return mrf;
	}


	/**
	 * Flush atoms states to the atom table.
	 */
	public void flushAtomStates(Collection<GAtom> atoms, String relAtoms){
		try {
			String fout = FileMan.getUniqueFileNameAbsolute();
			BufferedWriter writer = 
				new BufferedWriter(new OutputStreamWriter
						(new FileOutputStream(fout),"UTF8"), 50*1024*1024);
			
			if(!Config.snapshot_mode){
				for(GAtom n : atoms){
					writer.append(n.id + ", " + n.truth + ", " + n.prob + "\n");
				}
			}else{
				for(GAtom n : atoms){
					writer.append(n.id + ", " + n.truth + ", " + ((double)n.tallyTrue)/Config.currentSampledNumber + "\n");
				}
			}
			
			writer.close();
			String relt = "tmp_ptruth";
			String sql = "CREATE TEMP TABLE " + relt + 
			"(fatomID INT, ftruth BOOL, fprob FLOAT)";
			db.update(sql);

			sql = "COPY " + relt + " FROM STDIN CSV";
			FileInputStream in = new FileInputStream(fout);
			PGConnection con = (PGConnection)db.getConnection();
			con.getCopyAPI().copyIn(sql, in);
			in.close();
			
			FileMan.removeFile(fout);

			sql = "UPDATE " + relAtoms + " SET truth = ftruth, prob = fprob " +
			"FROM " + relt + " WHERE atomID = fatomID";
			db.update(sql);
			
			db.dropTable(relt);
		} catch (Exception e) {
			ExceptionMan.handle(e);
		}
	}
	
	/**
	 * Update predicate tables with the current states in the atom table.
	 * @param relAtoms the atom table
	 * @param asSoftEvidence mark updated tuples as soft evidence; otherwise as hard evidence
	 */
	public void copyStatesToPredicateTables(String relAtoms, boolean asSoftEvidence){
		String sql;
		for(Predicate p : mln.getAllPred()){
			if(asSoftEvidence){
				sql = "UPDATE " + p.getRelName() + " SET truth = NULL, prior = ra.prob, " + 
					"club = (CASE WHEN club<2 THEN club+2 ELSE club) " +
					"FROM " + relAtoms + " ra WHERE id = ra.tupleID and ra.predID = " + p.getID();
			}else{
				sql = "UPDATE " + p.getRelName() + " SET truth = ra.truth, prior = NULL, " + 
					"club = (CASE WHEN club<2 THEN club+2 ELSE club) " +
					"FROM " + relAtoms + " WHERE id = ra.tupleID and ra.predID = " + p.getID();
			}
			db.update(sql);
			db.vacuum(p.getRelName());
			db.analyze(p.getRelName());
		}
	}
	
	private String atomToString(Predicate p, ResultSet rs, HashMap<Integer,String> cmap){
		String line = p.getName() + "(";
		ArrayList<String> cs = new ArrayList<String>();
		try{
			for(int i=0; i<p.arity(); i++){
				String a = p.getArgs().get(i);
				Type t = p.getTypeAt(i);
				String v;
				if(!t.isNonSymbolicType()){
					v = cmap.get(rs.getInt(a));
				}else{
					v = rs.getString(a);
				}
				if(v.matches("^[0-9].*$") && !StringMan.escapeJavaString(v).contains(" ")){
					cs.add("" + StringMan.escapeJavaString(v) + "");
				}else{
					cs.add("\"" + StringMan.escapeJavaString(v) + "\"");
				}
			}
		}catch(Exception e){
			ExceptionMan.handle(e);
		}
		line += StringMan.commaList(cs) + ")";
		return line;
	}
	
	public void dumpTruthToFile(String relAtoms, String fout){
		HashMap<Integer,String> cmap = db.loadIdSymbolMapFromTable();
		try {
			String sql;
			BufferedWriter bufferedWriter = FileMan.getBufferedWriterMaybeGZ(fout);
			for(Predicate p : mln.getAllPred()) {
				if(p.isImmutable()){
					sql = "SELECT * FROM " + p.getRelName() +
					" WHERE club=3 AND truth";
				}else{
					sql = "SELECT * FROM " + p.getRelName() + " pt " + 
					" WHERE (pt.club=3 OR pt.club=1) AND ( pt.truth OR " +
					" pt.id IN (SELECT tupleID FROM " + relAtoms + " ra " +
					" WHERE ra.truth AND ra.predID = " + p.getID() +
					") )" +
					" ORDER BY " + StringMan.commaList(p.getArgs());
				}
				ResultSet rs = db.query(sql);
				while(rs.next()) {
					String satom = atomToString(p, rs, cmap);
					bufferedWriter.append(satom + "\n");
				}
				rs.close();
			}
			bufferedWriter.close();
		} catch (Exception e) {
			ExceptionMan.handle(e);
		}
	}
	
	public void updateOriTable(HashSet<String> relAtomsTables, Predicate p){

		String sql;
		try {
		
			for(String relAtoms : relAtomsTables){

				sql = "UPDATE " + p.getRelName() + " SET prior = ra.prob, truth = ra.truth " +
						"FROM " +relAtoms + " ra " +
				" WHERE id = ra.tupleID AND ra.predID = " + p.getID();
				db.execute(sql);
			}

		}catch (Exception e) {
			ExceptionMan.handle(e);
		}
	}


	public void dumpProbsToFile(String relAtoms, String fout){
		BufferedWriter bufferedWriter = FileMan.getBufferedWriterMaybeGZ(fout);
		HashMap<Integer,String> cmap = db.loadIdSymbolMapFromTable();
		int digits = 4;
		String sql;
		try {
			for(Predicate p : mln.getAllPredOrderByName()){			
				if(p.isImmutable()) continue;
				String atomTypeCond = (Config.mcsat_output_hidden_atoms ?					
						" " : " AND (club=1 OR club=3) ");
				String orderBy = " ORDER BY ";
				switch(Config.mcsat_output_order){
				case PRED_ARGS:
					orderBy += StringMan.commaList(p.getArgs());				
					break;
				case PROBABILITY:
					orderBy += " prob DESC ";
				}
				sql = "SELECT * FROM " + p.getRelName() + " pt, " + relAtoms + " ra " +
				" WHERE pt.id = ra.tupleID AND ra.predID = " + p.getID() +  
				" AND ra.prob >= " + Config.marginal_output_min_prob + " and ra.prob>0 " +
				atomTypeCond + orderBy;
				ResultSet rs = db.query(sql);
				while(rs.next()) {
					double prob = rs.getDouble("prob");
					double prior = rs.getDouble("prior");
					if(rs.wasNull()){
						prior = -1;
					}
					String satom = atomToString(p, rs, cmap);
					String line = null;
					if(Config.output_prolog_format){
						line = "tuffyPrediction(" + UIMan.decimalRound(digits, prob) +
							", " + satom + ").";
					}else{
						line = UIMan.decimalRound(digits, prob) + "\t" + satom;
					}
					if(Config.output_prior_with_marginals && prior >= 0){
						line += " // prior = " + UIMan.decimalRound(digits, prior);
						line += " ; delta = " + UIMan.decimalRound(digits, prob - prior);
					}
					bufferedWriter.append(line + "\n");
				}
				rs.close();
			}
			bufferedWriter.close();
		}catch (Exception e) {
			ExceptionMan.handle(e);
		}
	}


	/**
	 * Assign random truth values to the atoms.
	 */
	public void assignRandomTruth(String relAtoms){
		String sql = "UPDATE " + relAtoms + 
		" SET truth = (cast(random()*atomID as int)%2=0)";
		db.update(sql);
	}


	
}
