package tuffy.infer;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Random;

import tuffy.infer.ds.GAtom;
import tuffy.infer.ds.GClause;
import tuffy.infer.ds.KeyBlock;
import tuffy.learn.Learner;
import tuffy.mln.Clause;
import tuffy.mln.MarkovLogicNetwork;
import tuffy.util.Config;
import tuffy.util.DebugMan;
import tuffy.util.HashArray;
import tuffy.util.Timer;
import tuffy.util.UIMan;

/**
 * In-memory data structure representing an MRF.
 * 
 */
public class MRF {

	public static enum INIT_STRATEGY {ALL_FALSE, COIN_FLIP, GREEDY, COPY_LOW, TRAINING_DATA, NO_CHANGE};
	protected INIT_STRATEGY initStrategy = INIT_STRATEGY.COIN_FLIP;
	
	public INIT_STRATEGY getInitStrategy(){
		return initStrategy;
	}
	
	public void setInitStrategy(INIT_STRATEGY strategy){
		initStrategy = strategy;
	}
	
	public KeyBlock keyBlock = new KeyBlock();
	
	/**
	 * Map from GAtom ID to GAtom object.
	 */
	public HashMap<Integer, GAtom> atoms = new HashMap<Integer, GAtom>();
	
	public boolean ownsAllAtoms = false;
	private boolean usingBlocks = false;
	
	private HashSet<Integer> coreAtoms = new HashSet<Integer>();
	
	/**
	 * Array of all GClause objects in this MRF.
	 */
	public ArrayList<GClause> clauses = new ArrayList<GClause>();
	
	/**
	 * Array of unsatisfied GClauses under current atoms' truth setting.
	 */
	protected HashArray<GClause> unsat = new HashArray<GClause>();
	
	/**
	 * Index from GAtom ID to GClause.
	 */
	protected HashMap<Integer, ArrayList<GClause>> adj = new HashMap<Integer, ArrayList<GClause>>();

	int numCriticalNodesLocal = 0;

	int numClausesInCut = 0;
	
	double weightClausesInCut = 0;
	
	/**
	 * Atoms that have been flipped since last saving to low.
	 */
	protected HashSet<Integer> dirtyAtoms = new HashSet<Integer>();
	
	/**
	 * The total cost of this MRF under current atoms' truth setting.
	 */
	protected double totalCost = Double.MAX_VALUE;
	
	public double getCost(){
		return totalCost;
	}
	
	/**
	 * Lowest cost ever seen.
	 */
	public double lowCost = Double.MAX_VALUE;
	
	/**
	 * Number of GClauses that is selected, and therefore must be
	 * satisfied by next SampleSAT invocation of MCSAT.
	 */
	protected int totalAlive = 0;
	
	/**
	 * The flag indicating whether MCSAT is running WalkSAT 
	 * or SampleSAT.
	 */
	protected boolean sampleSatMode = false;
	
	int partID = 0;
	public long inferOps = 0;
	
	/**
	 * The MLN object.
	 */
	private MarkovLogicNetwork mln;
	
	
	public MarkovLogicNetwork getMLN(){
		return mln;
	}
	
	/**
	 * Reset low-cost to infinity.
	 */
	public void invalidateLowCost(){
		lowCost = Double.MAX_VALUE;
	}

	/**
	 * For research experiments!
	 * Split the MRF into multiple pieces by agglomerative clustering.
	 * Each piece contains up to 2/np of the total atoms.
	 * 
	 * @param np number of pieces
	 * @return the pieces, each as an individual MRF
	 */
	@SuppressWarnings("unused")
	private ArrayList<MRF> split(int np){
		Random r = new Random();
		ArrayList<MRF> pieces = new ArrayList<MRF>();
		ArrayList<Integer> as = new ArrayList<Integer>(atoms.keySet());
		HashSet<Integer> danglings = new HashSet<Integer>(atoms.keySet());
		Collections.shuffle(as);
		int imb = 1;// + r.nextInt(3);
		int maxs = (int) (atoms.size() * (1.0/np + r.nextDouble()*(1.0/np)));
		HashMap<Integer, MRF> amap = new HashMap<Integer, MRF>();
		int cur = 0;
		// assign seed nodes
		for(int i=0; i<np; i++){
			MRF m = new MRF(mln, i, atoms);
			m.ownsAllAtoms = false;
			pieces.add(m);
			m.numCriticalNodesLocal = 0;
			if(i==0){
				for(int k=imb; k>0; k--){
					int aid = as.get(cur++);
					GAtom a = atoms.get(aid);
					while(!a.critical()){
						aid = as.get(cur++);
						a = atoms.get(aid);
					}
					if(a.critical()){
						m.numCriticalNodesLocal ++;
					}
					amap.put(aid, m);
					m.addAtom(aid);
					danglings.remove(aid);
				}
			}else{
				int aid = as.get(cur++);
				GAtom a = atoms.get(aid);
				while(!a.critical()){
					aid = as.get(cur++);
					a = atoms.get(aid);
				}
				if(a.critical()){
					m.numCriticalNodesLocal ++;
				}
				amap.put(aid, m);
				m.addAtom(aid);
				danglings.remove(aid);
			}
		}
		for(int aid : atoms.keySet()){
			GAtom a = atoms.get(aid);
			a.cut = false;
		}
		
		// assign remaining nodes
		while(!danglings.isEmpty()){
			Iterator<Integer> it = danglings.iterator();
			while(it.hasNext()){
				int aid = it.next();
				HashSet<Integer> ns = getAtomNeighbors(aid);
				ArrayList<Integer> opts = new ArrayList<Integer>();
				for(int a : ns){
					if(amap.get(a) != null){
						opts.add(a);
					}
				}
				MRF m = pieces.get(r.nextInt(np));
				if(!opts.isEmpty()){
					m = amap.get(opts.get(r.nextInt(opts.size())));
					if(m.getCoreAtoms().size() > maxs){
						int ot = r.nextInt(np);
						while(ot == pieces.indexOf(m)){
							ot = r.nextInt(np);
						}
						m = pieces.get(ot);
					}
				}else{
					continue;
				}
				GAtom a = atoms.get(aid);
				if(a.critical()){
					m.numCriticalNodesLocal ++;
				}
				amap.put(aid, m);
				m.addAtom(aid);
				it.remove();
			}
		}
		
		double gcSize = 0;
		double gcWeight = 0;
		for(GClause gc : clauses){
			int aid = Math.abs(gc.lits[r.nextInt(gc.lits.length)]);
			MRF m = amap.get(aid);
			m.clauses.add(gc);
			gcSize ++;
			gcWeight += Math.abs(gc.weight);
		}
		ArrayList<Double> numLCs = new ArrayList<Double>();
		double cutSize = 0;
		double cutWeight = 0;
		for(MRF m : pieces){
			m.buildIndices();
			numLCs.add((double)m.numCriticalNodesLocal);
			cutSize += m.numClausesInCut;
			cutWeight += m.weightClausesInCut;
			DebugMan.log(m.numCriticalNodesLocal + ", ");
		}
		double gaSize = 0;
		double gaCritical = 0;
		double atomsCriticalCut = 0;
		double atomsCut = 0;
		for(GAtom a : atoms.values()){
			gaSize ++;
			if(a.critical()){
				gaCritical++;
			}
			if(a.cut){
				atomsCut++;
				if(a.critical()){
					atomsCriticalCut++;
				}
			}
		}

		DebugMan.log("np, " + np + ", ");
		double imbb = Collections.max(numLCs) / this.numCriticalNodesLocal;
		DebugMan.log("maxpart, " + UIMan.comma(imbb) + ", ");
		double atomsCutPortion = atomsCut / gaSize;
		DebugMan.log("atomsCut, " + UIMan.comma(atomsCutPortion) + ", ");
		double atomsCriticalCutPortion = atomsCriticalCut / gaCritical;
		DebugMan.log("atomsCriticalCut, " + UIMan.comma(atomsCriticalCutPortion) + ", ");
		double clausesCutPortion = cutSize / gcSize;
		DebugMan.log("clausesCut, " + UIMan.comma(clausesCutPortion) + ", ");
		double clausesWtCutPortion = cutWeight / gcWeight;
		DebugMan.log("clausesWtCut, " + UIMan.comma(clausesWtCutPortion) + ", ");
		
		return pieces;
	}
	
	/**
	 * For research experiments!
	 * Get all neighboring atoms of one atom.
	 * @param aid id of the core atom
	 * @return id of neighbors
	 */
	private HashSet<Integer> getAtomNeighbors(int aid){
		HashSet<Integer> ns = new HashSet<Integer>();
		for(GClause gc : adj.get(aid)){
			for(int lit : gc.lits){
				ns.add(Math.abs(lit));
			}
		}
		for(GClause gc : adj.get(-aid)){
			for(int lit : gc.lits){
				ns.add(Math.abs(lit));
			}
		}
		return ns;
	}
	
	
	/**
	 * Discard all data structures, in hope of facilitating faster GC.
	 */
	public void discard() {
		atoms = null;
		getCoreAtoms().clear();
		clauses.clear();
		unsat.clear();
		adj.clear();
	}
	
	/**
	 * Add an atom into this MRF.
	 * @param aid id of the atom
	 */
	public void addAtom(int aid){
		getCoreAtoms().add(aid);
	}
	
	/**
	 * Default constructor.
	 * Does not really do anything.
	 */
	public MRF(MarkovLogicNetwork mln){
		this.mln = mln;
	}

	/**
	 * 
	 * @param partID id of this MRF
	 * @param gatoms ground atoms
	 */
	public MRF(MarkovLogicNetwork mln, int partID, HashMap<Integer, GAtom> gatoms){
		this.partID = partID;
		atoms = gatoms;
	}

	/**
	 * Test if a given atom is "owned" by this MRF.
	 * An atom may not belong to this MRF if this MRF represents
	 * a partition of a component that has multiple partitions.
	 * 
	 * @param aid id of the atom
	 * 
	 */
	protected boolean ownsAtom(int aid){
		return ownsAllAtoms || getCoreAtoms().contains(aid);
	}
	

	/**
	 * If current truths have the lowest cost, save them.
	 * @param cost the current cost
	 */
	protected void saveLowTruth(double cost){
		if(!(cost < lowCost || (cost < 0.0001 && sampleSatMode))) return;
		if(cost < lowCost) lowCost = cost;
		if(dirtyAtoms.size() > 0){ // incremental dump
			for(int aid : dirtyAtoms){
				GAtom ga = atoms.get(aid);
				ga.lowTruth = ga.truth;
			}
			dirtyAtoms.clear();
		}else{ // force flushing
			if(ownsAllAtoms){
					for(GAtom n : atoms.values()){
						n.lowTruth = n.truth;
					}
			}else{
				for(int aid : getCoreAtoms()){
					GAtom n = atoms.get(aid);
					n.lowTruth = n.truth;
				}
			}
		}
	}


	private void saveTruthAsLow(){
		if(ownsAllAtoms){
				for(GAtom n : atoms.values()){
					n.lowTruth = n.truth;
				}
		}else{
			for(int aid : getCoreAtoms()){
				GAtom n = atoms.get(aid);
				n.lowTruth = n.truth;
			}
		}
	}

	
	/**
	 * Check if a given literal is true under current truth assignment.
	 * @param lit the literal represented as an integer
	 * 
	 */
	protected boolean isTrueLit(int lit){
		return (lit > 0 && atoms.get(lit).truth) 
		|| (lit < 0 && !atoms.get(-lit).truth);
	}

	/**
	 * Test if a clause is always true no matter how we flip flippable atoms.
	 * @param gc the clause
	 * 
	 */
	protected boolean isAlwaysTrue(GClause gc){
		int fixed = 0;
		for(int lit : gc.lits){
			if(!ownsAtom(Math.abs(lit))){
				if(isTrueLit(lit)){
					return true;
				}
				fixed ++;
			}
		}
		return fixed == gc.lits.length;
	}

	/**
	 * Fix the truth value of an atom.
	 * @param aid id of the atom
	 * @param t truth value to be fixed
	 */
	protected void fixAtom(int aid, boolean t){
		GAtom a = atoms.get(aid);
		a.truth = t;
		a.fixed = true;
	}



	/**
	 * Retain a subset of currently satisfied clauses, according
	 * to the sampling method of MC-SAT.
	 * 
	 * @return the number of retained clauses
	 */
	protected int retainSomeGoodClauses(){
		int numGood = 0;
		totalAlive = 0;
		for(GClause c : clauses){
			if(c.cost() == 0) numGood++;
			if(c.selectMCSAT()){
				c.dead = false;
				++ totalAlive;
			}else{
				c.dead = true;
			}
		}
		if(!Config.learning_mode)
			UIMan.verbose(2, "    Retained #clauses = " + UIMan.comma(totalAlive) + 
				" out of " + UIMan.comma(numGood) + " non-violated clauses (" + 
				UIMan.comma(clauses.size()) + " in total)");
		return totalAlive;
	}


	/**
	 * Unfix all atoms.
	 */
	protected void unfixAllAtoms(){
		for(GAtom a : atoms.values()){
			a.fixed = false;
		}
	}


	/**
	 * Assign the recorded low-cost truth values to current truth values.
	 */
	public void restoreLowTruth(){
		if(ownsAllAtoms){
			for(GAtom n : atoms.values()){
				n.truth = n.lowTruth;
			}
		}else{
			for(int aid : getCoreAtoms()){
				GAtom n = atoms.get(aid);
				n.truth = n.lowTruth;
			}
		}
	}

	/**
	 * Reset all clauses to be alive.
	 */
	protected void enableAllClauses(){
		totalAlive = clauses.size();
		for(GClause c : clauses){
			c.dead = false;
		}
	}

	/**
	 * Build literal-->clauses index.
	 * Used by WalkSAT.
	 */
	protected void buildIndices(){
		if(!adj.isEmpty()) return;
		adj.clear();
		for(GClause f : clauses){
			boolean isCut = false;
			for(int lit : f.lits){
				ArrayList<GClause> plist = adj.get(lit);
				if(plist == null){
					plist = new ArrayList<GClause>();
					adj.put(lit, plist);
				}
				plist.add(f);
				if(!ownsAllAtoms && !getCoreAtoms().contains(Math.abs(lit))){
					isCut = true;
				}
				if(Config.checkNumCriticalNodes){
					GAtom n = atoms.get(Math.abs(lit));
					if((lit > 0) == (f.weight > 0)){
						n.markWannaBeTrue();
					}else{
						n.markWannaBeFalse();
					}
				}
			}
			if(isCut){
				numClausesInCut++;
				weightClausesInCut += Math.abs(f.weight);
				for(int lit : f.lits){
					atoms.get(Math.abs(lit)).cut = true;
				}
			}
		}
	}
	
	private Random rand = new Random();
	
	/**
	 * Coin flipping.
	 * @param p probability of returning true
	 * 
	 */
	protected boolean testChance(double p){
		return rand.nextDouble() < p;
	}
	
	@SuppressWarnings("unused")
	private void testKeyConstraints(){
	
		HashMap<Integer, myInt> counter = new HashMap<Integer, myInt>();
		
		for(GAtom g : atoms.values()){

			Integer key = keyBlock.gatom2key.get(g);
			
			if(key == null){
				continue;
			}
			
			if(!counter.containsKey(key)){
				counter.put(key, new myInt(0));
			}
			if(g.truth == true){
				counter.get(key).addOne();
			}
		}
		
		for(Integer key : counter.keySet()){
			if(counter.get(key).value >= 2){
				System.err.print("DUPLICATION!!!");
			}
		}
		
	}

	/**
	 * Run WalkSAT.
	 * @param nTries number of tries
	 * @param nSteps number of steps per try
	 */
	public void inferWalkSAT(int nTries, long nSteps){
		if(Config.use_atom_blocking == false){
			this.inferWalkSATwithoutBlocking(nTries, nSteps);
		}else{
			this.inferWalkSATwithBlocks(nTries, nSteps);
		}
	}
	
	/**
	 * Run WalkSAT.
	 * @param nTries number of tries
	 * @param nSteps number of steps per try
	 */
	private void inferWalkSATwithoutBlocking(int nTries, long nSteps){
		if(atoms.size() == 0){
			return;
		}
		if(nSteps < 4) nSteps = 4;
		UIMan.println("    Running WalkSAT for " + nTries +
				" tries, " + UIMan.comma(nSteps) + " flips/try");
		if(adj.isEmpty()) buildIndices();
		Random rand = new Random();
		// low cost of periodic flushing
		for(long itry=1; itry<=nTries; itry++){
			UIMan.println("[Try #" + itry + "/" + nTries + "]");
			initMRF();
			for(long flip = 1; flip <= nSteps; flip++){
				// check if we have reached terminal condition
				if(unsat.isEmpty()){
					UIMan.println("Resolved!");
					saveLowTruth(totalCost);
					return;
				}
				// pick random unsat clause
				GClause lucky = unsat.getRandomElement();
				// pick an atom therein
				int picked = 0;
				if(lucky.isPositiveClause()){
					// a positive clause
					if(testChance(Config.walksat_random_step_probability)){
						// a random step
						picked = -1;
						while(picked == -1 || !ownsAtom(picked)){
							picked = Math.abs(lucky.lits[rand.nextInt(lucky.lits.length)]);
						}
					}else{
						// a greedy step
						double minDelta = Double.MAX_VALUE;
						int nrivals = 0;
						for(int lit : lucky.lits){
							int a = Math.abs(lit);
							if(!ownsAtom(a)) continue;
							if(atoms.get(a).delta() < minDelta){
								picked = a;
								minDelta = atoms.get(a).delta();
								nrivals = 1;
							}else if(atoms.get(a).delta() == minDelta){
								if(testChance(1.0/(++nrivals))){
									picked = a;
								}
							}
						}
					}
				}else{
					// a negative clause
					ArrayList<Integer> cands = new ArrayList<Integer>();
					for(int lit : lucky.lits){
						if(isTrueLit(lit) && ownsAtom(Math.abs(lit))){
							cands.add(Math.abs(lit));
						}
					}
					if(cands.size() == 1){
						picked = cands.get(0);
					}else{
						if(testChance(Config.walksat_random_step_probability)){
							// a random step
							picked = cands.get(rand.nextInt(cands.size()));
						}else{
							// a greedy step
							double minDelta = Double.MAX_VALUE;
							int nrivals = 0;
							for(int a : cands){
								if(atoms.get(a).delta() < minDelta){
									picked = a;
									minDelta = atoms.get(a).delta();
									nrivals = 1;
								}else if(atoms.get(a).delta() == minDelta){
									if(testChance(1.0/(++nrivals))){
										picked = a;
									}
								}
							}
						}
					}
				}
				
				//if(picked == 0) continue;
				inferOps ++;
				if(flip==1 || flip==nSteps || flip % (nSteps/4) == 0){
					UIMan.verbose(1, "At flip "+UIMan.comma(flip)+
							"/" + UIMan.comma(nSteps) + ";\tcost = " + UIMan.comma(lowCost));
				}
				if(!ownsAtom(picked)) continue;
				GAtom atom = atoms.get(picked);
				if(Config.avoid_breaking_hard_clauses && 
						atom.criticalForHardClauses()) continue;
				
				if(Config.apply_greedy_throttling && 
						atom.delta() > 0 && 
						testChance(0.95)){
					continue;
				}
				// flip this atom
				Timer.runStat.effectiveSteps++;
				
				//double oldTotal = totalCost;
				
				totalCost += atom.delta();
				atom.truth = !atom.truth;
				
				//double expected = totalCost;
				
				dirtyAtoms.add(picked);
				atom.invertDelta();
				if(totalCost < lowCost){
					saveLowTruth(totalCost);
				}
				// update stats
				ArrayList<GClause> tlfac, flfac;
				if(atom.truth){
					tlfac = adj.get(atom.id);
					flfac = adj.get(-atom.id);
					
					
				}else{
					flfac = adj.get(atom.id);
					tlfac = adj.get(-atom.id);
					
				}
				if(tlfac != null){
					for(GClause f : tlfac){
						++ f.nsat;
						int nsat = f.nsat;
						if(nsat==1){
							if(f.weight >= 0) {
								unsat.removeObj(f);
							}else{
								unsat.add(f);
							}
							for(int lit : f.lits){
								int a = Math.abs(lit);
								if(a==picked || !ownsAtom(a)) continue;
								atoms.get(a).revokeSatPotential(f);
							}
						}else if(nsat==2){
							for(int lit : f.lits){
								int a = Math.abs(lit);
								if(a==picked || !ownsAtom(a)) continue;
								GAtom n = atoms.get(a);
								if((lit>0) == n.truth && willChange(lit, f.lits)){
									n.revokeUnsatPotential(f);
									break;
								}
							}
						}
					}
				}
				if(flfac != null){
					for(GClause f : flfac){
						-- f.nsat;
						int nsat = f.nsat;
						if(nsat==0){
							if(f.weight <= 0) {
								unsat.removeObj(f);
							}else{
								unsat.add(f);
							}
							for(int lit : f.lits){
								int a = Math.abs(lit);
								if(a==picked || !ownsAtom(a)) continue;
								atoms.get(a).assignSatPotential(f);
							}
						}else if(nsat==1){
							for(int lit : f.lits){
								int a = Math.abs(lit);
								if(a==picked || !ownsAtom(a)) continue;
								GAtom n = atoms.get(a);
								if((lit>0) == n.truth && willChange(lit, f.lits)){
									n.assignUnsatPotential(f);
									break;
								}
							}
						}
					}
				}
				
				/*
				double newcost = this.recalcCostNotOverrideOldCost();
				if( newcost != expected){
					if(Math.abs(newcost - expected) > 0.1){
						System.out.println();
					}
					System.out.println(newcost - expected);
				}
				*/
			}
		}
	}
	
	public boolean willChange(int lit, int[] lits){
		for(int a : lits){
			if(a == -lit){
				return false;
			}
		}
		return true;
	}
	
	/**
	 * Run WalkSAT with blocks.
	 * @param nTries number of tries
	 * @param nSteps number of steps per try
	 */
	private void inferWalkSATwithBlocks(int nTries, long nSteps){
		
		//TODO: CHECK KEYBLOCK AND PARTITIONING
		
		if(atoms.size() == 0){
			return;
		}
		if(nSteps < 4) nSteps = 4;
		//System.out.print(".");
		UIMan.println(">>> Running WalkSAT for " + nTries +
				" tries, " + UIMan.comma(nSteps) + " flips/try");
		if(adj.isEmpty()) buildIndices();
		Random rand = new Random();
		// low cost of periodic flushing
		for(int itry=1; itry<=nTries; itry++){
			UIMan.println("[Try #" + itry + "/" + nTries + "]");
			initMRF();
			
			for(long flip = 1; flip <= nSteps; flip++){
				
			//	if(flip % 1000 == 0){
			//		this.testKeyConstraints();
			//	}
				
				// check if we have reached terminal condition
				if(unsat.isEmpty()){
					UIMan.println("resolved!");
					saveLowTruth(totalCost);
					return;
				}
				// pick random unsat clause
				GClause lucky = unsat.getRandomElement();
				// pick an atom therein
				int picked = 0;
				if(lucky.isPositiveClause()){
					// a positive clause
					if(testChance(Config.walksat_random_step_probability)){
						// a random step
						picked = -1;
						while(picked == -1 || !ownsAtom(picked)){
							picked = Math.abs(lucky.lits[rand.nextInt(lucky.lits.length)]);
						}
					}else{
						// a greedy step
						double minDelta = Double.MAX_VALUE;
						int nrivals = 0;
						for(int lit : lucky.lits){
							int a = Math.abs(lit);
							if(!ownsAtom(a)) continue;
							if(atoms.get(a).delta() < minDelta){
								picked = a;
								minDelta = atoms.get(a).delta();
								nrivals = 1;
							}else if(atoms.get(a).delta() == minDelta){
								if(testChance(1.0/(++nrivals))){
									picked = a;
								}
							}
						}
						if(minDelta > 0){
							continue;
						}
					}
				}else{
					// a negative clause
					ArrayList<Integer> cands = new ArrayList<Integer>();
					for(int lit : lucky.lits){
						if(isTrueLit(lit) && ownsAtom(Math.abs(lit))){
							cands.add(Math.abs(lit));
						}
					}
					if(cands.size() == 1){
						picked = cands.get(0);
						if(atoms.get(picked).delta() > 0){
							continue;
						}
					}else{
						if(testChance(Config.walksat_random_step_probability)){
							// a random step
							picked = cands.get(rand.nextInt(cands.size()));
						}else{
							// a greedy step
							double minDelta = Double.MAX_VALUE;
							int nrivals = 0;
							for(int a : cands){
								if(atoms.get(a).delta() < minDelta){
									picked = a;
									minDelta = atoms.get(a).delta();
									nrivals = 1;
								}else if(atoms.get(a).delta() == minDelta){
									if(testChance(1.0/(++nrivals))){
										picked = a;
									}
								}
							}
							if(minDelta > 0){
								continue;
							}
						}
					}
				}
				
				//if(picked == 0) continue;
				inferOps ++;
				if(flip % (nSteps/4) == 1){
					UIMan.verbose(1, "At flip "+UIMan.comma(flip)+
							"/" + UIMan.comma(nSteps));
				}
				if(!ownsAtom(picked)) continue;
				GAtom atom = atoms.get(picked);
				if(Config.avoid_breaking_hard_clauses && 
						atom.criticalForHardClauses()) continue;
				
				if(Config.apply_greedy_throttling && 
						atom.delta() > 0 && 
						testChance(0.7)){
					continue;
				}
				
				// flip this atom
				Timer.runStat.effectiveSteps++;
				
				totalCost += atom.delta();
				
				ArrayList<GAtom> influenced = atom.flip(keyBlock);
				//atom.truth = !atom.truth;
				for(GAtom n : influenced){
					dirtyAtoms.add(n.id);
				}

				if(totalCost < lowCost){
					saveLowTruth(totalCost);
					UIMan.verbose(1, "newLowCost=" + UIMan.comma(lowCost) + 
							" at flip #" + UIMan.comma(flip));
				}
								
				// update stats
				ArrayList<GClause> tlfac = new ArrayList<GClause>();
				ArrayList<GClause> flfac = new ArrayList<GClause>();
				
				//TODO
				if(influenced.isEmpty()){
					
					atom.invertDelta();
					
					if(atom.truth){
						if(adj.get(atom.id) != null)	tlfac.addAll(adj.get(atom.id));
						if(adj.get(-atom.id) != null)	flfac.addAll(adj.get(-atom.id));
					}else{
						if(adj.get(atom.id) != null)	flfac.addAll(adj.get(atom.id));
						if(adj.get(-atom.id) != null)	tlfac.addAll(adj.get(-atom.id));
					}
					
					this.adjustAtomClauseRelation(tlfac, flfac, picked);
					
				}else{
					
					// update f-nsat
					ArrayList<GClause> addNSAT = new ArrayList<GClause>();
					ArrayList<GClause> minusNSAT = new ArrayList<GClause>();
					HashSet<GClause> allGCs = new HashSet<GClause>();
					HashSet<Integer> needToReCalcCost = new HashSet<Integer>();
					
					for(GAtom toFlip : influenced ){
						if(toFlip.truth == false){
							if(adj.containsKey(toFlip.id)) minusNSAT.addAll(adj.get(toFlip.id));
							if(adj.containsKey(-toFlip.id)) addNSAT.addAll(adj.get(-toFlip.id));
						}else{
							if(adj.containsKey(-toFlip.id)) minusNSAT.addAll(adj.get(-toFlip.id));
							if(adj.containsKey(toFlip.id)) addNSAT.addAll(adj.get(toFlip.id));
						}
					}
					for(GClause gc : minusNSAT){
						gc.nsat --;
						allGCs.add(gc);
					}
					for(GClause gc : addNSAT){
						gc.nsat ++;
						allGCs.add(gc);
					}
					double realCost = 0.0;
					for(GClause gc : allGCs){
						for(int a : gc.lits){
							needToReCalcCost.add(Math.abs(a));
						}
						if(gc.weight >= 0){
							if(gc.nsat > 0){
								if(unsat.contains(gc)){
									realCost -= Math.abs(gc.weight);
								}
								unsat.removeObj(gc);
							}else{
								if(!unsat.contains(gc)){
									realCost += Math.abs(gc.weight);
								}
								unsat.add(gc);
							}
						}else{
							if(gc.nsat > 0){
								if(!unsat.contains(gc)){
									realCost += Math.abs(gc.weight);
								}
								unsat.add(gc);
							}else{
								if(unsat.contains(gc)){
									realCost -= Math.abs(gc.weight);
								}
								unsat.removeObj(gc);
							}
						}
					}
						
					if(Math.abs(realCost - atom.delta()) > 0.1 ){
						System.err.println("REALCOST != ESTIMATED COST ERROR");
					}
				
					//this.calcCosts();
					
					this.calcCostsForWalkSAT(needToReCalcCost);
					
				}
				
			}
		}
		UIMan.println("### lowest cost = " + UIMan.comma(lowCost));
	}
	
	
	private void adjustAtomClauseRelation(ArrayList<GClause> tlfac, ArrayList<GClause> flfac, int picked){
		if(tlfac.size() != 0){
			for(GClause f : tlfac){
				++ f.nsat;
				int nsat = f.nsat;
				if(nsat==1){
					if(f.weight >= 0) {
						unsat.removeObj(f);
					}else{
						unsat.add(f);
					}
					for(int lit : f.lits){
						int a = Math.abs(lit);
						if(a==picked || !ownsAtom(a)) continue;
						atoms.get(a).revokeSatPotential(f);
					}
				}else if(nsat==2){
					for(int lit : f.lits){
						int a = Math.abs(lit);
						if(a==picked || !ownsAtom(a)) continue;
						GAtom n = atoms.get(a);
						if((lit>0) == n.truth){
							n.revokeUnsatPotential(f);
							break;
						}
					}
				}
			}
		}
		if(flfac.size() != 0){
			for(GClause f : flfac){
				-- f.nsat;
				int nsat = f.nsat;
				if(nsat==0){
					if(f.weight <= 0) {
						unsat.removeObj(f);
					}else{
						unsat.add(f);
					}
					for(int lit : f.lits){
						int a = Math.abs(lit);
						if(a==picked || !ownsAtom(a)) continue;
						atoms.get(a).assignSatPotential(f);
					}
				}else if(nsat==1){
					for(int lit : f.lits){
						int a = Math.abs(lit);
						if(a==picked || !ownsAtom(a)) continue;
						GAtom n = atoms.get(a);
						if((lit>0) == n.truth){
							n.assignUnsatPotential(f);
							break;
						}
					}
				}
			}
		}
	}
	
	/**
	 * Run SweepSAT for MAP inference.
	 * @param nTries number of tries
	 * @param nSteps number of steps per try
	 * @deprecated
	 */
	public void inferSweepSAT(int nTries, int nSteps){
		DebugMan.checkPeakMem();
		UIMan.println(">>> Running SweepSAT for " + nTries +
				" tries, " + nSteps + " flips/try");
		
		for(int itry=1; itry<=nTries; itry++){
			UIMan.println("[Try #" + itry + "/" + nTries + "]");
			initMRF();
			int nflips = 0;
			while(nflips < nSteps){
				int numGood = 0;
				if(ownsAllAtoms){
					for(GAtom n : atoms.values()){
						if(n.delta() < 0){
							++ numGood;
							if(testChance(Config.sweepsat_greedy_probability)){
								n.flip();
								nflips ++;
								inferOps ++;
								Timer.runStat.effectiveSteps++;
							}
						}
					}
					if(numGood == 0){
						for(GAtom n : atoms.values()){
							if(testChance(0.5)){
								if(!Config.avoid_breaking_hard_clauses || !n.criticalForHardClauses()){
									n.flip();
									nflips ++;
									inferOps ++;
									Timer.runStat.effectiveSteps++;
								}
							}
						}
					}
				}else{
					for(int aid : getCoreAtoms()){
						GAtom n = atoms.get(aid);
						if(n.delta() < 0){
							++ numGood;
							if(testChance(Config.sweepsat_greedy_probability)){
								n.flip();
								nflips ++;
								inferOps ++;
								Timer.runStat.effectiveSteps++;
							}
						}
					}
					if(numGood == 0){
						for(int aid : getCoreAtoms()){
							GAtom n = atoms.get(aid);
							if(testChance(0.5)){
								if(!Config.avoid_breaking_hard_clauses || !n.criticalForHardClauses()){
									n.flip();
									nflips ++;
									inferOps ++;
									Timer.runStat.effectiveSteps++;
								}
							}
						}
					}
				}

				calcCosts();
				if(totalCost < lowCost){
					saveLowTruth(totalCost);
					UIMan.println("["+nflips+"]low=" + lowCost);
					if(lowCost == 0){
						UIMan.println("--all violations gone, cost = " + totalCost);
						return;
					}
				}
			}
		}
		DebugMan.checkPeakMem();
	}



	/**
	 * Initialize the state of the MRF.
	 */
	public void initMRF() {
		switch(initStrategy){
		case ALL_FALSE:
			assignAllFalseTruthValues();
			break;
		case COIN_FLIP:
			assignRandomTruthValues();
			break;
		case GREEDY:
			assignGreedyTruthValues();
			break;
		case COPY_LOW:
			restoreLowTruth();
			break;
		case NO_CHANGE:
			break;
		}
		if(usingBlocks)	maintainKeyConstraints();
		lowCost = calcCosts();
		dirtyAtoms.clear();
		saveTruthAsLow();
	}
	
	private void maintainKeyConstraints(){
		for(GAtom n : atoms.values()){
			if(!keyBlock.hasKey(n)){
				continue;
			}
			
			if(n.truth == true){
				ArrayList<GAtom> mates = keyBlock.getBlockMates(n);
				for(GAtom shouldBeFalse : mates){
					shouldBeFalse.truth = false;
				}
			}
		}
	}

	/**
	 * Set all atoms to false.
	 */
	private void assignAllFalseTruthValues(){
		if(ownsAllAtoms){
			for(GAtom n : atoms.values()){
				n.truth = false;
			}
		}else{
			for(int aid : getCoreAtoms()){
				atoms.get(aid).truth = false;
			}
		}
	}

	/**
	 * Set random atom truth values.
	 */
	private void assignRandomTruthValues(){
		Random rand = new Random();
		int nt = 0;
		if(ownsAllAtoms){
			for(GAtom n : atoms.values()){
				if(n.fixed || (Config.focus_on_critical_atoms && !n.critical())) continue;
				n.truth = rand.nextBoolean();
				if(n.truth) nt++;
			}
		}else{
			for(int aid : getCoreAtoms()){
				GAtom n = atoms.get(aid);
				if(n.fixed || (Config.focus_on_critical_atoms && !n.critical())) continue;
				n.truth = rand.nextBoolean();
				if(n.truth) nt++;
			}
		}
	}

	/**
	 * Assign inital truth values according to some
	 * ad hoc and heuristic stats.
	 */
	private void assignGreedyTruthValues(){
		HashMap<Integer, Double> wts = new HashMap<Integer, Double>();
		if(ownsAllAtoms){
			for(GAtom n : atoms.values()){
				wts.put(n.id, 0.0);
			}
		}else{
			for(int aid : getCoreAtoms()){
				wts.put(aid, 0.0);
			}
		}
		for(GClause f : clauses){
			if(!ownsAllAtoms && isAlwaysTrue(f)) continue;
			for(int lit : f.lits){
				int a = Math.abs(lit);
				Double prev = wts.get(a);
				if(prev != null){
					wts.put(a, prev + Math.signum(lit) * f.weight);
				}
			}
		}
		for(int aid : wts.keySet()){
			GAtom n = atoms.get(aid);
			if(wts.get(n.id) > 0){
				n.truth = true;
			}else{
				n.truth = false;
			}
		}
	}
	
	// TODO: NEED TO CHANGE THIS FUNCTION FOR READING DATA.
	// CURRENTLY, IT IS THE SAME OF RANDOM!!!!!!!
	///**
	// * Assign initial truth value according to training data. 
	// * 
	// * CURRENTLY, IT IS THE SAME OF RANDOM!!!!!!!
	// * NEED TO CHANGE.
	// */
	/*private void assignTruthValuesAccording2TrainingData(){
		Random rand = new Random();
		if(ownsAllAtoms){
			for(GAtom n : atoms.values()){
				if(n.fixed || (Config.focus_on_critical_atoms && !n.critical())) continue;
				n.truth = rand.nextBoolean();
			}
		}else{
			for(int aid : coreAtoms){
				GAtom n = atoms.get(aid);
				if(n.fixed || (Config.focus_on_critical_atoms && !n.critical())) continue;
				n.truth = rand.nextBoolean();
			}
		}
	}*/

	/**
	 * Calculate the number of true literals in a clause.
	 * @param f
	 */
	private void calcNSAT(GClause f){
		int nsat = 0;
		for(int lit : f.lits){
			if(isTrueLit(lit)){
				nsat++;
			}
		}
		f.nsat = nsat;
	}

	private ArrayList<GAtom> getFlipSequence(GAtom a){
		ArrayList<GAtom> ret = new ArrayList<GAtom>();
		
		if(a.truth == true || keyBlock.hasKey(a) == false){
			ret.add(a);
			return ret;
		}else{
		
			ArrayList<GAtom> mates = keyBlock.getBlockMates(a);
			for(GAtom g : mates){
				if(g == a || g.truth == true){
					ret.add(g);
				}
			}
			return ret;
		}
		
	}
	
	/**
	 * Compute total cost and per-atom delta cost.
	 * The delta cost of an atom is the change in the total cost if
	 * this atom is flipped.
	 * @return total cost
	 */
	protected double calcCosts(){
		totalCost = 0;
		unsat.clear();
		// reset stats
		if(ownsAllAtoms){
			for(GAtom n : atoms.values()){
				n.resetDelta();
			}
		}else{
			for(int aid : getCoreAtoms()){
				atoms.get(aid).resetDelta();
			}
		}

		// recompute stats
		for(GClause f : clauses){
			calcNSAT(f);
			if(f.dead) continue;
			totalCost += f.cost();
			if(!ownsAllAtoms && isAlwaysTrue(f)){
				continue;
			}
			if(f.cost() > 0){
				unsat.add(f);
			}
		}
		
		for(GAtom n : atoms.values()){
			
			if(!ownsAllAtoms){	//TODO: CHECK
				if(!getCoreAtoms().contains(n.id)){
					continue;
				}
			}
			
			ArrayList<GAtom> flips = this.getFlipSequence(n);
			ArrayList<GClause> addNSAT = new ArrayList<GClause>();
			ArrayList<GClause> minusNSAT = new ArrayList<GClause>();
			
			for(GAtom toFlip : flips ){
				if(toFlip.truth == true){
					if(adj.containsKey(toFlip.id)) minusNSAT.addAll(adj.get(toFlip.id));
					if(adj.containsKey(-toFlip.id)) addNSAT.addAll(adj.get(-toFlip.id));
				}else{
					if(adj.containsKey(-toFlip.id)) minusNSAT.addAll(adj.get(-toFlip.id));
					if(adj.containsKey(toFlip.id)) addNSAT.addAll(adj.get(toFlip.id));
				}
			}
			
			//TODO: CHANGE TO MORE EFFICIENT DATA STRUCTURE
			HashMap<GClause, Integer> delta = new HashMap<GClause, Integer>();
			for(GClause gc : minusNSAT){
				if(!delta.containsKey(gc)){
					delta.put(gc, 0);
				}
				delta.put(gc, delta.get(gc)-1);
			}
			for(GClause gc : addNSAT){
				if(!delta.containsKey(gc)){
					delta.put(gc, 0);
				}
				delta.put(gc, delta.get(gc)+1);
			}
			
			for(GClause gc : delta.keySet()){
				
				Integer del = delta.get(gc);
				
				if(gc.nsat > 0 && gc.nsat + del <= 0){
					n.assignUnsatPotential(gc);
				}else if (gc.nsat == 0 && gc.nsat + del > 0){
					n.assignSatPotential(gc);
				}
					
			}
		}
				
		return totalCost;
	}
	
	
	private double calcCostsForWalkSAT(HashSet<Integer> needToBeReset){
		
		HashSet<Integer> afterClosure = new HashSet<Integer>();
	/*	
		if(ownsAllAtoms){
			for(GAtom n : atoms.values()){
				afterClosure.add(n.id);
			}
		}else{
			for(int aid : coreAtoms){
				afterClosure.add(aid);
			}
		}
*/

		for(Integer _n : needToBeReset){
	
			GAtom n = atoms.get(_n);
			afterClosure.add(_n);
			ArrayList<GAtom> mates = keyBlock.getBlockMates(n);
			
			if(mates == null){
				//System.err.println("MATE = NULL");
				continue;
			}
			
			for(GAtom mate : mates){
				afterClosure.add(mate.id);
			}
					
		}

		for(Integer _n : afterClosure){

			GAtom n = atoms.get(_n);
			n.resetDelta();
			
			//if(!ownsAllAtoms){	//TODO: CHECK
			//	if(!coreAtoms.contains(n.id)){
			//		continue;
			//	}
			//}
			
			ArrayList<GAtom> flips = this.getFlipSequence(n);
			ArrayList<GClause> addNSAT = new ArrayList<GClause>();
			ArrayList<GClause> minusNSAT = new ArrayList<GClause>();
			
			for(GAtom toFlip : flips ){
				if(toFlip.truth == true){
					if(adj.containsKey(toFlip.id)) minusNSAT.addAll(adj.get(toFlip.id));
					if(adj.containsKey(-toFlip.id)) addNSAT.addAll(adj.get(-toFlip.id));
				}else{
					if(adj.containsKey(-toFlip.id)) minusNSAT.addAll(adj.get(-toFlip.id));
					if(adj.containsKey(toFlip.id)) addNSAT.addAll(adj.get(toFlip.id));
				}
			}
			
			//TODO: CHANGE TO MORE EFFICIENT DATA STRUCTURE
			HashMap<GClause, myInt> delta = new HashMap<GClause, myInt>();
			for(GClause gc : minusNSAT){
				if(!delta.containsKey(gc)){
					delta.put(gc, new myInt(0));
				}
				delta.get(gc).subOne();
			}
			for(GClause gc : addNSAT){
				if(!delta.containsKey(gc)){
					delta.put(gc, new myInt(0));
				}
				delta.get(gc).addOne();
			}
			
			for(GClause gc : delta.keySet()){
				
				Integer del = delta.get(gc).value;
				
				if(gc.nsat > 0 && gc.nsat + del <= 0){
					n.assignUnsatPotential(gc);
				}else if (gc.nsat == 0 && gc.nsat + del > 0){
					n.assignSatPotential(gc);
				}
					
			}
		}
				
		return totalCost;
	}

	/**
	 * Track ground clause violations to fo-clauses.
	 * Stats are records on a per fo-clause basis.
	 * 
	 * @see	tuffy.helper.Stats#reportMostViolatedClauses
	 */
	public void auditClauseViolations(){
		restoreLowTruth();
		totalCost = 0;
		for(Clause c : mln.getAllNormalizedClauses()){
			c.cost = 0;
			c.violations = 0;
			c.violatedGClauses = new ArrayList<GClause>();
		}
		for(GClause f : clauses){
			calcNSAT(f);
			double c = f.cost();
			int kv = 0;
			if(c > 0){
				totalCost += c;
				for(int fc : f.fcid){
					Clause cl = mln.getClauseById(fc);
					boolean pen0 = cl.isPositiveClause();
					if(fc < 0) pen0 = !pen0;
					if(pen0 == (f.nsat == 0)){
						kv ++;
					}
				}
				if(kv > 0)	{
					for(int fc : f.fcid){
						Clause cl = mln.getClauseById(fc);
						boolean pen0 = cl.isPositiveClause();
						if(fc < 0) pen0 = !pen0;
						if(pen0 == (f.nsat == 0)){
							cl.cost += (c/kv);
							cl.violations += 1.0/kv;
							cl.violatedGClauses.add(f);
						}
					}
				}else{
					// System.err.println("found unaccounted-for ground clauses");
				}
			}
		}
	}
	
	/**
	 * Recalculate total cost.
	 * @return updated total cost
	 */
	public double recalcCost(){
		totalCost = 0;
		for(GClause f : clauses){
			calcNSAT(f);
			totalCost += f.cost();
		}
		return totalCost;
	}

	public double recalcCostNotOverrideOldCost(){
		double atotalCost = 0;
		for(GClause f : clauses){
			calcNSAT(f);
			atotalCost += f.cost();
		}
		return atotalCost;
	}
	
	public HashSet<Integer> getCoreAtoms() {
		return coreAtoms;
	}

	private class myInt{
		int value = -1;
		
		public void addOne(){
			value ++;
		}
		
		public void subOne(){
			value --;
		}
		
		public myInt(int _value){
			value = _value;
		}
		
	}
	
	
	/****************************************
	 * MCSAT Below
	 ***************************************/
	

	/**
	 * This map records the expectation of #violation for
	 * each clause. This is filled by {@link MCSAT#calcExpViolation()}.
	 */
	public HashMap<String, Double> expectationOfViolation = null;

	/**
	 * This map records the expectation of square #violation for
	 * each clause. This is filled by {@link MCSAT#calcExpViolation()}.
	 */
	public HashMap<String, Double> expectationOfSquareViolation = null;

	/**
	 * This map records the tallies for calculating E(v_i*v_j).
	 */
	public HashMap<String, Long> clauseNiNjViolationTallies = null;

	/**
	 * This map records the expectation of E(v_i*v_j). 
	 * This is filled by {@link MCSAT#calcExpViolation()}.
	 */
	public HashMap<String, Double> expectationOfNiNjViolation = null;

	/**
	 * This array records the expection of #satisfaction for
	 * each clause. This is filled by {@link MCSAT#calcExpViolation()}.
	 */
	public HashMap<String, Double> expectationOfSatisfication = null;


	/**
	 * This array records total number of violation for a clause.
	 * Dividing this number by {@link MCSAT#nClauseVioTallies}
	 * will give the estimated expectation of #violation. 
	 */
	public HashMap<String, Long> clauseVioTallies = null;

	/**
	 * This array records total number of square violation for a clause.
	 * Dividing this number by {@link MCSAT#nClauseVioTallies}
	 * will give the estimated expectation of #violation. 
	 */
	public HashMap<String, Long> clauseSquareVioTallies = null;

	/**
	 * This array records total number of satisfaction for a clause.
	 */
	public HashMap<String, Long> clauseSatTallies = null;

	/**
	 * Number of iterations of tallies. 
	 */
	private int nClauseVioTallies = 0;

	/**
	 * Kill soft clauses.
	 * 
	 * @return the number of hard clauses
	 */
	protected int retainOnlyHardClauses(){
		int numHard = 0;
		for(GClause c : clauses){
			if(c.isHardClause() == c.dead){
				c.dead = !c.dead;
			}
			if(!c.dead){
				++ numHard;
			}
		}
		totalAlive = numHard;
		return numHard;
	}

	/**
	 * SampleSAT (with WalkSAT inside), used to uniformly sample a zero-cost world.
	 * WalkSAT is used as a SAT solver to find the first (quasi-)zero-cost world.
	 * Simulated annealing (SA) is stochastically performed to wander around.
	 * @param nSteps
	 * 
	 * @return true iff a zero-cost world was reached
	 */
	protected boolean sampleSAT(long nSteps){
		if(!Config.learning_mode)
			UIMan.println("    Running SampleSAT for " + UIMan.comma(nSteps) + " flips...");

		if(adj.isEmpty()) buildIndices();
		Random rand = new Random();

		initMRF();
		
		for(long flip = 1; flip <= nSteps; flip++){
			// check if we have reached terminal condition
			if(unsat.isEmpty()){
				saveLowTruth(totalCost);
				// // keep walking only if in sampleSAT mode
				if(Config.stop_samplesat_upon_sat)
					return true;
				if(!sampleSatMode || totalAlive == 0) return true;
			}
			// simulated annealing step
			boolean saStep = false;
			// id of atom to be flipped
			int picked = 0;

			/**
			 * pick an atom to flip in one of two ways:
			 * WalkSAT or SA
			 */
			if(sampleSatMode && (totalCost <= 0.0001 
					|| rand.nextDouble() <= 0.5)){ 
				// SA step: randomly pick an atom
				GClause c = clauses.get(rand.nextInt(clauses.size()));
				picked = Math.abs(c.lits[0]);
				saStep = true;
			}else{
				// WalkSAT step
				GClause lucky = unsat.getRandomElement();
				if(lucky.isPositiveClause()){
					// a positive clause
					if(testChance(Config.walksat_random_step_probability)){
						// random flip
						picked = -1;
						while(picked == -1 || !ownsAtom(picked)){
							picked = Math.abs(lucky.lits[rand.nextInt(lucky.lits.length)]);
						}
					}else{
						// greedy flip
						double minDelta = Double.MAX_VALUE;
						int nrivals = 0;
						for(int lit : lucky.lits){
							int a = Math.abs(lit);
							if(!ownsAtom(a)) continue;
							if(atoms.get(a).delta() < minDelta){
								picked = a;
								minDelta = atoms.get(a).delta();
								nrivals = 1;
							}else if(atoms.get(a).delta() == minDelta){
								if(testChance(1.0/(++nrivals))){
									picked = a;
								}
							}
						}
					}
				}else{
					// a negative clause
					ArrayList<Integer> cands = new ArrayList<Integer>();
					for(int lit : lucky.lits){
						if(isTrueLit(lit) && ownsAtom(Math.abs(lit))){
							cands.add(Math.abs(lit));
						}
					}
					if(cands.size() == 1){
						picked = cands.get(0);
					}else{
						if(testChance(Config.walksat_random_step_probability)){
							picked = cands.get(rand.nextInt(cands.size()));
						}else{
							double minDelta = Double.MAX_VALUE;
							int nrivals = 0;
							for(int a : cands){
								// TODO: FIXED
								if(atoms.get(a).delta() < minDelta){
									picked = a;
									minDelta = atoms.get(a).delta();
									nrivals = 1;
								}else if(atoms.get(a).delta() == minDelta){
									if(testChance(1.0/(++nrivals))){
										picked = a;
									}
								}
							}
						}
					}
				}
			}
			if(picked == 0) continue;
			if(!ownsAtom(picked)) continue;
			GAtom atom = atoms.get(picked);
			if(atom.fixed) continue;
			if(Config.avoid_breaking_hard_clauses && 
					atom.criticalForHardClauses()) continue;
			if(Config.apply_greedy_throttling && 
					atom.delta() > 0 && 
					testChance(0.95)){
				continue;
			}


			// SA is just a filter to actual flipping
			if(saStep){
				if(rand.nextDouble() > Math.exp(-atom.delta()*Config.samplesat_sa_coef)){
					continue;
				}
			}

			/**
			 *  flip the picked atom
			 */
			totalCost += atom.delta();
			atom.truth = !atom.truth;
			dirtyAtoms.add(picked);
			atom.invertDelta();

			if(totalCost <= lowCost){
				saveLowTruth(totalCost);
			}

			// update stats
			ArrayList<GClause> tlfac, flfac;
			if(atom.truth){
				tlfac = adj.get(atom.id);
				flfac = adj.get(-atom.id);
			}else{
				flfac = adj.get(atom.id);
				tlfac = adj.get(-atom.id);
			}

			if(tlfac != null){
				for(GClause f : tlfac){
					///////////////////////////////////////////////
					// Ce flip the following two lines on Nov. 29
					///////////////////////////////////////////////
					++ f.nsat;
					if(f.dead) continue;

					int nsat = f.nsat;
					if(nsat==1){
						if(f.weight >= 0) {
							unsat.removeObj(f);
						}else{
							unsat.add(f);
						}
						for(int lit : f.lits){
							int a = Math.abs(lit);
							if(a==picked || !ownsAtom(a)) continue;
							atoms.get(a).revokeSatPotential(f);
						}
					}else if(nsat==2){
						for(int lit : f.lits){
							int a = Math.abs(lit);
							if(a==picked || !ownsAtom(a)) continue;
							GAtom n = atoms.get(a);
							if((lit>0) == n.truth && willChange(lit, f.lits)){
								n.revokeUnsatPotential(f);
								break;
							}
						}
					}
				}
			}
			if(flfac != null){
				for(GClause f : flfac){
					///////////////////////////////////////////////
					// Ce flip the following two lines on Nov. 29
					///////////////////////////////////////////////
					-- f.nsat;
					if(f.dead) continue;

					int nsat = f.nsat;
					if(nsat==0){
						if(f.weight <= 0) {
							unsat.removeObj(f);
						}else{
							unsat.add(f);
						}
						for(int lit : f.lits){
							int a = Math.abs(lit);
							if(a==picked || !ownsAtom(a)) continue;
							atoms.get(a).assignSatPotential(f);
						}
					}else if(nsat==1){
						for(int lit : f.lits){
							int a = Math.abs(lit);
							if(a==picked || !ownsAtom(a)) continue;
							GAtom n = atoms.get(a);
							if((lit>0) == n.truth && willChange(lit, f.lits)){
								n.assignUnsatPotential(f);
								break;
							}
						}
					}
				}
			}
		}
		return false;
	}

	public void updateAtomMarginalProbs(int numSamples){
		if(ownsAllAtoms){
			for(GAtom n : atoms.values()){
				n.prob = ((float)(n.tallyTrue)) / numSamples;
				//TODO: CHECK n.truth = (n.prob >= 0.5);
			//	n.truth = (n.prob >= 0.5);
				n.truth = true;
			}
		}else{
			for(int aid : getCoreAtoms()){
				GAtom n = atoms.get(aid);
				n.prob = ((float)(n.tallyTrue)) / numSamples;
			//	n.truth = (n.prob >= 0.5);
				n.truth = true;
			}
		}
	}

	/**
	 * For each atom, increment its truth tally by one if it's currently true.
	 */
	private void updateAtomTruthTallies(){
		if(ownsAllAtoms){
			for(GAtom n : atoms.values()){
				if(n.truth) n.tallyTrue++;
			}
		}else{
			for(int aid : getCoreAtoms()){
				GAtom n = atoms.get(aid);
				if(n.truth) n.tallyTrue++;
			}
		}
	}

	/**
	 * Execute the MC-SAT algorithm.
	 * @param numSamples number of MC-SAT samples
	 * @param numFlips number of SampleSAT steps in each iteration
	 */
	@SuppressWarnings("unused")
	public double mcsat(int numSamples, long numFlips, DataMover... dmovers){
		initStrategy = INIT_STRATEGY.COIN_FLIP;
		if(!Config.learning_mode) Config.stop_samplesat_upon_sat = true;
		UIMan.println(">>> Running MC-SAT for " + numSamples + " samples...");
		// init
		sampleSatMode = false;

		boolean isFirstTime = true;

		if(isFirstTime && !Config.snapshoting_so_do_not_do_init_flip){
			UIMan.verbose(1, ">>> MC-SAT INIT: running WalkSAT on hard clauses...");
			int x = retainOnlyHardClauses();
			UIMan.verbose(1, "### hard clauses = " + x);
			sampleSAT(numFlips);
		}

		if(!Config.snapshoting_so_do_not_do_init_flip){
			isFirstTime = false;
			enableAllClauses();
			restoreLowTruth();
		}

		sampleSatMode = true;
		// clear tallies of clauses.
		this.nClauseVioTallies = 0;
		this.clauseVioTallies = null;
		this.clauseSatTallies = null;
		this.expectationOfViolation = null;
		this.expectationOfSatisfication = null;
		this.clauseSquareVioTallies = null;
		this.expectationOfSquareViolation = null;
		this.clauseNiNjViolationTallies = null;
		this.expectationOfNiNjViolation = null;

		int dumpingTime = 0;
		int lastTime = (int)Timer.elapsedSeconds();
		
		HashSet<Integer> historyTime = new HashSet<Integer>();
		
		double sumCost = 0;
		
		// sample
		for(int i=1; i<=numSamples; i++){
			if(Config.learning_mode)
				UIMan.print("*");
			else
				UIMan.println(">>> MC-SAT Sample #" + i + "");
			sumCost += performMCSatStep(numFlips);
			int curTime = (int) Timer.elapsedSeconds();
			
			if(dmovers.length == 1){
				if( i < 100  || (Math.abs(((curTime - dumpingTime) % 100)) < 5 && !historyTime.contains(curTime - dumpingTime)) ){
					historyTime.add(curTime - dumpingTime);
					String fout = "mcsat.results.samples." + i + ".seconds." + (curTime - dumpingTime);
					UIMan.println("MCSAT-SAMPLE-" + i + ": " + (curTime-dumpingTime) + " " + Timer.elapsed());
					UIMan.println("Writing MCSAT result to file " + fout);
					
					updateAtomMarginalProbs(i);
					dmovers[0].flushAtomStates(this.atoms.values(), mln.relAtoms);
					dmovers[0].dumpProbsToFile(mln.relAtoms, fout);
					
					int aTime = (int)Timer.elapsedSeconds();
					dumpingTime += (aTime - curTime);
				}
			}
			
/*			if(Config.mcsatDumpPeriodSamples == 46886221 && i % Config.mcsatDumpPeriodSamples == 0){
				String fout = "mcsat.results.samples." + i + ".seconds." + (curTime-dumpingTime);
				UIMan.println("MCSAT-SAMPLE-" + i + ": " + (curTime-dumpingTime) + " sec; " + Timer.elapsed());
				UIMan.println("Writing MCSAT result to file " + fout);
				dumpAtomProb(i, fout);
				int aTime = (int)Timer.elapsedSeconds();
				dumpingTime += (aTime - curTime);
				lastTime = aTime;
			}else if((i<=100 && i%10==0) || i > 100 && Config.mcsatDumpPeriodSeconds > 0 && curTime - lastTime >= Config.mcsatDumpPeriodSeconds){
				String fout = "mcsat.results.samples." + i + ".seconds." + (curTime-dumpingTime);
				UIMan.println("MCSAT-SAMPLE-" + i + ": " + (curTime-dumpingTime) + " " + Timer.elapsed());
				UIMan.println("Writing MCSAT result to file " + fout);
				dumpAtomProb(i, fout);
				int aTime = (int)Timer.elapsedSeconds();
				dumpingTime += (aTime - curTime);
				lastTime = aTime;
			}*/
			//if(curTime-dumpingTime >= Config.timeout){
			//	System.out.println("!!! TIME OUT AT " + (curTime - dumpingTime) + " sec.");
			//	break;
			//}
		}

		if(Config.learning_mode){
			this.calcExpViolation();
		}
		updateAtomMarginalProbs(numSamples);
		
		return sumCost;
	}

	/**
	 * Update the number of violations of a clause. For each GClause, their
	 * value can increase at most 1 for each MCSAT iteration. For Clause,
	 * their value can increase more, because there may be more than one
	 * GClauses associated with it. 
	 */
	public void updateClauseVoiTallies(){

		// at least one clause in this system
		// assert(GClause.maxFCID >= 0);

		if(this.clauseVioTallies == null){
			this.clauseVioTallies = new HashMap<String, Long>();
			this.clauseSatTallies = new HashMap<String, Long>();
			this.clauseSquareVioTallies = new HashMap<String, Long>();
			this.clauseNiNjViolationTallies = new HashMap<String, Long>();
			//this.nGClauseEachClause = new HashMap<String, Long>();
		}
		HashMap<String, Long> tmpSquareViolation = new HashMap<String, Long>();

		this.nClauseVioTallies ++ ;

		// HERE THE METHOD FOR CALCULATION OF UNSATISFIED GCLAUSE
		// CAN BE INFLUENCED BY CE'S CHANGE IN FUNCTION SAMPLESAT().
		for(GClause c : clauses){

			if(c.ffcid[0].equals("4.0") && c.nsat==0){
				//System.out.print("");
			}

			if( c.nsat == 0){
				// This GClause == false.
				for(String cid : c.ffcid){
					String newCID = cid;
					if(newCID.charAt(0) == '-'){
						newCID = newCID.substring(1, newCID.length());
						if( -Learner.currentWeight.get(newCID) > 0 ){
							// -x is po2.0=11sitive clause
							// => -x is violated
							// => x is violated
							Long tmp = this.clauseVioTallies.get(newCID);
							if(tmp == null) tmp = 0l;
							this.clauseVioTallies.put(newCID, tmp+1);

							tmp = tmpSquareViolation.get(newCID);
							if(tmp == null) tmp = 0l;
							tmpSquareViolation.put(newCID, tmp+1);

							tmp = this.clauseSatTallies.get(newCID);
							if(tmp == null)
								this.clauseSatTallies.put(newCID, 0l);
						}else{
							// -x is negative clause
							// => -x is satisfied
							// => x is satisfied
							Long tmp = this.clauseSatTallies.get(newCID);
							if(tmp == null) tmp = 0l;
							this.clauseSatTallies.put(newCID, tmp+1);

							tmp = this.clauseVioTallies.get(newCID);
							if(tmp == null)
								this.clauseVioTallies.put(newCID, 0l);

							tmp = tmpSquareViolation.get(newCID);
							if(tmp == null)
								tmpSquareViolation.put(newCID, 0l);
						}	
					}else{
						if( Learner.currentWeight.get(newCID) > 0 ){
							// x is positive clause
							// => x is violated
							Long tmp = this.clauseVioTallies.get(newCID);
							if(tmp == null) tmp = 0l;
							this.clauseVioTallies.put(newCID, tmp+1);

							tmp = tmpSquareViolation.get(newCID);
							if(tmp == null) tmp = 0l;
							tmpSquareViolation.put(newCID, tmp+1);

							tmp = this.clauseSatTallies.get(newCID);
							if(tmp == null)
								this.clauseSatTallies.put(newCID, 0l);
						}else{
							// x is negative clause
							// => x is satisfied
							Long tmp = this.clauseSatTallies.get(newCID);
							if(tmp == null) tmp = 0l;
							this.clauseSatTallies.put(newCID, tmp+1);

							tmp = this.clauseVioTallies.get(newCID);
							if(tmp == null)
								this.clauseVioTallies.put(newCID, 0l);

							tmp = tmpSquareViolation.get(newCID);
							if(tmp == null)
								tmpSquareViolation.put(newCID, 0l);
						}	
					}
				}
			}else{
				// This GClause == true.
				for(String cid : c.ffcid){
					String newCID = cid;
					if(newCID.charAt(0) == '-'){
						newCID = newCID.substring(1, newCID.length());
						if( -Learner.currentWeight.get(newCID) > 0 ){
							// -x is positive clause
							// => -x is satisfied
							// => x is satisfied
							Long tmp = this.clauseSatTallies.get(newCID);
							if(tmp == null) tmp = 0l;
							this.clauseSatTallies.put(newCID, tmp+1);

							tmp = this.clauseVioTallies.get(newCID);
							if(tmp == null)
								this.clauseVioTallies.put(newCID, 0l);

							tmp = tmpSquareViolation.get(newCID);
							if(tmp == null)
								tmpSquareViolation.put(newCID, 0l);
						}else{
							// -x is negative clause
							// => -x is violated
							// => x is violated
							Long tmp = this.clauseVioTallies.get(newCID);
							if(tmp == null) tmp = 0l;
							this.clauseVioTallies.put(newCID, tmp+1);

							tmp = tmpSquareViolation.get(newCID);
							if(tmp == null) tmp = 0l;
							tmpSquareViolation.put(newCID, tmp+1);

							tmp = this.clauseSatTallies.get(newCID);
							if(tmp == null)
								this.clauseSatTallies.put(newCID, 0l);
						}	
					}else{
						if( Learner.currentWeight.get(newCID) > 0 ){
							// x is positive clause
							// => x is satisfied
							Long tmp = this.clauseSatTallies.get(newCID);
							if(tmp == null) tmp = 0l;
							this.clauseSatTallies.put(newCID, tmp+1);

							tmp = this.clauseVioTallies.get(newCID);
							if(tmp == null)
								this.clauseVioTallies.put(newCID, 0l);

							tmp = tmpSquareViolation.get(newCID);
							if(tmp == null)
								tmpSquareViolation.put(newCID, 0l);
						}else{
							// x is negative clause
							// => x is violated
							Long tmp = this.clauseVioTallies.get(newCID);
							if(tmp == null) tmp = 0l;
							this.clauseVioTallies.put(newCID, tmp+1);

							tmp = tmpSquareViolation.get(newCID);
							if(tmp == null) tmp = 0l;
							tmpSquareViolation.put(newCID, tmp+1);

							tmp = this.clauseSatTallies.get(newCID);
							if(tmp == null)
								this.clauseSatTallies.put(newCID, 0l);
						}	
					}
				}

			}
		}

		for(String cid : tmpSquareViolation.keySet()){
			Long value = tmpSquareViolation.get(cid);
			Long tmp = this.clauseSquareVioTallies.get(cid);
			if(tmp == null){
				this.clauseSquareVioTallies.put(cid, value*value);
			}else{
				this.clauseSquareVioTallies.put(cid, tmp + value*value);
			}
		}

		for(String cid1 : tmpSquareViolation.keySet()){
			Long value1 = tmpSquareViolation.get(cid1);
			for(String cid2 : tmpSquareViolation.keySet()){
				Long value2 = tmpSquareViolation.get(cid2);
				Long tmp = this.clauseNiNjViolationTallies.get(cid1+","+cid2);
				if(tmp == null){
					this.clauseNiNjViolationTallies.put(cid1+","+cid2, value1*value2);
				}else{
					this.clauseNiNjViolationTallies.put(cid1+","+cid2, tmp + value1*value2);
				}
			}
		}

	}

	/**
	 * Calculating the different expectations by filling the HashMaps related to
	 * expectations in this class.
	 */
	public void calcExpViolation(){
		// has run updateClauseVoiTallies() before
		assert(this.clauseVioTallies != null);
		assert(this.nClauseVioTallies > 0);

		this.expectationOfViolation = new HashMap<String, Double>();
		this.expectationOfSatisfication = new HashMap<String, Double>();
		this.expectationOfSquareViolation = new HashMap<String, Double>();
		this.expectationOfNiNjViolation = new HashMap<String, Double>();

		for(String k : this.clauseVioTallies.keySet()){
			this.expectationOfViolation.put(k, 
					((double)this.clauseVioTallies.get(k)) / this.nClauseVioTallies);
			this.expectationOfSatisfication.put(k, 
					((double)this.clauseSatTallies.get(k)) / this.nClauseVioTallies);
			this.expectationOfSquareViolation.put(k, 
					((double)this.clauseSquareVioTallies.get(k)) / this.nClauseVioTallies );
		}

		for(String k : this.clauseNiNjViolationTallies.keySet()){
			this.expectationOfNiNjViolation.put(k, 
					((double)this.clauseNiNjViolationTallies.get(k)) 
					/ this.nClauseVioTallies);
		}
	}


	/**
	 * Change the weight of GClause based on updated weight
	 * of Clause. This new weight will be aware by MCSAT. The
	 * cost of flipping atom and the unsat set for GClause will
	 * be calculated automatically by this function.
	 * 
	 * @param currentWeight The weight of clauses to be flushed
	 * in this MCSAT instance.
	 */
	public void updateClauseWeights(HashMap<String, Double> currentWeight){
		for(GClause c : this.clauses){
			String[] ffcid = c.ffcid;
			c.weight = 0;
			for(int i=0;i<ffcid.length;i++){
				String newCID = ffcid[i];
				int signal = 1;
				if(newCID.charAt(0) == '-'){
					newCID = newCID.substring(1, newCID.length());
					signal = -1;
				}

				// TODO: HERE IS THE HARD-CLAUSE TRICK. CHECK WHETHER
				// WE NEED IT.
				// it is almost impossible for a non-zero-violation
				// clause to have such a big value...
				//if(currentWeight.get(newCID)*signal > 19){
				//	Learner.isHardMappings.put(newCID, true);
				//	c.weight = Config.hard_weight + 100;
				//	break;
				//}else
				c.weight += currentWeight.get(newCID) * signal;
			}
		}
		calcCosts();
		buildIndices();
	}

	/**
	 * Perform one sample of MC-SAT
	 * @param numFlips number of sampleSAT flips
	 */
	private double performMCSatStep(long numFlips){

		this.invalidateLowCost();

		retainSomeGoodClauses();
		// TODO: CURRENTLY, CE FINDS ADDING THIS FUNCTION
		// MAY INFLUENCE THE SPEED OF SAMPLESAT. CHECK IT
		// IN THE FUTURE
		// unitPropagation();

		sampleSAT(numFlips);

		unfixAllAtoms();
		enableAllClauses();
		restoreLowTruth();
		updateAtomTruthTallies();
		
		this.recalcCost();
		
		if(Config.learning_mode){
			updateClauseVoiTallies();
		}
		
		return this.getCost();

	}


	/**
	 * Try to satisfy as many clauses as possible with unit propagation.
	 * Used as a preprocessing step of SampleSAT, which tries to uniformly 
	 * sample among all zero-cost worlds.
	 */
	@SuppressWarnings("unused")
	private void unitPropagation(){
		/**
		 *  Satisfy negative clauses.
		 *  We have to make sure all literals are false.
		 */
		for(GClause cee : clauses){
			if(cee.dead || cee.isPositiveClause()) continue;
			if(isAlwaysTrue(cee)) continue;
			for(int lit : cee.lits){
				if(lit > 0){
					fixAtom(lit, false);
				}else{
					fixAtom(-lit, true);
				}
			}
		}

		/**
		 *  Propagate to positive clauses.
		 *  We only need to fix one literal of each positive clause.
		 *  Keep sweeping the clauses until no fixing can be done in the whole round.
		 *  TODO: improve the efficiency of this procedure
		 */
		boolean done = false;
		while(!done){
			done = true;
			for(GClause cee : clauses){
				if(cee.dead || !cee.isPositiveClause()) continue;
				if(isAlwaysTrue(cee)) continue;

				int numCand = 0, lastCand = 0;
				for(int lit : cee.lits){
					GAtom atom = atoms.get(Math.abs(lit));
					if(!atom.fixed){
						lastCand = lit;
						numCand ++;
						if(numCand > 1) break;
					}
				}
				if(numCand == 1){
					fixAtom(Math.abs(lastCand), lastCand>0);
					done = false;
				}
			}
		}
	}

	
	
	

}
