package tuffy.learn;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;

import tuffy.infer.MRF;
import tuffy.main.Infer;
import tuffy.mln.Clause;
import tuffy.mln.Predicate;
import tuffy.parse.CommandOptions;
import tuffy.util.Config;
import tuffy.util.FileMan;
import tuffy.util.StringMan;
import tuffy.util.Timer;
import tuffy.util.UIMan;

// TODO: CURRENTLY ASSUME THE ACTIVE SET IS NOT CHANGED
// WHEN FLIPPING A POSITIVE CLAUSE TO NEGATIVE CLAUSE.

/**
 * The abstract class of learning a weight of MLN.
 * A runnable learner should extend this abstract
 * class by specifying some ad-hoc functions.
 * This class extends Infer class, because it
 * uses inference as subroutines.
 * 
 * NOTE: this class has static variables.
 * Parallell running of multiple learner instances may
 * cause problems.
 * 
 * @author Ce Zhang
 *
 */
public abstract class Learner extends Infer {
	
	/**
	 * Map from clause name to current clause weight.
	 */
	static public HashMap<String, Double> currentWeight = null;
	
	/**
	 * Map from clause name to the clause weight read originally from
	 * MLN program.
	 */
	static public HashMap<String, Double> oriWeight = null;
	
	/**
	 * Map from clause name to whether it is assigned to hard
	 * weight clause while learning.
	 */
	static public HashMap<String, Boolean> isHardMappings = null;
	
	/**
	 * Map from clause name to clause weight learned in last
	 * iteration.
	 */
	public HashMap<String, Double> _oldWeight = null;
	
	//public HashMap<String, Double> randomWeight = null;
	//
	//public HashMap<String, Double> oriWeight = null;
	
	///**
	// * Map from clause name to the corresponding H^(-1)g value.
	// */
	//public static HashMap<String, Double> currentD = null;
	
	/**
	 * Map from clause name to final weight. Here by final
	 * weight, it means the the average of the weights
	 * in each iteration.
	 */
	public static HashMap<String, Double> finalWeight = null;
	
	/**
	 * FOR JUNIT TEST ONLY.
	 */
	public double odds = 0;
	
	/**
	 * Whether current step is a backtracked step.
	 */
	public boolean backtracked = false;
	
	/**
	 * Number of past backtracked steps.
	 */
	public int backtrackCount_ = 0;
	
	/**
	 * Map from clause name to current training data violation.
	 */
	public HashMap<String, Long> trainingViolation = null;
	
	/**
	 * Map from clause name to current training data satisfaction.
	 */
	public HashMap<String, Long> trainingSatisification = null;
	
	/**
	 * Map from clause name to training data violation of positive weight.
	 */
	HashMap<String, Long> positiveWeightViolation = null;
	
	/**
	 * Map from clause name to training data satisfaction of positive weight.
	 */
	HashMap<String, Long> positiveWeightSatisfication = null;
	
	/**
	 * Map from clause name to training data violation of negative weight.
	 */
	HashMap<String, Long> negativeWeightViolation = null;
	
	/**
	 * Map from clause name to training data satisfaction of negative weight.
	 */
	HashMap<String, Long> negativeWeightSatisfication = null;
	
	/**
	 * run the learner
	 * @param opt Command line Options
	 * @throws SQLException 
	 */
	@SuppressWarnings("unchecked")
	public void run(CommandOptions opt) throws SQLException{
		// tally clause
		Config.learning_mode = true;
		Config.stop_samplesat_upon_sat = true;
		Config.track_clause_provenance = true;
		
		UIMan.println(">>> Learning clause weight with MC-SAT.");
		setUp(opt);
		ground();
		
		Timer.runStat.markGroundingDone();

		
		if(options.maxFlips == 0){
			options.maxFlips = 100 * grounding.getNumAtoms();
		}
		if(options.maxTries == 0){
			options.maxTries = 1;
		}
	
		Learner.isHardMappings = new HashMap<String, Boolean>();
		//String outputStr = "";
		
		// INIT CURRENT WEIGHT, LOAD FROM CBUFFER TABLE
		Learner.currentWeight = new HashMap<String, Double>();
		Learner.oriWeight = new HashMap<String, Double>();
		Learner.finalWeight = new HashMap<String, Double>();
		//Learner.currentD = new HashMap<String, Double>();
		String sql = "SELECT DISTINCT weight, ffcid FROM " + "mln" + mln.getID() + "_cbuffer" + ";";
		ResultSet rs = db.query(sql);
		while(rs.next()){
			String ffcid = rs.getString("ffcid");
			Double wght = rs.getDouble("weight");
			String newCID = ffcid;
			if(newCID.charAt(0) == '-'){
				newCID = newCID.substring(1, newCID.length());
				wght = -wght;
			}
			Learner.currentWeight.put(newCID, wght);
			Learner.oriWeight.put(newCID, wght);
			//Learner.currentD.put(newCID, 0.0);
		}
		rs.close();
		
		// INIT TRAINING VIOLATION AND WEIGHT
		MRF _mcsat = new MRF(mln);
		dmover.loadMrfFromDb(_mcsat, mln.relAtoms, mln.relClauses);
		this.loadingTrainingData(_mcsat);
		
		// all positive violation
		for(String s : currentWeight.keySet()){
			currentWeight.put(s, 1.0);
		}
		_mcsat.updateClauseWeights(currentWeight);
		_mcsat.updateClauseVoiTallies();
		positiveWeightViolation = (HashMap<String, Long>)
			(_mcsat.clauseVioTallies.clone());
		positiveWeightSatisfication = (HashMap<String, Long>)
			(_mcsat.clauseSatTallies.clone());	
		
		// all negative violation
		for(String s : currentWeight.keySet()){
			currentWeight.put(s, -1.0);
		}
		_mcsat.updateClauseWeights(currentWeight);
		_mcsat.clauseVioTallies.clear();
		_mcsat.clauseSatTallies.clear();
		_mcsat.updateClauseVoiTallies();
		negativeWeightViolation = (HashMap<String, Long>)
			(_mcsat.clauseVioTallies.clone());
		negativeWeightSatisfication = (HashMap<String, Long>)
			(_mcsat.clauseSatTallies.clone());
		
		Learner.currentWeight = (HashMap<String, Double>) Learner.oriWeight.clone();
		this.calcCurrentTrainingViolation();
		this.fillInCurrentWeight(_mcsat);
		_mcsat.updateClauseWeights(currentWeight);
		this.calcCurrentTrainingViolation();

		//OUTPUT THE INIT. WEIGHT AND ID/CLAUSE MAPPING 
		Object[] keySet = currentWeight.keySet().toArray();
		java.util.Arrays.sort(keySet);
		UIMan.println("#################INIT. WEIGHT#################");
		for(Object ss : keySet){
			String s = (String) ss;
			UIMan.println(s + "\t" + Learner.currentWeight.get(s) + ":" + 
					this.trainingSatisification.get(s) + "/" + this.trainingViolation.get(s) + 
					"\t" + Clause.mappingFromID2Desc.get(s));
		}
		
		UIMan.println(">>> Iteration Begins...");
		
		_oldWeight = (HashMap<String, Double>) currentWeight.clone();
	
		// options.nDIteration = 1;
		for(int i=0;i<options.nDIteration;i++){

			_mcsat.mcsat(options.mcsatSamples, options.maxFlips);
			
			if(updateWeight(_mcsat) == true){
				break;
			}
			
			_mcsat.updateClauseWeights(Learner.currentWeight);
			
			if(this.backtracked == true){
				UIMan.println("####BACKTRACKED####");
				this.calcCurrentTrainingViolation();
				i--;
				continue;
			}

			//CALC THE FINAL WEIGHT
			for(String s : Learner.finalWeight.keySet()){
				double fw = Learner.finalWeight.get(s);
				fw = fw* (i+1) + Learner.currentWeight.get(s);
				fw /= (i+2);
				Learner.finalWeight.put(s, fw);
			}
			
			Object[] keySet1 = currentWeight.keySet().toArray();
			java.util.Arrays.sort(keySet1);
			UIMan.println("#################ITERATION + " + i +"#################");
			for(Object ss : keySet1){
				String s = (String) ss;
				UIMan.println(s + "\t" + Learner.currentWeight.get(s) + "\t" +
						(_oldWeight.get(s)<currentWeight.get(s)? "larger\t" : "smaller\t") +
						Learner.finalWeight.get(s) + "\t" + 
						_mcsat.expectationOfViolation.get(s) + "->" + this.trainingViolation.get(s));
			}

			this.calcCurrentTrainingViolation();

			_oldWeight = (HashMap<String, Double>) currentWeight.clone();
			
		}
				
		Object[] keySet1 = currentWeight.keySet().toArray();
		java.util.Arrays.sort(keySet1);
		UIMan.println("#################FINAL WEIGHT#################");
		for(Object ss : keySet1){
			String s = (String) ss;
			UIMan.println(s + "\t" + Learner.currentWeight.get(s) + "\t" +
					Learner.finalWeight.get(s) + "\t" + 
					_mcsat.expectationOfViolation.get(s) + "->" + this.trainingViolation.get(s));
		}
		
		Timer.runStat.markInferDone();
		UIMan.println(">>> Writing answer to file: " + options.fout);
		this.dumpAnswers(options.fout);
		
		cleanUp();		
	}	
	
	/**
	 * Dump the learning result to file {@link CommandOptions#fout}.
	 * The format of this file is consistent with inference part.
	 */
	public void dumpAnswers(String fout){
		ArrayList<String> lines = new ArrayList<String>();
		DecimalFormat twoDForm = new DecimalFormat("#.####");
		
		HashSet<Predicate> allp = mln.getAllPred();
		for(Predicate p : allp){
			String s = "";
			if(p.isClosedWorld()){
				//System.out.print("*");
				s += "*";
			}
			//System.out.print(p.getName() + "(");
			s += p.getName() + "(";
			for(int i=0;i<p.arity();i++){
				//System.out.print(p.getTypeAt(i).name());
				s += p.getTypeAt(i).name();
				if(i!=p.arity()-1){
					//System.out.print(",");
					s += ",";
				}
			}
			//System.out.println(")");
			s += ")";
			lines.add(s);
		}
		lines.add("\n");
		//System.out.println();
		
		lines.add("//////////////AVERAGE WEIGHT OF ALL THE ITERATIONS//////////////");
		Object[] keySet = currentWeight.keySet().toArray();
		java.util.Arrays.sort(keySet);
		for(Object ss : keySet){
			String s = (String) ss;
			//System.out.println(s + "\t" + Learner.currentWeight.get(s) + ":" + 
			//		this.trainingSatisification.get(s) + "/" + this.trainingViolation.get(s) + 
			//		"\t" + Clause.mappingFromID2Desc.get(s));
			lines.add(twoDForm.format(Learner.finalWeight.get(s)) + " " + Clause.mappingFromID2Desc.get(s) + " //" + s);
		}
		
		lines.add("\n");
		
		lines.add("//////////////WEIGHT OF LAST ITERATION//////////////");
		keySet = currentWeight.keySet().toArray();
		java.util.Arrays.sort(keySet);
		for(Object ss : keySet){
			String s = (String) ss;
			//System.out.println(s + "\t" + Learner.currentWeight.get(s) + ":" + 
			//		this.trainingSatisification.get(s) + "/" + this.trainingViolation.get(s) + 
			//		"\t" + Clause.mappingFromID2Desc.get(s));
			lines.add("// " + twoDForm.format(Learner.currentWeight.get(s)) + " " + Clause.mappingFromID2Desc.get(s) + " //" + s);
		}
		
		FileMan.writeToFile(fout, StringMan.join("\n", lines));
	}
	

	/**
	 * Update {@link Learner#currentWeight} to new weights
	 * according to information provided by mcsat instance, e.g.,
	 * {@link MCSAT#expectationOfViolation}; This is a virtual
	 * function, each instance of {@link Learner} should materialize
	 * an adhoc version of this function. This function should return
	 * whether the learner thinks this iteration should terminate.
	 * 
	 * @return Whether terminate.
	 * @param mcsat MCSAT instance after this iteration.
	 */
	public abstract boolean updateWeight(MRF mcsat);
	
	/**
	 * Reading from training data and fill it into {@link MRF#atoms}.
	 * This function should be materialized by an instance
	 * of abstract class Learner.
	 * @param _mcsat The MCSAT object to be filled in.
	 * 
	 */
	public abstract void loadingTrainingData(MRF _mcsat);
	
	/**
	 * Calculate current training violations according to current sign of
	 * weight. This function will fill {@link Learner#trainingViolation}
	 * and {@link Learner#trainingSatisification}.
	 */
	public void calcCurrentTrainingViolation(){
		
		if(this.trainingViolation == null){
			this.trainingViolation = new HashMap<String, Long>();
			this.trainingSatisification = new HashMap<String, Long>();
		}
		
		for(String s : Learner.currentWeight.keySet()){
			if(Learner.currentWeight.get(s)>=0){
				this.trainingViolation.put(s, this.positiveWeightViolation.get(s));
				this.trainingSatisification.put(s, this.positiveWeightSatisfication.get(s));
			}else{
				this.trainingViolation.put(s, this.negativeWeightViolation.get(s));
				this.trainingSatisification.put(s, this.negativeWeightSatisfication.get(s));
			}
		}
	}
	
	/**
	 * Initialize weight according to the log odd of training data.
	 * 
	 * @param _mcsat MCSAT instance containing the vio/sat
	 * informaiton.
	 */
	public void fillInCurrentWeight(MRF _mcsat){
		for(String s : _mcsat.clauseVioTallies.keySet()){
			
			double priorodds = 1;
			double tc,fc;
			//positive clause
			if(Learner.currentWeight.get(s) > 0){
				tc = this.trainingSatisification.get(s);//_mcsat.clauseSatTallies.get(s);
				fc = this.trainingViolation.get(s);//_mcsat.clauseVioTallies.get(s);
				if(tc == 0)
					tc = 0.00001;
				if(fc == 0)
					fc = 0.00001;
				if(tc == fc)
					tc = fc + 0.001;
				Learner.finalWeight.put(s, (Math.log(tc/fc))- Math.log(priorodds));
				Learner.currentWeight.put(s, (Math.log(tc/fc))- Math.log(priorodds));
				
			}else{
				tc = this.trainingViolation.get(s);//_mcsat.clauseVioTallies.get(s);
				fc = this.trainingSatisification.get(s);//_mcsat.clauseSatTallies.get(s);
				if(tc == 0)
					tc = 0.00001;
				if(fc == 0)
					fc = 0.00001;
				if(tc == fc)
					tc = fc + 0.001;
				Learner.finalWeight.put(s, (Math.log(tc/fc)) - Math.log(priorodds));
				Learner.currentWeight.put(s, (Math.log(tc/fc))- Math.log(priorodds));
			}
		}
		// FOR TESTING ONLY
		if(Learner.currentWeight.containsKey("1.0"))
			this.odds = Learner.currentWeight.get("1.0");
	}
	
	// JUSTIFICATION OF THIS MEASURE SHOULD FOLLOW
	///**
	// * DEBUG ONLY!. This function measures the quality of learning. Instead of
	// * using pseudo-likelihood, it uses the overlapping between
	// * MAP inference in current weight-settings and training data.
	// * 
	// * @param mcsat MCSAT instance after this iteration.
	// * @param _mcsat MCSAT instance saving training data.
	// * @return the performance measure.
	// */
	/*public double validatingLearningPerformance(MCSAT mcsat, MCSAT _mcsat){
		mcsat.invalidateLowCost();
		mcsat.inferWalkSAT(options.maxTries, options.maxFlips);
		
		//double rs = 0;
		//double all = 0;
		double[][] rs = new double[2][2];
		rs[0][1] = 1;
		rs[1][1] = 1;
		rs[1][0] = 1;
		for(Integer aid : mcsat.atoms.keySet()){
			//System.out.println("\t~~~~~~~~~~~~~~" + aid + ":" + mcsat.atoms.get(aid).truth + 
			//		"\t" + mcsat.atoms.get(aid).delta());
			//if(mcsat.atoms.get(aid).truth == _mcsat.atoms.get(aid).truth)
			//	rs ++;
			rs[ mcsat.atoms.get(aid).truth == true? 0 : 1]
			   [_mcsat.atoms.get(aid).truth == true ? 0 : 1]++;
			//all ++;
		}
		
		double r = rs[0][0] / (rs[0][0]+rs[1][0]);
		double p = rs[0][0] / (rs[0][0]+rs[0][1]);
		return 2*p*r/(p+r);
		
		//return rs/all;
	}*/
	
	///**
	// * DEBUG ONLY.
	// */
	/*public double validatingDifferenceBetweenCurrentWeightAndRandomWeight(){
		double product;
		double all=0;
		double same=0;
		for(String s1 : currentWeight.keySet()){
			for(String s2 : currentWeight.keySet()){
				product = (currentWeight.get(s1) - currentWeight.get(s2)) * 
					(randomWeight.get(s1) - randomWeight.get(s2));
				all ++;
				same += Math.signum(product);
			}
		}
		return same / all;
	}*/
	
	///**
	// * DEBUG ONLY.
	// */
	/*public double validatingWeightCosine(){
		
		double cN = 0;
		double rN = 0;
		double rs = 0;
		
		for(String s : currentWeight.keySet())
			cN += currentWeight.get(s) * currentWeight.get(s);
		for(String s : randomWeight.keySet())
			rN += randomWeight.get(s) * randomWeight.get(s);
		
		for(String s : currentWeight.keySet())
			rs += currentWeight.get(s) * randomWeight.get(s);
		
		return rs/Math.sqrt(cN)/Math.sqrt(rN);
	}*/
	
}





